:- module(_, _, [bundlehooks]).

:- doc(title,  "Bundle Hooks for CiaoDE").
:- doc(author, "Jose F. Morales").
:- doc(author, "Edison Mera (first version)").

% ===========================================================================
% Bundle description

:- use_module(library(bundletool/paths_extra), [fsR/2]).

this_bundle_id(ciaode).
bundle_dname(ciaode, 'CiaoDE'). % the whole system...
%
bundle_manual_dir(ciaode) := _ :- fail.
bundle_readme_path(ciaode) := as('doc/readmes/README_CIAODE', 'README').

% ===========================================================================
% Backends for pbundle generation

% TODO: disable 'unused module' warnings
:- use_module(library(bundletool/pbundle_gen_win32)).
:- use_module(library(bundletool/pbundle_gen_rpm)).
:- use_module(library(bundletool/pbundle_gen_mac)).
:- use_module(library(bundletool/pbundle_gen_src)).
:- use_module(library(bundletool/pbundle_gen_bin)).
:- use_module(library(bundletool/pbundle_gen_common)).

% ---------------------------------------------------------------------------
% (lpmake targets for the pbundle generation)

% TODO: src and bin targets have WRONG names ('nothing' implies tgz and tbz)

% (hooks for gen_pbundle)
:- include(library(bundletool/pbundle_gen_hookdefs)).

'$bundle_own_make_target'(gen_pbundle__win32) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(win32, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__descfile) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(descfile, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__rpm) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(rpm, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__src) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(src, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__tgz) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(tgz, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__tbz) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(tbz, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__bin) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(bin, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__bin_tgz) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(bin_tgz, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__bin_tbz) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(bin_tbz, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__noa) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(noa, Bundle, []). 
'$bundle_own_make_target'(gen_pbundle__noa_tgz) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(noa_tgz, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__noa_tbz) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(noa_tbz, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__pkg) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(pkg, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__macport) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(macport, Bundle, []).
'$bundle_own_make_target'(gen_pbundle__app) :-
	this_bundle_id(Bundle),
	gen_pbundle_hook(app, Bundle, []).

% ============================================================================
% (lpmake interface)

'$bundle_own_make_target'(build_extra) :- rec_bundles_do(ciaode, extra, build_nodocs).

'$bundle_own_make_target'(build_nolibs_extra) :- rec_bundles_do(ciaode, extra, build_nolibs).

'$bundle_own_make_target'(build_extra_libraries) :- rec_bundles_do(ciaode, extra, build_libraries).

'$bundle_own_make_target'(build_platdep_extra) :- rec_bundles_do(ciaode, all, build_platdep).

% ============================================================================

% TODO: overwritten in bundletool_aux.pl
bundle_build_nodocs_hook(ciaode).

% TODO: overwritten in bundletool_aux.pl
bundle_build_platdep_hook(ciaode).

bundle_build_nolibs_hook(ciaode). % TODO: not for ciaode bundle
bundle_build_libraries_hook(ciaode). % TODO: not for ciaode bundle
bundle_prebuild_libraries_hook(ciaode). % TODO: not for ciaode bundle

bundle_prepare_docs_hook(ciaode).

bundle_register_hook(ciaode).
bundle_unregister_hook(ciaode).

bundle_install_hook(ciaode).
bundle_uninstall_hook(ciaode).

% Run bundle_custom_hook in all sub-bundles
bundle_custom_hook(Bundle, calibrateciaopp) :- Bundle = ciaode, !, bundle_do_hookcmd(ciaopp, calibrateciaopp).
bundle_custom_hook(Bundle, gen_timingmodel) :- Bundle = ciaode, !, bundle_do_hookcmd(ciao, gen_timingmodel).
bundle_custom_hook(Bundle, CustomAction) :- Bundle = ciaode, !, rec_bundles_do(Bundle, all, CustomAction).

% ============================================================================
% Miscallaneous actions

% ---------------------------------------------------------------------------

:- use_module(library(bundletool/bundle_aux), [ensure_builddir/0]).
:- use_module(library(bundle/bundle_info), [
	gen_bundle_commit_info/1]).

% TODO: Used from ciaobot
'$bundle_own_make_target'(gen_bundle_commit_info) :-
	this_bundle_id(Bundle),
	ensure_builddir,
	gen_bundle_commit_info(Bundle).

% ---------------------------------------------------------------------------

:- use_module(library(bundle/bundle_info), [show_bundles/0]).

% (available from the command line)
'$bundle_own_make_target'(show_bundles) :-
	show_bundles.

