:- module(_, _, [bundlehooks, dcg]).

:- doc(title,  "Bundle Hooks for Ciao").
:- doc(author, "Jose F. Morales").
:- doc(author, "Edison Mera (first version)").

% ============================================================================
% Bundle description

this_bundle_id(ciao).
bundle_dname(ciao, 'Ciao').
%
bundle_manual_dir(ciao) := as('doc/reference', 'ciao').
bundle_manual_dir(ciao) := as('doc/compiler', 'ciao_compiler').
bundle_manual_dir(ciao) := 'doc/bundletool'.
%
bundle_readme_path(ciao) := as('doc/common/INSTALLATION_CIAO', 'INSTALLATION').
bundle_readme_path(ciao) := as('doc/common/INSTALLATION_CIAO_Win32', 'INSTALLATION_Win32').
bundle_readme_path(ciao) := as('doc/common/README_CIAO', 'README').
% TODO: Strange, use a prepare_doc or remove 
bundle_readme_path(ciao) := 'doc/common/NewUser'.

% ===========================================================================

% TODO: refine dependencies, move to bundletool(bundle_actions_aux)
:- use_module(library(aggregates), [findall/3]).
:- use_module(library(write), [portray_clause/1, portray_clause/2]).
:- use_module(library(streams), [open_output/2, close_output/1]).
:- use_module(library(messages), [note_message/1]).
:- use_module(library(file_utils), [string_to_file/2]).

% ============================================================================

:- use_module(library(bundle/bundle_config), [bundle_config_value/3]).

% ============================================================================

% TODO: This should be part of bundletool, not config
:- use_module(library(bundletool/config_common), [
    instype/1,
    with_docs/1,
    get_installdir_lib/2,
    installdir_include/1,
    perms/1,
    get_platformdeb/2,
    build_mod_options/2,
    % (lpdoc bundle)
    docdir/1
]).
:- use_module(library(bundletool/bundle_configure), [
    foreign_config_var/3,
    foreign_config_version/2,
    set_prolog_flag_from_bundle_config/1
]).

:- use_module(library(bundletool/paths_extra), [fsR/2]).

% ===========================================================================

:- use_module(library(process), [process_call/3]).
:- use_module(library(system),
	[cyg2win/3,
	 winpath/2,
	 touch/1,
	 working_directory/2,
	 getenvstr/2,
	 file_exists/1,
	 copy_file/2, copy_file/3
	 ]).
:- use_module(library(system_extra),
	[try_finally/3, (-)/1,
	 del_file_nofail/1,
	 del_files_nofail/1,
	 set_file_perms/2]).
:- use_module(library(glob), [glob/3]).
:- use_module(library(pathnames)).

:- use_module(library(bundletool/bundle_aux), [
        b_copy/4,
        b_link/3,
        b_link_as/4,
        b_install_dir/1,
        b_install_file_exec/2,
        b_install_file_link_as/2
	]).
:- use_module(library(bundletool/ciaoc_aux), [
	% TODO: use bundle defs instead
        invoke_ciaosh_batch/1,
	build_mods/2,
 	compile_module_list/3,
	compile_platdep_modules/1,
	scan_platdep_modules/1]).
:- use_module(library(bundletool/register_in_script), [
 	register_in_script/3, unregister_from_script/2]).

% ============================================================================

:- doc(section, "Build").
% (engine, libraries, and compiler)

% (hook)
bundle_build_nodocs_hook(Bundle) :- Bundle = ciao, !,
	build_item(Bundle, exec_header), % TODO: why here?
	build_item(Bundle, ciaoc),
	build_item(Bundle, shell),
	build_item(Bundle, optim_comp),
	%
	build_etc,
	%
	bundle_lldo(Bundle, build_libraries),
	%
	build_item(Bundle, emacs_mode).

optimizing_compiler := ~bundle_config_value(ciao, 'OPTIMIZING_COMPILER').

% TODO: This is not the right way; integrate in the rest of the build
bundleitem_build_hook_decl(optim_comp).
bundleitem_build_hook(optim_comp) :-
	( optimizing_compiler(yes) ->
	    bold_message("Building 'optim_comp'", []),
	    process_call('./ciaotool', [build],
	           [cwd(~fsR(bundle_src(ciaode)/'optim_comp'))])
	; bold_message("Not building 'optim_comp'", [])
	).

% (invoked from 'bundle_actions__*', for each bundle)
bundle_build_libraries_hook(Bundle) :- Bundle = ciao, !,
        bundle_lldo(Bundle, prebuild_libraries),
	build_library.

% (invoked from 'bundle_actions__*', for each bundle)
bundle_build_nolibs_hook(Bundle) :- Bundle = ciao, !,
	bundle_lldo(Bundle, prebuild_libraries),
	bundle_lldo(Bundle, build_platdep).

% Libraries that require customized installation before they are built
% (such as automatically generated code, foreign interfaces, etc.)
% TODO: annotate which items need prebuild
bundle_prebuild_libraries_hook(Bundle) :- Bundle = ciao, !,
	prebuild_item(Bundle, [pillow, mathlibs, ppl, mathematica, persdb_mysql, timingmodel]).

% (invoked from 'bundle_actions__*', for each bundle)
bundle_build_platdep_hook(Bundle) :- Bundle = ciao, !,
	build_item(Bundle, exec_header), % TODO: why here?
	build_etc,
	build_tabling, % TODO: in build_library too, why?
	build_item(Bundle, emacs_mode),
	build_platdep_libs.

build_platdep_libs :-
	compile_platdep_modules(~build_mod_options(ciao)),
	% TODO: Scan in a better way (read .itf or after all is
	%   compiled and save for future use).
	%
	%   (original comment) This is fail prone, we update the .so
	%   list to ensure the next time the libraries containing
	%   externals are compiled, if the list has changed -- EMM.
	scan_platdep_modules(~findall(D, modulesdir(D))).

% TODO: extract from previous builditem and Manifest!
modulesdir(lib).
modulesdir(library).
modulesdir(contrib).

% ============================================================================

% (used only from 'ciaosetup')
'$bundle_own_make_target'(build_ciaoc) :- Bundle = ciao, build_item(Bundle, ciaoc).

bundleitem_def_decl_hook(ciaoc).
bundleitem_def_hook(ciaoc) :=
        standalone_list(ciao, bundle_src(ciao)/ciaoc, [
          'ciaoc'-[
            name="Standalone compiler",
            plexe,
	    bootstrap_ciaoc, static
          ]
        ]).
	
% (used only from 'ciaosetup')
'$bundle_own_make_target'(build_shell) :- build_shell.
build_shell :- Bundle = ciao, build_item(Bundle, shell).

bundleitem_def_decl_hook(shell).
bundleitem_def_hook(shell) :=
	standalone_list(ciao, bundle_src(ciao)/shell, [
          'ciaosh'-[plexe, name="Interactive toplevel", final_ciaoc],
          'ciao-shell'-[plexe, name="ciao-script runtime", final_ciaoc, static]
        ]).

% ===========================================================================

:- doc(section, "Windows-specific").

:- use_module(library(bundletool/bundle_params),
	[set_bundle_param_value/3]).

% TODO: review, this should not be necessary
% 'environment_and_windows_bats' required by runwin32.bat, CiaoDE.iss.skel
'$bundle_own_make_target'(environment_and_windows_bats) :-
	% TODO: move these param set operations to emacs_aux.pl
	set_bundle_param_value(ciao, 'EMACS_TYPE', 'Win32'), % TODO: strange
	set_bundle_param_value(ciao, 'WITH_EMACS_MODE', 'yes'), % TODO: strange
	environment,
	windows_bats.

environment :-
	Bundle = ciao,
	build_item(Bundle, emacs_mode),
	install_item(Bundle, emacs_mode), % (put ciao-mode-init.el in place)
	build_item(Bundle, exec_header). % TODO: why here?

win32_engine_path := F :-
	% TODO: Check this (see ~local_engine)
	% TODO: is this path correct? why not the installed engine?
	F = ~fsR(concat_k(exec, builddir(build)/objs/(~get_platformdeb(build))/'ciaoengine')).

windows_bats :-
	bold_message("Creating .bat files for the top-level and compiler", []),
	( % (failure-driven loop)
	  win_cmd_and_opts(BatCmd, EOpts, Opts, OrigCmd),
	    bat_file(BatCmd, OrigCmd, EOpts, Opts, BatFile, Tail),
	    %
	    open_output(BatFile, Out),
	    % TODO: missing quotation (e.g., of \")
	    display('@"'), display(~winpath(~win32_engine_path)), display('"'),
	    ( getenvstr('OS', "Windows_NT") ->
	        AllArgs = ' %*'
	    ; AllArgs = ' %1 %2 %3 %4 %5 %6 %7 %8 %9'
	    ),
	    display(AllArgs),
	    display(Tail),
	    nl,
	    close_output(Out),
	    fail
	; true
	).

% TODO: I do not know why '-i' is not necessary in 'ciaoc'
% TODO: why ciao_extra_commands?
win_cmd_and_opts(ciaosh, '', '-i', ciaosh).
win_cmd_and_opts(ciaoc, '', '', ciaoc).
win_cmd_and_opts(ciao, Atm, '-i', ciaosh) :- 
	atom_codes(Atm, ~ciao_extra_commands).

:- use_module(library(terms), [atom_concat/2]).

bat_file(BatCmd, OrigCmd, EOpts, Opts, BatFile, Tail) :-
	BatFile = ~fsR(concat_verk(ciao, ext('.bat'), builddir_bin(build)/(BatCmd))),
	OrigFile = ~fsR(concat_ver(ciao, builddir_bin(build)/(OrigCmd))),
	( EOpts = '' -> EOptsS = '' ; EOptsS = ' ' ),
	( Opts = '' -> OptsS = '' ; OptsS = ' ' ),
	Tail = ~atom_concat([' ', EOpts, EOptsS,
	                     '-C ', Opts, OptsS,
			     '-b \"', OrigFile, '\"']).

:- doc(subsection, "Special header for Windows").
% TODO: merge with do_exe_header code?

bundleitem_build_hook_decl(exec_header).
bundleitem_build_hook(exec_header) :-
	make_exec_header.

% TODO: see builddir_aux.sh:do_exec_header_for_builddir (almost duplicated)
make_exec_header :-
	( get_os('Win32') ->
	    % TODO: why?
	    Eng = ~win32_engine_path
	; Eng = ~fsR(concat_k(exec, ~installdir_engine/(~name_ciaoengine)))
	),
	%
	bold_message("Executables assume engine at ~w", [Eng]),
	%
	HeaderPath = ~fsR(bundle_src(ciaode)/'ciao'/'lib'/'compiler'/'header'),
	%
	open_output(HeaderPath, Out),
	display('#!/bin/sh\n'),
	display_list(['INSTENGINE=\"', Eng, '\"\n']),
	display('ENGINE=${CIAOENGINE:-${INSTENGINE}}\n'),
	display('exec "$ENGINE" "$@" -C -b "$0"\n'),
	put_code(0xC), % ^L control character
	display('\n'),
	close_output(Out).

% ===========================================================================

:- doc(section, "Compilation of Libraries").

% TODO: Kludge: defaulttype is always dyn, see ENGINE_LINK_TYPE variable in
%       config-sysdep.sh

% TODO: Rename, bad name
:- meta_predicate proc_if_dir_exists(?, goal).
proc_if_dir_exists(Dir, Goal) :-
	NCDir = ~fsR(Dir/'NOCOMPILE'), % TODO: full path?
	( file_exists(Dir), \+ file_exists(NCDir) -> call(Goal)
	; true
	).

% ===========================================================================

build_mods(BaseDir) :-
	build_mods(BaseDir, ~build_mod_options(ciao)).

build_library :-
	build_mods('lib'),
	build_tabling, % TODO: in build_platdep too, why?
	%
	build_mods('library/clpq'),
	build_mods('library/clpr'),
	build_javalibs,
	%
	build_mods('library'),
	%
	build_toplevel,
	%
	build_contrib.

build_toplevel :-
	Dir = 'library/toplevel/',
	FileName = 'library/toplevel/toplevel__scope.pl',
	compile_module_list([m(_, _, FileName)], Dir, ~build_mod_options(ciao)).

build_contrib :-
	build_mods('contrib').

% (called from 'ciaosetup' (for ciaobot))
'$bundle_own_make_target'(build_profiler) :- build_profiler.
build_profiler :- build_mods('contrib/profiler').

tabled_execution := ~bundle_config_value(ciao, 'TABLED_EXECUTION').

build_tabling :-
	( % We can not compile this if tabled execution is disabled
	  tabled_execution(no) ->
	    true
	; build_mods('contrib/tabling')
	).

% ---------------------------------------------------------------------------

:- doc(section, "CiaoJava bundle").

javalibs_dir := bundle_src(ciao)/library/javall.
with_java_interface := ~bundle_config_value(ciao, 'WITH_JAVA_INTERFACE').

% TODO: missing target for 'build_docs'
build_javalibs :-
	( with_java_interface(yes) ->
	    invoke_gmake_javall(build),
	    % TODO: need hooks for foreign docs
	    ( with_docs(yes) ->
	        invoke_gmake_javall(docs)
	    ; true
	    )
	; true
	).

'$bundle_own_make_target'(clean_javalibs) :- clean_javalibs.
clean_javalibs :-
	( with_java_interface(yes) ->
	    bold_message("Cleaning CiaoJava...", []),
	    invoke_gmake_javall(distclean) % TODO: 'clean' or 'distclean'?
	; true
	).

:- use_module(library(bundletool/bundle_aux), [invoke_gmake/2]).

invoke_gmake_javall(Cmd) :-
	invoke_gmake(~fsR(~javalibs_dir),
	             [~atom_concat('CIAOBUILDDIR=', ~fsR(builddir(build))),
		      ~atom_concat('CIAODESRC=', ~fsR(bundle_src(ciaode))),
		      Cmd]).

% ---------------------------------------------------------------------------

:- doc(section, "timingmodel bundle").
% TODO: split in two bundles: miniprolog and timingmodel

timingmodel_dir := bundle_src(ciao)/contrib/timingmodel.

bundleitem_prebuild_hook_decl(timingmodel).
bundleitem_prebuild_hook(timingmodel) :-
	proc_if_dir_exists(~fsR(~timingmodel_dir), do_timingmodel),
 	% TODO: is this call necessary? (do_timingmodel does it)
	copy_mp_auto.

% (called from 'ciaosetup', called from 'ciaobot' as a test)
% (hook)
bundle_custom_hook(ciao, gen_timingmodel) :- !,
	do_timingmodel.

timingmodel_cmd := bench|estimate.

do_timingmodel :-
	% TODO: redundant check?
	( file_exists(~fsR(~timingmodel_dir)) -> do_timingmodel_ ; true ).

do_timingmodel_ :-
	bold_message("Compiling mini prolog engine"),
	invoke_gmake_miniprolog(all),
	copy_file(~fsR(~timingmodel_dir/'miniprolog'/'bin'/'timingmodel_auto.pl'),
	          ~fsR(~timingmodel_dir/'timingmodel_pre.pl'),
		  [overwrite]),
	%
	bold_message("Generating timing model for mini prolog"),
	( % (failure-driven loop)
	  timingmodel_cmd(Cmd),
	    invoke_gmake_timingmodel(Cmd),
	    fail
	; true
	).

% This extra step is to ensure the generation of timingmodel_auto.pl
% even if miniprolog has not been configured
copy_mp_auto :-
	Orig = ~fsR(~timingmodel_dir/'timingmodel_pre.pl'),
	( file_exists(Orig) ->
	    copy_file(Orig,
	              ~fsR(~timingmodel_dir/'timingmodel_auto.pl'),
		      [overwrite])
	; true
	).

benchmp :-
	bold_message("Running timing model benchmarks"),
	invoke_gmake_timingmodel('bench').

estimatemp :-
	bold_message("Running timing model estimate"),
	invoke_gmake_timingmodel('estimate').

% (called from 'ciaosetup')
'$bundle_own_make_target'(clean_timingmodel) :- clean_timingmodel.
clean_timingmodel :-
	bold_message("Cleaning timingmodel"),
	invoke_gmake_miniprolog(clean),
	invoke_gmake_timingmodel(clean).

:- use_module(library(bundletool/bundle_aux), [invoke_gmake/2]).

invoke_gmake_miniprolog(Cmd) :-
	invoke_gmake(~fsR(~timingmodel_dir/'miniprolog'),
	             ['-s', '-j1',
		      ~atom_concat('MPARCH=', ~get_platform),
		      Cmd]).

invoke_gmake_timingmodel(Cmd) :-
	invoke_gmake(~fsR(~timingmodel_dir),
	             ['-s', '-j1',
		      ~atom_concat('MPARCH=', ~get_platform),
		      ~atom_concat('CIAOBUILDDIR=', ~fsR(builddir(build))),
		      ~atom_concat('CIAODESRC=', ~fsR(bundle_src(ciaode))),
		      Cmd]).

% ---------------------------------------------------------------------------

:- doc(section, "MySQL bundle").

:- use_module(library(bundletool/bundle_aux), [wr_template/4]).

with_mysql := ~bundle_config_value(ciao, 'WITH_MYSQL').
mysql_client_directory := ~bundle_config_value(ciao, 'MYSQL_CLIENT_DIRECTORY').

bundleitem_prebuild_hook_decl(persdb_mysql).
bundleitem_prebuild_hook(persdb_mysql) :-
	( with_mysql(yes) ->
	    bold_message("Configuring MySQL Libraries"),
	    wr_template(origin,
	        bundle_src(ciao)/'library'/(~mysql_directory),
	        'linker_opts_auto.pl',
	        ['where_mysql_client_lives' = ~mysql_client_directory]),
	    % TODO: why?
	    ( file_exists(~fsR(bundle_src(ciao)/library/(~mysql_directory_op))) ->
	        wr_template(origin,
		    bundle_src(ciao)/'library'/(~mysql_directory_op),
		    'linker_opts_auto.pl',
		    ['where_mysql_client_lives' = ~mysql_client_directory])
	    ; true
	    )
	; true
	).

mysql_directory := 'persdb_mysql'.
mysql_directory_op := 'persdb_mysql_op'.

% ---------------------------------------------------------------------------

:- doc(section, "PiLLoW bundle").

webimagespath := ~bundle_config_value(ciao, 'WEB_IMAGES_PATH').
webimagesurl := ~bundle_config_value(ciao, 'WEB_IMAGES_URL').

pillow_dir := bundle_src(ciao)/'library'/'pillow'.

bundleitem_prebuild_hook_decl(pillow).
bundleitem_prebuild_hook(pillow) :-
	icon_address_auto.

icon_address_auto :-
	try_finally(
	    open(~fsR(~pillow_dir/'icon_address_auto.pl'), write, OS),
	    portray_clause(OS, icon_base_address(~webimagesurl)),
	    close(OS)
	).

bundleitem_def_decl_hook(pillow_images).
bundleitem_def_hook(pillow_images) :=
  group("PiLLoW images", 
    dir(~webimagespath, [do_not_del, files_from(~pillow_dir/images)])).

:- doc(bug, "The current installation method do not uninstall the
   installed PiLLoW images. Fix.").

% ---------------------------------------------------------------------------

:- doc(section, "GSL bundle").
% TODO: Share code for PPL, GMP, GSL.

% Specification of GSL (third-party component)
:- def_third_party(gsl, [
    version('1.16'),
    source_url('http://ftp.gnu.org/gnu/gsl/gsl-1.16.tar.gz'),
    source_md5("e49a664db13d81c968415cd53f62bc8b"),
    %
    build_system(gnu_build_system)
]).

:- use_module(library(llists), [flatten/2]).

gsl_dir := bundle_src(ciao)/'contrib'/'gsl_imports'.
with_gsl := ~bundle_config_value(ciao, 'WITH_GSL').
auto_install_gsl := ~bundle_config_value(ciao, 'AUTO_INSTALL_GSL').

% TODO: rename by 'gsl' (no mathlibs)?
bundleitem_prebuild_hook_decl(mathlibs).
bundleitem_prebuild_hook(mathlibs) :-
	do_auto_install_gsl,
	prebuild_gsl_bindings.

:- use_module(library(bundletool/third_party_install), [auto_install/1]).

do_auto_install_gsl :-
	( auto_install_gsl(yes) -> 
	    bold_message("Auto-install GSL (third party)"),
	    third_party_install:auto_install(gsl)
	; true
	).

prebuild_gsl_bindings :-
	( with_gsl(yes) ->
	    bold_message("Configuring GSL Library"),
	    S = ":- use_module(library(gsl_imports)).\n",
 	    foreign_config_var(gsl, 'cflags', CompilerOpts),
 	    foreign_config_var(gsl, 'libs', LinkerOpts0),
	    fix_linker_opts(LinkerOpts0, LinkerOpts1),
	    (
		% If gsl is installed as a third party, add ./third-party/lib
		% to  the runtime library search path
		auto_install_gsl(yes) ->
		% TODO: better way to compute RelativeLibDir.
		fsR(bundle_src(ciaode),  CIAOSRC),
		third_party_path(libdir, LibDir),
		path_relocate(CIAOSRC, '.', LibDir, RelativeLibDir),
		atom_codes(RelativeLibDir, RelativeLibDir_),
		append("-Wl,-rpath,"||RelativeLibDir_, " "||LinkerOpts1, LinkerOpts)
	    ;
	        LinkerOpts1 = LinkerOpts
	    ),
	    T = ~flatten([
		    ":- extra_compiler_opts(\'"||CompilerOpts, "\').\n"||
		    ":- extra_linker_opts(\'"||LinkerOpts, "\').\n"]),
	    string_to_file(T, ~fsR(~gsl_dir/'gsl_imports_decl_auto.pl')),
	    % List of static libraries from GSL
	    % TODO: generalize for any other library
	    M = ~flatten(["GSL_STAT_LIBS=\'"||LinkerOpts, "\'\n"])
	;
	    LinkerOpts = "",
	    bold_message("Ignoring GSL Library"),
	    S = ":- use_module(library(gsl_imports/gsl_imports_dummy)).\n",
	    M = ""
	),
	% TODO: Simplify
	% TODO: Options of CONFIG_GSL are included in the final builddir(build)/'CONFIG_mkf' file
        % TODO: Those options are not immediately used, correct?
	string_to_file(M, ~fsR(builddir(build)/'CONFIG_GSL')),
	string_to_file(S, ~fsR(~gsl_dir/'gsl_imports_auto.pl')).

:- use_module(library(lists), [append/3]).

% Remove the -L option, hack that allows to run in LINUXi86_64 --EMM:
fix_linker_opts(LinkerOpts0, LinkerOpts) :-
	get_platform('LINUXi86_64'),
	append("-L"||_, " "||LinkerOpts, LinkerOpts0),
	!.
fix_linker_opts(LinkerOpts, LinkerOpts).

% ---------------------------------------------------------------------------

:- doc(section, "PPL (Parma Polyhedra Library) bundle").
% TODO: Share code for PPL, GMP, GSL.

% TODO: Look at PPL-related code in bundle_actions__ciaopp.pl

% Specification of M4 (third-party component)
:- def_third_party(m4, [
    version('1.4.17'),
    source_url('http://ftp.gnu.org/gnu/m4/m4-1.4.17.tar.bz2'),
    source_md5("8a1787edcba75ae5cd1dc40d7d8ed03a"),
    %
    build_system(gnu_build_system)
]).

% Specification of GMP (third-party component)
:- def_third_party(gmp, [
    version('5.0.5'),
    source_url('https://gmplib.org/download/gmp/gmp-5.0.5.tar.bz2'),
    source_md5("041487d25e9c230b0c42b106361055fe"),
    %
    build_system(gnu_build_system),
    option1('enable-cxx') % for PPL
]).

% Specification of PPL (third-party component)
:- def_third_party(ppl, [
    version('1.0'),
    source_url('http://bugseng.com/products/ppl/download/ftp/releases/1.0/ppl-1.0.tar.bz2'),
    source_md5("8a90e0b0b3e9527609a6e5ed3616fab1"),
    %
    build_system(gnu_build_system),
    option2('enable-interfaces','c cxx')
]).

:- use_module(library(lists), [append/3]).
:- use_module(library(llists), [flatten/2]).

ppl_dir := bundle_src(ciao)/'contrib'/'ppl'.
with_ppl := ~bundle_config_value(ciao, 'WITH_PPL').
auto_install_ppl := ~bundle_config_value(ciao, 'AUTO_INSTALL_PPL').

remove_all_substrings(Input, String, Output):-
	append(String, Postfix, Input), !,
	remove_all_substrings(Postfix, String, Output).
remove_all_substrings([H|T1], String, [H|T2]):-
	remove_all_substrings(T1, String, T2).
remove_all_substrings([], _String, []).

ppl_version(Version) :-
	foreign_config_version(ppl, Version).

bundleitem_prebuild_hook_decl(ppl).
bundleitem_prebuild_hook(ppl) :-
	do_auto_install_ppl,
	prebuild_ppl_bindings.

:- use_module(library(bundletool/third_party_install), [auto_install/1, third_party_path/2]).

do_auto_install_ppl :-
	( auto_install_ppl(yes) -> 
	    % TODO: add dependencies between PPL and GMP
	    % bold_message("Auto-install M4 (third party)"),
	    % third_party_install:auto_install(m4),
	    bold_message("Auto-install GMP (third party)"),
	    third_party_install:auto_install(gmp),
	    bold_message("Auto-install PPL (third party)"),
	    third_party_install:auto_install(ppl)
	; true
	).

prebuild_ppl_bindings :-
	( with_ppl(yes) ->
	    bold_message("Configuring PPL Interface"),
 	    foreign_config_var(ppl, 'cppflags', CompilerOpts1),
 	    foreign_config_var(ppl, 'cxxflags', CompilerOpts2),
	    append(CompilerOpts1, " "||CompilerOpts2, CompilerOpts3),
 	    foreign_config_var(ppl, 'ldflags', LinkerOpts1),
            ( get_platform('DARWINi86') ->
              % Remove "-arch x86_64" option if ppl is an MacOS universal binary
                remove_all_substrings(CompilerOpts3, "-arch x86_64", CompilerOpts),
                remove_all_substrings(LinkerOpts1, "-arch x86_64", LinkerOpts2) ;
                CompilerOpts3 = CompilerOpts, LinkerOpts1 = LinkerOpts2
            ),list(CompilerOpts),
	    (
		% If ppl is installed as a third party, add ./third-party/lib
		% to the runtime library search path
		auto_install_ppl(yes) ->
		% Better way to compute RelativeLibDir
		fsR(bundle_src(ciaode),  CIAOSRC),
		third_party_path(libdir, LibDir),
		path_relocate(CIAOSRC, '.', LibDir, RelativeLibDir),
		atom_codes(RelativeLibDir, RelativeLibDir_),
		append("-Wl,-rpath,"||RelativeLibDir_, " "||LinkerOpts2, LinkerOpts3)
	    ;
	        LinkerOpts2 = LinkerOpts3
	    ),
        append(LinkerOpts3, " -lstdc++", LinkerOpts),
	    flatten(["%Do not edit generated automatically\n\n",
		    ":- extra_compiler_opts('", CompilerOpts, "').\n",
		    ":- extra_linker_opts('", LinkerOpts, "').\n"], T),
	    % TODO: generalize, share with GSL
	    string_to_file(T, ~fsR(~ppl_dir/'ppl_decl_auto.pl')),
	    ppl_version(Version),
	    ( Version @< [0, 9] ->
		fail
	    ; Version @< [0, 10] ->
		ppl_interface_version("0_9")
	    ;  Version @< [1, 0] ->
		ppl_interface_version("0_10")
	    ; ppl_interface_version("1_0")
	    )
	;
	    string_to_file(
		":- initialization(error('PPL library not installed')).\n\n",
		~fsR(~ppl_dir/'ppl_auto.pl'))
	).

% This selects one of the two versions
% TODO: maybe we can use the package 'condcomp' to make this simpler?
ppl_interface_version(StrVer) :-
	S = ~flatten([
           "%Do not edit generated automatically.\n\n"||
           ":- include(library(ppl/'", StrVer, "'/ppl_ciao)).\n"]),
 	string_to_file(S, ~fsR(~ppl_dir/'ppl_auto.pl')).

% TODO: Clean code is missing here

% ---------------------------------------------------------------------------

:- doc(section, "Mathematica bundle").
% TODO: There is even more code in the bundle_actions__ciaopp.pl files!

:- use_module(library(llists), [flatten/2]).

mathematica_dir := bundle_src(ciao)/'contrib'/'mathematica'.
with_mathematica := ~bundle_config_value(ciao, 'WITH_MATHEMATICA').

bundleitem_prebuild_hook_decl(mathematica).
bundleitem_prebuild_hook(mathematica) :-
	( with_mathematica(yes) ->
	    bold_message("Configuring Mathematica Interface"),
	    foreign_config_var(mathematica, 'mathlink', MathLink),
 	    foreign_config_var(mathematica, 'cppflags', CompilerOpts),
 	    foreign_config_var(mathematica, 'ldflags', LinkerOpts),
	    T = ~flatten(["%Do not edit generated automatically\n\n",
		    "mathematica_link_path('", MathLink, "').\n",
		    ":- extra_compiler_opts('", CompilerOpts, "').\n",
		    ":- extra_linker_opts('", LinkerOpts, "').\n"]),
	    % TODO: share with GSL
	    string_to_file(T, ~fsR(~mathematica_dir/'mathematica_decl_auto.pl')),
	    string_to_file(
	        ":- reexport(library(mathematica/mathematica_link)).\n", 
		~fsR(~mathematica_dir/'mathematica_auto.pl'))
	;
	    string_to_file(
		":- initialization(error('Mathematica interface not installed')).",
		~fsR(~mathematica_dir/'mathematica_auto.pl'))
	).

% TODO: Clean code is missing here


% ============================================================================

:- doc(section, "Documentation").

bundle_prepare_docs_hook(ciao) :-
	prepare_docs_emacs_hook. % (generates a .lpdoc with the emacs documentation)

% ============================================================================

:- doc(section, "Register in the System").
% TODO: Should this be just a subtask in the installation? Users
%       should not invoke it...
% TODO: This should be done for each anchor (bash, csh, emacs, etc.)
%       and each bundle

% Modifies the .bashrc/.cshrc/.emacs files to let Ciao run from the
% installed lib files.
bundle_register_hook(ciao) :-
        bundleitem_register_hook(bashrc),
	bundleitem_register_hook(cshrc),
	bundleitem_register_hook(emacs_mode).

% Leaves the .bashrc/.cshrc/.emacs file in its original state.
bundle_unregister_hook(ciao) :-
        bundleitem_unregister_hook(bashrc),
	bundleitem_unregister_hook(cshrc),
	bundleitem_unregister_hook(emacs_mode).

% ============================================================================

:- doc(section, "Installation").

bundle_install_hook(Bundle) :- Bundle = ciao, !,
	install_item(Bundle, ~ciao_desc).

bundle_uninstall_hook(Bundle) :- Bundle = ciao, !,
	uninstall_item(Bundle, ~ciao_desc).

ciao_desc := 
  big_group("Ciao", [
    only_global_ins([
      % TODO: use the following two lines for all bundles
      dir(installdirbase(ciao), [del_if_empty]),
      dir(installdir_lib(ciao), [del_rec]), % why not del_if_empty?
      % TODO: use bundle_reg_dir/2
      % TODO: really? installation rewrites the bundlereg of each installed bundle
      dir(installdir_lib(ciao)/lib/bundlereg__auto, [do_not_del]),
      %
      engine,
      ciaoc,
      shell,
      etc,
      emacs_mode,
      %
      lib(ciao, 'lib'),
      lib(ciao, 'library'),
      pillow_images,
      lib(ciao, 'contrib'),
      src(ciao, 'examples')
    ]),
    docs(ciao)
  ]).

% ===========================================================================
% Engine installation/uninstallation

% TODO: Make sure that CIAOHDIR points to the right place when the engine 
%   is installed in instype=global

% TODO: move to ciaode_config_options
:- use_module(library(bundletool/config_common), [
	name_ciaoengine/1,
	name_ciaoengine_arch/1,
	installdir_engine/1]).
	
% (only for instype=global)
bundleitem_install_hook_decl(engine).
bundleitem_install_hook(engine) :-
	Bundle = ciao,
	bold_message("Installing engine and C header files for ~w", [~get_platform]),
	%
	install_item(Bundle, [
          dir(~fsR(installdir_lib(ciao))),
	  dir(~installdir_engine) % set_file_perms or set_exec_perms?
        ]),
	b_install_file_exec(~fsR(concat_k(exec, builddir(build)/objs/(~get_platformdeb(build))/(~name_ciaoengine))), ~fsR(concat_k(exec, ~installdir_engine/(~name_ciaoengine_arch)))),
	b_install_file_link_as(concat_k(exec, (~name_ciaoengine_arch)), ~fsR(concat_k(exec, ~installdir_engine/(~name_ciaoengine)))),
	%
	install_c_headers.

install_c_headers :-
	Bundle = ciao,
	b_install_dir(~fsR(installdir_lib(ciao)/include)), % perms?
	SrcDir = builddir(build)/include/(~get_platformdeb(build)),
	install_item(Bundle, [
          dir(installdir_lib(ciao)/include/(~get_platformdeb(build)), [files_from(SrcDir, '*'), del_rec])
        ]),
	% installdir_include
	% TODO: This does not allow several versions installed at the same time
	b_install_dir(~installdir_include),
	Header = 'ciao_prolog.h',
	b_install_file_link_as(~fsR(installdir_lib(ciao)/include/(~get_platformdeb(build))/(Header)),
	                       ~installdir_include/(Header)).

% TODO: (hook)
% (only for instype=global)
bundleitem_uninstall_hook_decl(engine).
bundleitem_uninstall_hook(engine) :-
	Bundle = ciao,
	bold_message("Uninstalling engine and C header files for ~w", [~get_platform]),
	uninstall_item(Bundle, [
	  dir(installdir_lib(ciao)/include/(~get_platformdeb(build)), [del_rec]),
	  file(~installdir_include/'ciao_prolog.h'),
	  %
	  file(~fsR(concat_k(exec, ~installdir_engine/(~name_ciaoengine_arch)))),
	  file(~fsR(concat_k(exec, ~installdir_engine/(~name_ciaoengine))))
        ]).
%	rm -f ${ROOTPREFIX} ~m_installed_ciaoengine_sta,
%	rm -f ${ROOTPREFIX} ~m_installed_ciaoengine_arch_sta.

% ===========================================================================

:- doc(section, "Shell Script Configuration").

% The configuration of shell scripts defines the necessary environment
% variables to make the system locate the installed version of Ciao
% code and binaries (executables, libraries, documentation, etc.) in
% Unix-like systems.

update_bashrc := ~bundle_config_value(ciao, 'UPDATE_BASHRC').
dotbashrc := ~bundle_config_value(ciao, 'DOTBASHRC').

bundleitem_register_hook(bashrc) :- !,
	( update_bashrc(yes) ->
	    (-register_in_script(~dotbashrc, "#", ~bashrc_lines))
	; true
	).
bundleitem_unregister_hook(bashrc) :- !,
	( update_bashrc(yes) ->
	    (-unregister_from_script(~dotbashrc, "#"))
	; true
	).

update_cshrc := ~bundle_config_value(ciao, 'UPDATE_CSHRC').
dotcshrc := ~bundle_config_value(ciao, 'DOTCSHRC').

bundleitem_register_hook(cshrc) :- !,
	( update_cshrc(yes) ->
	    (-register_in_script(~dotcshrc, "#", ~cshrc_lines))
	; true
	).
bundleitem_unregister_hook(cshrc) :- !,
	( update_cshrc(yes) ->
	    (-unregister_from_script(~dotcshrc, "#"))
	; true
	).

bashrc_lines(S) :-
	ciaolibetcsrc(LibRoot),
	shell_config_code(bash, LibRoot, S, []).
cshrc_lines(S) :-
	ciaolibetcsrc(LibRoot),
	shell_config_code(csh, LibRoot, S, []).

% TODO: Obtain the path from the code that installs the configuration
%       files!
ciaolibetcsrc(LibRoot) :-
	Bundle = ciao,
	( instype(local) ->
	    LibRoot = ~fsR(bundle_src(Bundle)/'etc')
	; LibRoot = ~fsR(installdirbase(Bundle))
	).

% Configuration code for the shell script interpreters
shell_config_code(bash, LibRoot) -->
	"if [ -f ", emit_atom(LibRoot), "/DOTprofile ] ; then\n"||
	"  . ", emit_atom(LibRoot), "/DOTprofile\n"||
	"fi\n".
shell_config_code(csh, LibRoot) -->
	"if ( -e ", emit_atom(LibRoot), "/DOTcshrc ) then\n"||
	"  source ", emit_atom(LibRoot), "/DOTcshrc\n"||
	"endif\n".

:- use_module(library(lists), [append/3]).

% (emit an atom codes in a DCG)
emit_atom(X, S, S0) :-
	atom_codes(X, Codes),
	append(Codes, S0, S).

% ===========================================================================

:- doc(section, "Tests and Benchmarks").
% TODO: Add bundle defs for unit tests, integration tests, regression
%   tests, etc.

% Run tests
bundle_custom_hook(ciao, runtests) :- !,
	bold_message("Running Ciao unit tests"),
	runtests_unittests_hook,
	bold_message("Running Ciao integration tests"),
	runtests_ciaotests_hook,
	bold_message("Running Ciao acceptance tests"),
	runtests_isotests_hook.

% Run Ciao unit tests
runtests_unittests_hook :-
	invoke_run_test_dir(~fsR(bundle_src(ciao)/'lib')),
	invoke_run_test_dir(~fsR(bundle_src(ciao)/'library')),
	invoke_run_test_dir(~fsR(bundle_src(ciao)/'contrib')).

invoke_run_test_dir(Dir) :-
	invoke_ciaosh_batch([
          use_module(library(unittest), [run_test_dir/2]),
	  run_test_dir(Dir, [rtc_entry])
	]).

% Run Ciao integration tests
runtests_ciaotests_hook :-
	proc_if_dir_exists(~fsR(bundle_src(ciao)/'tests'), do_ciaotests).

do_ciaotests :-
	working_directory(ThisDir, tests),
	invoke_ciaosh_batch([
	  use_module(ciao_tests(run_tests), [run_tests/0]),
	  run_tests:run_tests
	]),
	working_directory(_, ThisDir).

% Run ISO-prolog tests
runtests_isotests_hook :-
	IsoTestsDir = ~fsR(bundle_src(ciao)/'contrib'/'iso_tests'),
	proc_if_dir_exists(IsoTestsDir, do_isotests(IsoTestsDir)).

do_isotests(IsoTestsDir) :-
	bold_message("Running ISO-prolog tests"),
	invoke_run_test_dir(IsoTestsDir).

% Run benchmarks
bundle_custom_hook(ciao, runbenchmarks) :- !,
	invoke_ciaosh_batch([
	  use_module(library(benchmarks/ecrc), [main/1]),
	  ecrc:main([])
        ]).

% TODO: also include (or merge) these:
%   optim_comp/testsuite/multisystemtests/quick_test.sh
%   ciao/examples/misc/ (see Makefile) -- too small, need at least scaling

% ============================================================================

:- doc(section, "Emacs Mode").

:- use_module(library(bundle/bundle_info),
	[enum_bundle/2,
	 bundle_version/2,
	 bundle_version_patch/2,
	 bundle_manual_base/2]).

:- use_module(library(bundletool/emacs_aux), [
	emacsstylepath/3,
	emacs_type/1,
        emacs_batch_call/3,
	emacs_batch_byte_compile/3,
	emacs_update_autoloads/3,
        clean_log/2]).

emacsmode_dir := bundle_src(ciaode)/'emacs-mode'.

% TODO: Why? 
script_extension('Win32', '.bat') :- !.
script_extension(_,       '').

% Here is how this all works: 
% 
% - During build and install of 'ciao/emacs_mode'
%
%   * The ciao-config.el.skel file is filled with configuration
%     parameters for the installed system to produce ciao-config.el,
%     which is installed in the libraries. The current Ciao version is
%     also included in ciao-config.el at this time.
%
%   * All .el files in the libraries are byte-compiled.
%    
% - Generating the documentation ('ciaosetup build_docs'):
%
%   * CiaoMode.lpdoc is generated from ciao-documentation.el using
%     emacs (contains and extracts the documentation for all the elisp
%     functions). CiaoMode.pl is included as a chapter in the Ciao
%     manual.

% TODO: Move to a specific installation for emacs-mode
% (used only from 'ciaosetup')
'$bundle_own_make_target'(build_emacs_mode) :- Bundle = ciao, build_item(Bundle, emacs_mode).
'$bundle_own_make_target'(install_emacs_mode) :-
	Bundle = ciao,
        install_item(Bundle, ~emacs_mode_desc).
'$bundle_own_make_target'(uninstall_emacs_mode) :-
	Bundle = ciao,
        uninstall_item(Bundle, ~emacs_mode_desc).

emacs_mode_desc := 
  big_group("Emacs mode for Ciao", [
    only_global_ins([
      emacs_mode
    ])
  ]).

% TODO: Each bundle should be able to register parts of the CiaoMode
%       That will solve @bug{lpdoclibdir_emacs_mode}
bundleitem_build_hook_decl(emacs_mode).
bundleitem_build_hook(emacs_mode) :-
	build_emacs_mode.

% ---------------------------------------------------------------------------
% (in order to share code, make the emacs kind a parameter)

bundle_param_or_config_value(Bundle, Name, Value) :-
	bundle_param_value(Bundle, Name, Value),
	!.
bundle_param_or_config_value(Bundle, Name, Value) :-
	bundle_config_value(Bundle, Name, Value).

with_emacs_mode := ~bundle_param_or_config_value(ciao, 'WITH_EMACS_MODE').
with_xemacs_mode := ~bundle_param_or_config_value(ciao, 'WITH_XEMACS_MODE').

with_emacs_mode_k(emacs) :- with_emacs_mode(yes).
with_emacs_mode_k(xemacs) :- with_xemacs_mode(yes).

update_dotemacs := ~bundle_config_value(ciao, 'UPDATE_DOTEMACS').
update_dotxemacs := ~bundle_config_value(ciao, 'UPDATE_DOTXEMACS').

update_dotemacs_k(emacs) :- update_dotemacs(yes).
update_dotemacs_k(xemacs) :- update_dotxemacs(yes).

dotemacs := ~bundle_config_value(ciao, 'DOTEMACS').
dotxemacs := ~bundle_config_value(ciao, 'DOTXEMACS').

dotemacs_k(emacs) := ~dotemacs.
dotemacs_k(xemacs) := ~dotxemacs.

with_any_emacs_mode :-
        ( with_emacs_mode_k(emacs) -> true
	; with_emacs_mode_k(xemacs)
	).

emacs_site_start := ~bundle_config_value(ciao, 'EMACS_SITE_START').
xemacs_site_start := ~bundle_config_value(ciao, 'XEMACS_SITE_START').

emacs_site_start_k(emacs) := ~emacs_site_start.
emacs_site_start_k(xemacs) := ~xemacs_site_start.

% ---------------------------------------------------------------------------

:- use_module(library(llists), [flatten/2]).

% TODO: keep filters in library(source_tree) synchronized (e.g., more files added here)
build_emacs_mode :-
	\+ with_any_emacs_mode,
	!,
	note_message("Emacs support will not be installed").
build_emacs_mode :-
	% TODO: Use better log names (append them?)
	bold_message("Building the Emacs mode for Ciao"),
	% First, generate 'ciao-config.el' from 'ciao-config.el.skel'
	generate_emacs_config,
	% Generate autoloads automatically with 'batch-update-autoloads'
	Dir = ~emacsmode_dir,
	Init = ~fsR(Dir/'ciao-site-file.el'),
	emacs_update_autoloads(Dir, 'emacs_mode3', Init),
	% Compile to elisp bytecode the .el files
        EL = ~ciao_mode_el_files,
        ELC = ~ciao_mode_elc_files,
	emacs_batch_byte_compile(Dir, 'emacs_mode', EL),
	( % (failure-driven loop)
	  member(ELCFile, ~fsRlist_rel(Dir, ELC)),
	    set_file_perms(ELCFile, ~perms),
	    fail
	; true
	),
	bold_message("Emacs mode for Ciao build completed").

get_bindir_for_emacs(EmacsType, EmacsDir) :-
	( EmacsType = 'Win32' ->
	    % TODO: Why?
	    Dir = ~fsR(builddir_bin(build))
	; Dir = ~fsR(installdir_bin(build))
	),
	get_dir_for_emacs(EmacsType, Dir, EmacsDir).

% TODO: use ciao-root-dir for all EmacsType?
get_dir_for_emacs('MacOSBundle', Dir, EmacsDir) :- !,
	flatten(["(concat ciao-root-dir \"", ~atom_codes(Dir), "\")"], EmacsDir).
get_dir_for_emacs(EmacsType, Dir, EmacsDir) :-
	flatten(["\"", ~emacsstylepath(EmacsType, Dir), "\""], EmacsDir).

get_app_for_emacs(EmacsType, App, Bundle, Kind, AppShell) :-
	enum_bundle(_, Bundle), % (Bundle exists)
	get_app_for_emacs_(EmacsType, App, Bundle, Kind, AppShell),
	!.
% Avoid failure if the App is not being installed:
get_app_for_emacs(_, App, _Version, _, AppShell) :-
	flatten(["\"", ~atom_codes(App), "\""], AppShell).

get_app_for_emacs_(EmacsType, App, Bundle, Kind, AppShell) :-
	( Kind = plexe -> K = plexe
	; Kind = script -> K = ext(~script_extension(EmacsType))
	),
	App2 = ~fsR(concat_verk(Bundle, K, App)),
	AppShell = ~flatten(["(concat ciao-bin-dir \"/", ~atom_codes(App2), "\")"]).

% Avoid failure if the bundle is not there
get_bundle_manual_base_for_emacs(Bundle, NameVersion):-
	enum_bundle(_, Bundle), % (Bundle exists)
	bundle_manual_base(Bundle, NameVersion), 
	!.
get_bundle_manual_base_for_emacs(Bundle, Bundle).

% ---------------------------------------------------------------------------
% Generate ciao-config.el from ciao-config.el.skel

:- use_module(library(bundletool/bundle_aux), [load_template/4]).

generate_emacs_config :-
	emacs_type(EmacsType),
	EmacsModeDir = ~emacsmode_dir,
	%
	% TODO: 'EmacsType' probably should not be necessary
	emacs_type_specific(EmacsType, String, Tail3),
	%
	% TODO: this should not be necessary; the emacs-mode should
	% call lpdoc to do things such as generation of SETTINGS.pl
	load_template(EmacsModeDir/'ciao-config.el', [
            'CIAODE_VERSION' = ~bundle_version_patch(ciaode),
	    % Paths
	    'CIAOBINDIR' = ~get_bindir_for_emacs(EmacsType),
	    'CIAOREALLIBDIR' = ~get_dir_for_emacs(EmacsType, ~get_installdir_lib(ciao)),
	    'LPDOCDIR' = ~get_dir_for_emacs(EmacsType, ~docdir),
	    'LPDOCLIBDIR' = ~get_dir_for_emacs(EmacsType, ~get_installdir_lib(lpdoc)),
	    % Manuals
            'CIAO_MANUAL_BASE' = ~get_bundle_manual_base_for_emacs('ciao'),
            'CIAOPP_MANUAL_BASE' = ~get_bundle_manual_base_for_emacs('ciaopp'),
            'LPDOC_MANUAL_BASE' = ~get_bundle_manual_base_for_emacs('lpdoc'),
	    % Binaries
	    'PLINDENT' = ~get_app_for_emacs(EmacsType, 'plindent', ciao, plexe),
	    'CIAOSHELL' = ~get_app_for_emacs(EmacsType, 'ciao', ciao, script),
	    'CIAOPPSHELL' = ~get_app_for_emacs(EmacsType, 'ciaopp', ciaopp, script),
	    'LPDOCEXEC' = ~get_app_for_emacs(EmacsType, 'lpdoc', lpdoc, plexe)
        ], Tail3, ""),
	%
	string_to_file(String, ~fsR(EmacsModeDir/'ciao-config.el')).

% TODO: Use only one template and detect from elisp if we are in MacOSX or Win32

:- use_module(library(llists), [append/2]).

% TODO: use ciao-root-dir for all EmacsType?
emacs_type_specific('MacOSBundle', String, Tail) :- !,
	% TODO: do not use a hardwired value for Root!
	Root = '/usr/lib/ciao',
	% TODO: do not use a hardwired value for Engine!
	Lib = ~fsR(concat_ver(ciao, Root/'ciao')),
	Engine = ~fsR(Lib/'engine'/(~name_ciaoengine_arch)),
	append([
	    ";; Beginning of the specific part to MacOS Application Bundle\n"||
	    "(defvar ciao-root-dir "||
	    "\"/Applications/Emacs.app/Contents/Resources\"\n"||
	    "        \"path of Emacs Application Bundle\")\n\n"||
	    "(setenv \"CIAOENGINE\" (concat ciao-root-dir \"", ~atom_codes(Engine), "\"))\n"||
	    "(setenv \"CIAOLIB\" (concat ciao-root-dir "||
	    "\"", ~atom_codes(Lib), "/\"))\n\n"||
	    ";; End of the specific part to MacOS Application Bundle\n\n", Tail],
	    String).
emacs_type_specific('Win32', String, Tail) :- !,
	emacsstylepath('Win32', ~fsR(builddir_doc(build)), SDirS),
	append([
		";; Specific to Windows installation:\n"||
		";; Location of info manuals\n"||
		"(setq Info-default-directory-list  (cons \n"||
		"      \""|| SDirS, "\" \n"||
		"      Info-default-directory-list))\n"||
%%% Should put this in a separate file in CIAODESRC/emacs-mode and cat it now:
		";; Make things nicer (but check if you are already doing it)\n",
		"(global-font-lock-mode)\n"||
		"(transient-mark-mode t)\n"||
		";; Help for using the Windows command.com as your shell\n"||
		";; (comment out if you use bash, etc.):\n"||
		"(setq process-coding-system-alist \n"||
		"'((\"cmdproxy\" . (raw-text-dos . raw-text-dos))))\n"||
		";; Preventing ctrln-m's from being printed in the shell\n"||
		"(add-hook 'comint-output-filter-functions "||
		"  'shell-strip-ctrl-m nil t)\n"||
		"; -----"||
                "----------------------------------------------------------------\n\n"||
		Tail],
	    String).
emacs_type_specific(_, T, T).

% ---------------------------------------------------------------------------

% (called from 'ciaosetup' (for testing))
'$bundle_own_make_target'(register_emacs_mode) :- bundleitem_register_hook(emacs_mode).
% (called from 'ciaosetup' (for testing))
'$bundle_own_make_target'(unregister_emacs_mode) :- bundleitem_unregister_hook(emacs_mode).

bundleitem_register_hook(emacs_mode) :-
	register_emacs(emacs),
	register_emacs(xemacs).

register_emacs(EmacsKind) :-
	( with_emacs_mode_k(EmacsKind) ->
	    ( emacs_init_file(EmacsKind, InitFile) ->
	        (-register_in_script(InitFile, ";", ~emacs_config))
	    ; % Do not register
	      true
	    )
	; true
	).

bundleitem_unregister_hook(emacs_mode) :-
	unregister_emacs(emacs),
	unregister_emacs(xemacs).

unregister_emacs(EmacsKind) :-
	( with_emacs_mode_k(EmacsKind) ->
	    ( emacs_init_file(EmacsKind, InitFile) ->
	        (-unregister_from_script(InitFile, ";"))
	    ; true
	    )
	; true
	).

emacs_config(S) :-
	Lib = ~ciaolibemacs,
	emacs_type(EmacsType),
	Lib2S = ~emacsstylepath(EmacsType, Lib),
	atom_codes(Lib2, Lib2S),
	emacs_config_(Lib2, S, []).

emacs_config_(Lib) -->
	"(if (file-exists-p \"", emit_atom(Lib), "\")\n"||
	"(load-file \"", emit_atom(Lib), "\")\n"||
	")\n".

% The absolute path for the 'ciao-site-file.el' file
ciaolibemacs(LibEmacs) :-
	( instype(local) ->
	    LibEmacs = ~fsR(~emacsmode_dir/'ciao-site-file.el')
	; % TODO: Place the version in the right place automatically?
	  % TODO: Verify that the rest of .el files are in the correct directory.
	  LibEmacs = ~fsR(installdir_lib(ciao)/'ciao-site-file.el')
	).

% ---------------------------------------------------------------------------

% TODO: (transform into a bundleitem hook)
% TODO: docs always must be done after build, not before!
% Creation of LPdoc documentation files for emacs-mode
prepare_docs_emacs_hook :-
	EmacsModeDir = ~emacsmode_dir,
	emacs_batch_call(EmacsModeDir, 'emacs_mode2', % TODO: right log name?
	  ['--eval', '(setq load-path (cons "." load-path))',
	   '-l', 'ciao-documentation.el',
	   '-f', 'ciao-mode-documentation']),
	%
	set_file_perms(~fsR(EmacsModeDir/'CiaoMode.lpdoc'), ~perms),
	touch(~fsR(EmacsModeDir/'CiaoMode.pl')).

icon_dir := ~fsR(installdir_lib(ciao)/'icons').

%-----------------------------------------------------------------------------

% TODO: (hook)
% (only for instype=global)
bundleitem_install_hook_decl(emacs_mode).
bundleitem_install_hook(emacs_mode) :-
	bold_message("Installing Ciao emacs-mode (Emacs based IDE)"),
        % (only for global installation)
	Dir = ~emacsmode_dir,
	Loader = ~fsR(Dir/'ciao-mode-init.el'),
        string_to_file(~emacs_config, Loader),
	( with_any_emacs_mode ->
	    mkf_emacs_install
	; true
	).

% TODO: (hook)
% (only for instype=global)
bundleitem_uninstall_hook_decl(emacs_mode).
bundleitem_uninstall_hook(emacs_mode) :-
	bold_message("Uninstalling Ciao emacs-mode (Emacs based IDE)"),
	( with_any_emacs_mode ->
	     mkf_emacs_uninstall
	; true
	).

mkf_emacs_install :-
	Bundle = ciao,
	install_item(Bundle, ~mkf_emacs_mode_desc).

mkf_emacs_uninstall :-
	Bundle = ciao,
	uninstall_item(Bundle, ~mkf_emacs_mode_desc).

mkf_emacs_mode_desc := [
	  dir(~fsR(installdir_lib(ciao)), [do_not_del]),
	  dir(~icon_dir, [files_from(~emacsmode_dir/icons), del_rec]),
	  ~ciao_mode_lisp_desc,
	  ~ciao_mode_init_desc(xemacs),
	  ~ciao_mode_init_desc(emacs)
	].

% TODO: Remember to 'update ciao/library/bundletool/win32/CiaoDE.iss.skel'
%       if the files here are modified. Ideally, that file should not
%       contain any hardwired list of files.

% TODO: Merge this part with register. Make emacsinitfile a particular
% case.
ciao_mode_init_desc(EmacsKind) := Desc :-
	with_emacs_mode_k(EmacsKind),
        MidDir = ~emacs_site_start_k(EmacsKind),
        is_site_start_d(MidDir),
	!,
        Mid = MidDir/(~emacsinitfile),
	Desc = [
          dir(MidDir, [do_not_del]),
          lib_file_list(~emacsmode_dir, [
            'ciao-mode-init.el'-[mid(Mid)] % TODO: why?
          ])
        ].
ciao_mode_init_desc(_) := [].

emacsinitfile := ~bundle_config_value(ciao, 'EMACSINITFILE').

% Obtain the appropriate configuration file for this system or
% installation (.emacs or site-start.el). This predicate fails if no
% change is required, because the mode is installed through the
% site-start.d/ directory (Debian only?), or because the Ciao Emacs
% Mode is disabled.
emacs_init_file(EmacsKind) := InitFile :-
        ( % Local installation, register in your .emacs file
	  instype(local), update_dotemacs_k(EmacsKind) ->
	    InitFile = ~dotemacs_k(EmacsKind)
	; % Register in the site-start file (just in case that)
	  % the site-start.d directory was not found; see
	  % `ciao_mode_init_desc/2`)
	  Dir = ~emacs_site_start_k(EmacsKind),
	  \+ is_site_start_d(Dir) ->
	    InitFile = ~fsR(Dir/'site-start.el')
	; % No init file has to be modified
	  fail
	).

% Check that Dir is a site-start.d directory
is_site_start_d(Dir) :-
        atom_concat(_, '/site-start.d', Dir).

:- use_module(library(llists), [append/2]).

ciao_mode_lisp_desc :=
	lib_file_list(~emacsmode_dir, 
          ~append([
            ~addprops(['ciao-site-file.el'], [copy_and_link]),
            ~addprops(~ciao_mode_el_files, [copy_and_link]),
            ~addprops(~ciao_mode_elc_files, [copy_and_link])
          ])
        ).

ciao_mode_lisp_files := [
	'word-help',
	'ciao-help',
	'ciao-faces',
	'ciao-syntax',
	'ciao-parsing',
	'ciao-aux',
	'ciao-font-lock',
	'ciao-vc',
	'ciao-scratchpad',
	'ciao-process',
	'ciao-compile',
	'ciao-loading',
	'ciao-testing',
	'ciao-debugger',
	'ciao-lpdoc',
	'ciao-ciaopp',
	'java-ciaopp',
	'ciao-bundle',
	'ciao-optim-comp',
	'ciao-org',
	'ciao-widgets',
	'ciao-common',
	'ciao-config',
	'ciao-splash',
	'ciao-bindings',
	'ciao'].

ciao_mode_el_files := ~add_suffix(~ciao_mode_lisp_files, '.el').
ciao_mode_elc_files := ~add_suffix(~ciao_mode_lisp_files, '.elc').

add_suffix([],     _Suffix, []).
add_suffix([L|Ls], Suffix,  [R|Rs]) :-
	atom_concat(L, Suffix, R),
	add_suffix(Ls, Suffix, Rs).

% ----------------------------------------------------------------------------

'$bundle_own_make_target'(clean_emacs_mode) :- clean_emacs_mode.
clean_emacs_mode :-
	bold_message("Cleaning up emacs_mode...", []),
	EmacsModeDir = ~emacsmode_dir,
	% TODO: necessary? repeated?
	del_file_nofail(~fsR(EmacsModeDir/'ciao-site-file.el')),
	% clean log files
	clean_log(EmacsModeDir, 'emacs_mode'),
	clean_log(EmacsModeDir, 'emacs_mode2'),
	clean_log(EmacsModeDir, 'emacs_mode3'),
%	del_file_nofail(~fsR(EmacsModeDir/'*~' pattern?)).
	%
	% (automatically generated files)
	del_file_nofail(~fsR(EmacsModeDir/'CiaoMode.lpdoc')),
	del_file_nofail(~fsR(EmacsModeDir/'ciao-config.el')),
	% TODO: Use common implementation, do not specify suffixes by hand
	del_files_nofail(~add_prefix(~glob(~fsR(EmacsModeDir), '*.itf|*.po|*.asr|*.elc'),
	                             ~atom_concat(~fsR(EmacsModeDir), '/'))).

add_prefix([],     _Preffix, []).
add_prefix([L|Ls], Preffix,  [R|Rs]) :-
	atom_concat(Preffix, L, R),
	add_prefix(Ls, Preffix, Rs).

% ---------------------------------------------------------------------------
% (Auxiliary predicates)

% Add the property list Prop to each of Xs (for bundle description)
addprops([], _Props) := [].
addprops([X|Xs], Props) := [X-Props | ~addprops(Xs, Props)].

% Resolve a list of files, relative to Dir
fsRlist_rel(Dir, Xs) := ~fsRlist_rel_(Xs, Dir).

fsRlist_rel_([], _) := [].
fsRlist_rel_([X|Xs], Dir) := [Y| ~fsRlist_rel_(Xs, Dir)] :-
	Y = ~fsR(Dir/X).

% ===========================================================================
% Enumeration of the standalone utilities in etc/ and etc_contrib/
% TODO: Generalize and split

etc_dir := bundle_src(ciao)/etc.
etc_contrib_dir := bundle_src(ciao)/etc_contrib.

bundleitem_def_decl_hook(etc_utilities).
bundleitem_def_hook(etc_utilities) :=
	group("etc/ standalone utilities",
          standalone_list(ciao, ~etc_dir, ~etc_utilities)).

etc_utilities := ~findall(B-[K], etc_utility(B, K)).

etc_utility(B, K) :-
	( plutility(B), K = plexe
	; shscript(B), K = shscript
	).

plutility :=
	'fileinfo'|
	'pldiff'|
	'viewpo'|
	'lpmake'|
	'plindent'|
	'show_asr'|
	'compiler_output'|
	'checkline'.

% TODO: strange... enumerated for installation, add a table of exec+kind instead
shscript :=
	'ciao_get_arch'|
        'ciao'.

bundleitem_def_decl_hook(etc_contrib_utilities).
bundleitem_def_hook(etc_contrib_utilities) :=
	group("etc_contrib/ standalone utilities",
	  standalone_list(ciao, ~etc_contrib_dir, ~etc_contrib_utilities)).

etc_contrib_utilities := ~findall(B-[K], etc_contrib_utility(B, K)).

etc_contrib_utility(B, K) :-
	plutility_contrib(B), K = plexe.

% TODO: distribute those utils?
plutility_contrib :=
	'synch_actions'|
	'cleandirs'.

% ============================================================================

% (hook)
build_etc :-
	Bundle = ciao,
	build_item(Bundle, dot_shell(csh)),
	build_item(Bundle, dot_shell(sh)),
	build_item(Bundle, etc_utilities),
	build_item(Bundle, etc_contrib_utilities),
	build_item(Bundle, ciao_get_arch),
	build_item(Bundle, ciaocl).

:- use_module(library(bundletool/bundle_aux), [wr_template/4]).

% Generate shell initialization files
bundleitem_build_hook_decl(dot_shell(_)).
bundleitem_build_hook(dot_shell(Sh)) :-
	normal_message("Creating ~w", [~dot_shell_file(Sh)]),
	wr_template(origin, ~etc_dir, ~dot_shell_file(Sh), [
	    'CiaoDocDir' = ~docdir,
	    'CiaoBinDir' = ~fsR(installdir_bin(build)),
	    'ConfigOptimizingCompiler' = ~optimizing_compiler_string(Sh)
        ]).

dot_shell_file(csh) := 'DOTcshrc'.
dot_shell_file(sh) := 'DOTprofile'.

:- use_module(library(llists), [flatten/2]).

% TODO: Merge
optimizing_compiler_string(Shell) :=
	~optimizing_compiler_string_(Shell, ~optimizing_compiler).
optimizing_compiler_string_(_, no) := "" :- !.
optimizing_compiler_string_(csh, yes) :=
	~flatten(["eval `", ~atom_codes(~ciao_lib_dir), "/optim_comp/ciaotool csh-env`"]).
optimizing_compiler_string_(sh, yes) :=
	~flatten(["eval $(", ~atom_codes(~ciao_lib_dir), "/optim_comp/ciaotool bash-env)"]).

install_prolog_name := ~bundle_config_value(ciao, 'INSTALL_PROLOG_NAME').

:- use_module(library(format), [sformat/3]).

% TODO: 'ciao' script is needed to set the value of ciaosetup
%   configurable prolog_flags. This is a very ugly way of doing it.
%   Instead, add a parameter to 'ciaosh' to load/ignore value of
%   Prolog flags from saved configuration.

% Generate 'ciao' startup script
bundleitem_build_hook_decl(ciaocl).
bundleitem_build_hook(ciaocl) :-
	normal_message("Building '~w' (command line)", ['ciao']),
	wr_template(kver(shscript, ciao), ~etc_dir, 'ciao', [
	    'ExtraCommands' = ~ciao_extra_commands,
	    'CiaoVersion' = ~bundle_version(ciao),
	    'CiaoSuffix' = ~get_ciao_ext,
	    'CiaoBinDir' = ~fsR(installdir_bin(build))
        ]),
	b_link(shscript, ciao, 'ciao'),
 	( install_prolog_name(yes) ->
 	    b_link_as(shscript, ciao, 'ciao', 'prolog')
 	; true
 	).

ciao_extra_commands(ExtraCommands) :-
	sformat(ExtraCommands, "-e '~w'",
	    [~list_to_lits(~set_prolog_flag_from_bundle_config)]).

% TODO: import from formulae
list_to_lits([],     true).
list_to_lits([X|Xs], Lits) :-
	list_to_lits2(Xs, X, Lits).

list_to_lits2([],     X,  X).
list_to_lits2([X|Xs], X0, (X0, Lits)) :-
	list_to_lits2(Xs, X, Lits).

bundleitem_build_hook_decl(ciao_get_arch).
bundleitem_build_hook(ciao_get_arch) :-
	normal_message("Building '~w' (command line)", ['ciao_get_arch']),
	% TODO: we build nothing here (just copy) but the user does not want to know
	b_copy(shscript, ciao, ~etc_dir, 'ciao_get_arch'),
	b_link(shscript, ciao, 'ciao_get_arch').

% ============================================================================

bundleitem_def_decl_hook(dot_shell_).
bundleitem_def_hook(dot_shell_) :=
  small_group("shell initialization scripts",
    lib_file_list(~etc_dir, [
      'DOTprofile'-[copy_and_link],
      'DOTcshrc'-[copy_and_link]
    ])).

bundleitem_def_decl_hook(ciaocl_).
bundleitem_def_hook(ciaocl_) := R :-
	( install_prolog_name(yes) ->
	    Opts = [link_as('prolog')]
	; Opts = []
	),
	R = small_group("'ciao' (command line)", 
	      bin_copy_and_link(shscript, ciao, 'ciao', Opts)).

% (only for instype=global)
bundleitem_install_hook_decl(etc).
bundleitem_install_hook(etc) :-
	Bundle = ciao,
	install_item(Bundle, ~etc_desc).

% (only for instype=global)
bundleitem_uninstall_hook_decl(etc).
bundleitem_uninstall_hook(etc) :-
	Bundle = ciao,
	uninstall_item(Bundle, ~etc_desc).

etc_desc := [
	dir(~fsR(installdir_bin(build)), [do_not_del]),
	dot_shell_,
	ciaocl_,
	etc_utilities,
	etc_contrib_utilities
        ].

% TODO: 'NewUser' (like the rest of README files) is not
%       installed together with the documentation. Is it right?
%%	b_uninstall_file(~lib_dir/'NewUser'),





