:- module(abs_solver_rt,
	[
	    'abs'/2,
	    '.<-.'/2,
	    '.->.'/2,
	    'abs_atom'/1,
	    'abs_is'/2,
	    '.=.'/2,
	    '.<.'/2,
	    '.>.'/2,
	    '.=:=.'/2,
	    '.@<.'/2,
	    '.@>.'/2,
	    '.==.'/2,
	    'join'/3,
	    'meet'/3,
	    put_abs_list/2,
	    get_abs_list/2,
	    check_if_bot/1,
	    make_bot/1,
	    join_info_att/2,
	    entailment/2,
	    entailment_list/2
	]).

:- use_package(attr).
:- use_module(engine(attributes)).

:- use_module(library(aggregates)).
:- use_module(library(format)).
:- use_module(library(terms_vars)).

:- use_module(library(abs_solver/abstraction)).
:- include(library(abs_solver/abs_solver_syntax)).

%% OPERATORS $$
V1 .<-. AbsV1 :-
	abs(V1, AbsV1).

V1 .->. AbsV1 :-
	abs(V1, AbsV1).

abs_atom(V1) :-
	abs(Temp, atm),
	(
	    meet(V1, Temp, Temp) ->
	    abs(V1, atm)
	;
	    abs(V1, bot)
	).

V1 abs_is E :-
	nonvar(E),
	E =.. [Op, S1, S2],
 	abs(Num, num),
 	meet(S1, Num, N1),
 	meet(S2, Num, N2),
	setof( [P1, P2, F1],
	(
	    meet(N1, P1, P1),
	    meet(N2, P2, P2),
	    V1 .=. F1,
	    operation_att(Op, P1, P2, F1)
	), Set),
	join_info_att(Set, JoinSet),
	copy_list(JoinSet, [S1, S2, V1]).

operation_att(Op, S1, S2, V1) :-
	abs(S1, AbsS1),
	abs(S2, AbsS2),
	abs(V1, AbsV1),
	abs_operation(Op, AbsS1, AbsS2, AbsV1),
	AbsV1 \= bot.	

% ?- abs(X,top), abs(Y,pos+z), abs(Z,pos), X abs_is Y + Z.
% abs(X,pos),
% abs(Y,pos+z),
% abs(Z,pos) ? .
% ?- abs(X,var), abs(Y,pos+z), abs(Z,pos), X abs_is Y + Z.
% abs(X,pos),
% abs(Y,pos+z),
% abs(Z,pos) ? .
% ?- abs(X,neg), abs(Y,pos+z), abs(Z,pos), X abs_is Y + Z.
% no
% ?- abs(X,neg), abs(Y,pos+z), abs(Z,top), X abs_is Y + Z.
% abs(X,neg),
% abs(Y,zero),
% abs(Z,neg) ? .
% ?- abs(X,top), abs(Y,top), abs(Z,top), X abs_is Y + Z.
% abs(X,num),
% abs(Y,num),
% abs(Z,num) ? .

V1 .=. V2 :-
	abs(V1, AbsV1),
	abs(V2, AbsV2),
	abs_condition(==, AbsV1, AbsV2, [NewAbs, _]),
	(
	    AbsV1 == var ->
	    FinalAbs = AbsV2
	;
	    (
		AbsV2 == var ->
		FinalAbs = AbsV1
	    ;
		FinalAbs = NewAbs
	    )
	),
	FinalAbs \= bot,
	abs(V1, FinalAbs),
	abs(V2, FinalAbs).

V1 .<. V2 :-	condition_arith_att(V1, @<, V2) .
V1 .>. V2 :-	condition_arith_att(V1, @>, V2) .
V1 .=:=. V2 :-	condition_arith_att(V1, ==, V2) .

condition_arith_att(V1, Cond, V2) :-
	abs(Temp, num),
	meet(V1, Temp, V1), abs(V1, AbsV1),
	meet(V2, Temp, V2), abs(V2, AbsV2),
	abs_condition(Cond, AbsV1, AbsV2, [NewAbsV1, NewAbsV2]),
	NewAbsV1 \= bot,
	abs(V1, NewAbsV1),
	abs(V2, NewAbsV2).


%% var @< num @< atm @< str %%
V1 .@<. V2 :-	condition_att(V1, @<, V2) .
V1 .@>. V2 :-	condition_att(V1, @>, V2) .
V1 .==. V2 :-	condition_att(V1, ==, V2) .

condition_att(V1, Cond, V2) :-
	abs(V1, AbsV1),
	abs(V2, AbsV2),
	abs_condition(Cond, AbsV1, AbsV2, [NewAbsV1, NewAbsV2]),
	NewAbsV1 \= bot,
	abs(V1, NewAbsV1),
	abs(V2, NewAbsV2).

%% join /\ lub
join(A,B,S) :-
	abs(A, AbsA),
	abs(B, AbsB),
	abs(S, AbsS),
	abs_join(AbsA, AbsB, AbsS),
	AbsS \= bot.

%% meet \/ glb
meet(A,B,S) :-
	abs(A, AbsA),
	abs(B, AbsB),
	abs(S, AbsS),
	abs_meet(AbsA, AbsB, AbsS),
	AbsS \= bot.

% Check or give the abstract value of a ground-value
% Give the current abstraction of X or introduce a variable as abstract value
% if exist current abs value check if is equal to the new one.
% if not introduce the new abstract value
abs(X, AbsX) :-
	ground(X), !,
	abstract_constant(X, AbsCurrent),
	% take the Current value, or check that the abstract is ok
	AbsX = AbsCurrent.
abs(X,AbsX) :-
	var(AbsX),
	(
	    attvar(X) ->
	    get_attr_local(X, AbsX)
	;
	    put_attr_local(X, AbsX)
	).
abs(X,AbsX) :-
	ground(AbsX),
	in_abs_domain(AbsX, _),
	(
	    attvar(X) ->
	    % What to do ?
	    put_attr_local(X, AbsX)
%	    get_attr_local(X, AbsCurrent),
%	    AbsX = AbsCurrent
	;
	    put_attr_local(X, AbsX)
	).

%% Entailment X |- Y
entailment(X, Y) :-
	meet(X, Y, X).

entailment_list([], []).
entailment_list([X | RestX], [Y | RestY]) :-
	entailment(X, Y),
	entailment_list(RestX, RestY).

:- multifile attr_portray_hook/2.
:- multifile attr_unify_hook/2.
:- multifile attribute_goals/3.

attr_portray_hook(Att,Var) :-
	display(Var), display(', '), display(Att).

attr_unify_hook(_Att, _X) :-
	true.

attribute_goals(X, [S | T], T) :-
	get_attr_local(X, Att),
 	S = (abs(X,Att)).

% Which is better ?
% X abs_is top ? .
% X.=.top ? .
% abs(X,top) ? .
% X<-top ? .


join_info_att([A], Copy) :-
	copy_list(A, Copy).
join_info_att([A, B | Rest], Rs) :-
	join_info_list(A, B, Join),
	join_info_att([Join | Rest], Rs).
join_info_list([], [], []).
join_info_list([A | As], [B | Bs], [Join | Js]) :-
	join(A, B, Join),
	join_info_list(As, Bs, Js).
copy_list([],[]).
copy_list([A | Rest], [Copy | RestCopy]) :-
	abs(A, AbsA),
	abs(Copy,AbsA),
	copy_list(Rest, RestCopy).

put_abs_list([],[]).
put_abs_list( [V | Vs], [abs(V, AbsV) | AbsList]) :-
	var(V),
	(
	    ground(AbsV) ->
	    put_attr_local(V, AbsV)
	;
	    put_attr_local(V, var)
	),
	put_abs_list(Vs, AbsList).

get_abs_list([],[]).
get_abs_list([V | Vs], [abs(V, AbsV) | AbsList]) :-
	(
	    get_attr_local(V, AbsV) ->
	    detach_attribute(V)
	;
	    AbsV = var
	),
	get_abs_list(Vs, AbsList).

check_if_bot(Body) :-
	varset(Body, Xs),
	member(X,Xs),
	get_attr_local(X, bot).
make_bot([]).
make_bot([V |Rest]) :-
	put_attr_local(V, bot),
	make_bot(Rest).


%% Try to solve this
% ?- X .<. Y, abs(X, pos).
% abs(X,pos),
% abs(Y,num) ? .
% ?- abs(X,pos), X .<. Y.
% abs(X,pos),
% abs(Y,pos) ? .


% V1 abs_is E :-
% 	interp_arith_att(E, AbsE),
% 	abs(V1, AbsV1),
% 	(
% 	    AbsV1 \= var ->
% 	    abs_meet(AbsV1, AbsE, Abs_join)
% 	;
% 	    Abs_join = AbsE
% 	),
% 	(
% 	    abs_meet(Abs_join, AbsE, AbsR), AbsR \= AbsE ->
% 	    update_expresion_att(AbsV1, E)
% 	;
% 	    true
% 	),
% 	abs(V1, Abs_join).
% interp_arith_att(E, AbsE) :-
% 	E =.. [Op, S1, S2],
% 	abs(S1, AbsS1),
% 	abs(S2, AbsS2),
% 	abs_operation(Op, AbsS1, AbsS2, AbsE).
% interp_arith_att(Constant, AbsE) :-
% 	Constant =.. [_],
% 	abs(Constant, AbsE),
% 	in_abs_domain(AbsE,_).

% update_expresion_att(_AbsV1, Constant) :-
% 	Constant =.. [_].
% update_expresion_att(AbsV1, E) :-
% 	E =.. [Operator, S1, S2],
% 	abs(S1, AbsS1),
% 	abs(S2, AbsS2),
% 	findall(X, abs_operation(Operator, X, AbsS2, AbsV1), Xs),
% 	(
% 	    apply_glb(AbsS1, Xs, Ys),  join_list(Ys, Res), Res \= bot ->
% 	    abs(S1, Res)
% 	;
% 	    true
% 	).

% apply_glb(_, [], []).
% apply_glb(AbsV1, [Abs1 | Xs], [Res | Ys]) :-
% 	abs_meet(AbsV1, Abs1, Res),
% 	apply_glb(AbsV1, Xs, Ys).
% join_list([X], X).
% join_list([Abs1, Abs2 | As], Rs):-
% 	abs_join(Abs1, Abs2, Res),
% 	join_list([Res | As], Rs).