#!/bin/bash

detect_mathkernel_on_darwin() {
     local MATHEMATICA_APP_PATH="$(mdfind 'kMDItemCFBundleIdentifier == "com.wolfram.Mathematica"' | head -n1)"
 
     if [ -n "$MATHEMATICA_APP_PATH" ]; then 
	 MATHKERNEL="${MATHEMATICA_APP_PATH}/Contents/MacOS/MathKernel"
	 if [ ! -f "${MATHKERNEL}" ]; then 
	     MATHKERNEL=
	 fi
     fi

}

config_darwin() {
    local APPNAME

    if [ x"${MATHKERNEL}" == x"" ]; then
	echo "Mathematica not detected" >&2
	exit 1
    fi

    APPNAME="$(dirname "$(dirname "$(dirname "${MATHKERNEL}")")")"
    if [ ! -d "${APPNAME}" ] ; then
	echo "Mathematica not detected" >&2
	exit 1
    fi

    INCLUDEDIR="${APPNAME}/SystemFiles/Links/MathLink/DeveloperKit/MacOSX-x86-64/CompilerAdditions"
    LIBS="${APPNAME}/SystemFiles/Links/MathLink/DeveloperKit/MacOSX-x86-64/CompilerAdditions/lib*.a"
}

config_unix() {
    echo  config_unix not implemented >&2
    exit 1
}


help() {
    echo "Usage: mathematica-config [OPTION]
Options:
  -h, --help        prints this help text to stdout
  --mathlink        prints the command to start mathlink
  --cppflags        prints preprocessor flags
  --ldflags         prints linker flags
 "
}


MATHKERNEL="$(which MathKernel)"

if [ "$(uname)" == "Darwin" ]; then
    if [ x"${MATHKERNEL}" == x"" ]; then
	detect_mathkernel_on_darwin
	config_darwin
    fi
else
    config_unix
fi

case ${1} in
    "--mathlink") 
	echo \"${MATHKERNEL}\" -mathlink
	;;
    "--cppflags") 
	echo -I\"${INCLUDEDIR}\"
	;;
    "--ldflags")
	echo \"${LIBS}\" -lstdc++ -framework Foundation
	;;
    "-h" | "--help")
	help 
	;;
    *)
	help 
	exit 1
	;;
esac

 
