/*
 *  bignum.h
 *
 *  Bignum arithmetics (support for arbitrary size integers)
 *
 *  Copyright (C) 1996,1997,1998, 1999, 2000, 2001, 2002 UPM-CLIP
 *
 *  Authors:
 *    Torbjorn Granlund, Johan Andersson, and Mats Carlsson
 *    Ref: Knuth vol. 2 sec. 4.3.1
 */
 
#ifndef _CIAO_BIGNUM_H
#define _CIAO_BIGNUM_H

#include "threads.h"
#include "datadefs.h"
#include "support_macros.h"

bool_t bn_positive(register Bignum *x);
int bn_add(register Bignum *x, register Bignum *y, register Bignum *z, Bignum *zmax);
int bn_incr(register Bignum *x, Bignum *ignore, register Bignum *z, Bignum *zmax);
int bn_plus(register Bignum *x, Bignum *ignore, register Bignum *z, Bignum *zmax);
int bn_subtract(register Bignum *x, register Bignum *y, register Bignum *z, Bignum *zmax);
int bn_decr(register Bignum *x, Bignum *ignore, register Bignum *z, Bignum *zmax);
int bn_minus(register Bignum *x, Bignum *ignore, register Bignum *z, Bignum *zmax);
int bn_and(register Bignum *x, register Bignum *y, register Bignum *z, Bignum *zmax);
int bn_or(register Bignum *x, register Bignum *y, register Bignum *z, Bignum *zmax);
int bn_xor(register Bignum *x, register Bignum *y, register Bignum *z, Bignum *zmax);
int bn_not(register Bignum *x, register Bignum *y, register Bignum *z, Bignum *zmax);
int bn_lshift(register Bignum *x, Bignum *dist, register Bignum *z, Bignum *zmax);
int bn_rshift(register Bignum *x, Bignum *dist, register Bignum *z, Bignum *zmax);
int bn_compare(register Bignum *x, register Bignum *y);
int bn_multiply(Bignum *x, Bignum *y, Bignum *z, Bignum *zmax);
int bn_quotient_remainder_quot_wanted(Bignum *x, Bignum *y, Bignum *z, Bignum *zmax);
int bn_quotient_remainder_quot_not_wanted(Bignum *x, Bignum *y, Bignum *z, Bignum *zmax);
int bn_from_float(register Bignum *x, Bignum *ignore, register Bignum *z, Bignum *zmax);
int bn_from_string(register char *x, register Bignum *z, Bignum *zmax, int base);
CVOID__PROTO_N(bn_to_string, register Bignum *x, int base);

/* To be used by the test routines in ciao_prolog.c */

intmach_t bn_length(CIAO_REGISTER Bignum *x);

#endif /* _CIAO_BIGNUM_H */
