#!/bin/bash

# Test for emugen

# TODO: integrate with other tests (and optim_comp)

set -e

# TODO: hardwired path!
BUILD_OBJDIR="../../build/objs/DARWINi86"
FILES="wamloop.c instrdefs.h absmachdef.h Makefile-absmach_def"
OFILES="ciaoengine"

function compare_() { # (env DIFFOPT)
    echo "(Autogenerated C)"
    for i in ${FILES}; do
	echo -n "$i: "
	diff ${DIFFOPT} ${BUILD_OBJDIR}/$i $i-old && echo "[OK]" || echo "[?]"
    done
    echo "(Disassembled native code)"
    for i in ${OFILES}; do
	echo -n "$i: "
	diff ${DIFFOPT} $i.s $i.s-old && echo "[OK]" || echo "[?]"
    done
}

function briefcompare() {
    DIFFOPT="-q" compare_
}

function compare() {
    DIFFOPT= compare_
}

function save() {
    for i in ${FILES}; do
	cp ${BUILD_OBJDIR}/$i $i-old
    done
    for i in ${OFILES}; do
	cp $i.s $i.s-old
    done
}

function clean() {
    ../../ciaosetup clean engine
}

function gen() {
    ../../ciaosetup build engine
    # TODO: otool only works in macosx
    for i in ${OFILES}; do
	otool -tv ${BUILD_OBJDIR}/$i > $i.s
    done
}

function promote_emugen() {
    echo "Are you sure? (this is a dangerous operation) (press C-c to abort)"
    read
    for i in ${FILES}; do
	echo "Coping $i"
	cp ${BUILD_OBJDIR}/$i "../../ciao/bootstrap/$i"
    done
}

function help() {
    cat <<EOF
Usage: emugen-test.sh ACTION

  Diff based regression testing of emulator generated C files and
  disassembled native code.

Actions:

  save          Save the current engine files (for regression)
  compare       Compare current output with saved files
  briefcompare  

  gen           Build the engine. Shortcut for 'ciaosetup build engine'
  clean         Clean the engine. Shortcut for 'ciaosetup clean engine'

  promote-emugen  Promote autogenerated files for bootstrap
EOF
}

case $1 in
    gen) gen && briefcompare ;;
    clean) clean ;;
    save) save ;;
    briefcompare) briefcompare ;;
    compare) compare ;;
    promote-emugen) promote_emugen ;;
    *) help ;;
esac
    
    
    


