/*
 *  qget.c
 *
 *  Token reader for the quickload bytecode format.
 *
 *  Copyright (C) 1996,1997,1998, 1999, 2000, 2001, 2002 UPM-CLIP
 */

#include <stdlib.h>

#include "threads.h"
#include "datadefs.h"
#include "support_macros.h"
#include "alloc.h"
#include "objareas.h"

/* declarations for global functions accessed here */

#include "qget.h"
#include "qread.h"
#include "bignum.h"
#include "start.h"

/* local declarations */

/* Shared? Might be; then, only one thread may read in ql's at a time.
   Since it is not a very common situation, we might as well lock ql
   reading.  In any case, access to program area should be locked, and so we
   do not loose anything. */

#define WORKSTRINGLEN (STATICMAXATOM)

/* Called just once at the beginning of the execution */
/* static int workstringlen;*/

static char workstring[WORKSTRINGLEN]; 

/*
void init_workstring(void)
{
  workstringlen = STATICMAXATOM;    
  workstring    = malloc(workstringlen+1); 
} 
*/  

/* Increase the size of the (temporary) string which holds the .po code */
/*
static void expand_workstring(void)
{
  workstringlen *= 2;
  workstring     = realloc(workstring, workstringlen+1);
}
*/

int getshort(FILE *f)
{
  CIAO_REGISTER char *ws;
#if defined(DEBUG)
  CIAO_REGISTER int used_length = 0;
#endif

  for(ws = workstring; (*ws++ = GETC(f)); )
#if defined(DEBUG)
    if (++used_length > WORKSTRINGLEN)
      SERIOUS_FAULT("workstring length exceeded in getshort()")
#else
    ;
#endif
  return atoi(workstring);
}

intmach_t getlong(FILE *f)
{
  CIAO_REGISTER char *ws = workstring;

  while ((*ws++ = GETC(f)))
#if defined(DEBUG)
    if ((ws - workstring) > WORKSTRINGLEN)
      SERIOUS_FAULT("workstring length exceeded in getlong()")
#endif
    ;
    
  return atol(workstring);
}

CFUN__PROTO_N(getlarge, tagged_t, FILE *f)
{
  CIAO_REGISTER int used_length = 0;
  CIAO_REGISTER char *ws = Atom_Buffer;


  while ((ws[used_length++] = GETC(f)))
    if (used_length == Atom_Buffer_Length){
      EXPAND_ATOM_BUFFER(Atom_Buffer_Length*2);
      ws = Atom_Buffer; /* New buffer */
    }

  if (bn_from_string(ws,w->global_top,Heap_End,GetSmall(current_radix))) {
      SERIOUS_FAULT("$qload: miscalculated heap usage");
  } else {
      tagged_t *h = w->global_top;
      int ar = LargeArity(h[0]);
      
      if (ar==2 && IntIsSmall((int)h[1]))
	return MakeSmall(h[1]);
      else {
	  w->global_top += ar+1;
	  h[ar] = h[0];
	  return Tag(STR,h);
	}
    }
}


flt64_t getdouble(FILE *f)
{
  CIAO_REGISTER char *ws = workstring;

  while((*ws++ = GETC(f)))
#if defined(DEBUG)
    if ((ws - workstring) > WORKSTRINGLEN)
      SERIOUS_FAULT("workstring length exceeded in getdouble()")
#endif
    ;
  return atof(workstring);
}

CFUN__PROTO_N(getstring, char *, FILE *f)
{
  CIAO_REGISTER int used_length = 0;
  CIAO_REGISTER char *ws = Atom_Buffer;   /* Try to avoid indirection through WAM */

  while ((ws[used_length++] = GETC(f)))
    if (used_length == Atom_Buffer_Length){
      EXPAND_ATOM_BUFFER(Atom_Buffer_Length*2);
      ws = Atom_Buffer;
    }
  return ws;
}

/* Enable bytecode rewriting on load (experimental).
 *
 * This enables changes in size of internal data structures from same
 * bytecode.
 */
/* TODO: incomplete, fix for 64-bit support */
//#define USE_REWRITE_BYTECODE 1
//#define TRACE_REWRITE 1

#if defined(USE_REWRITE_BYTECODE)
#define USE_ABSMACHDEF_INFO 1
#endif

#if defined(USE_ABSMACHDEF_INFO)
#include "absmachdef.h"
#endif

#if defined(USE_REWRITE_BYTECODE)
CFUN__PROTO_N(rewrite_instr, bcp_t, bcp_t p, bcp_t begin);

CVOID__PROTO_N(bytecode_rewrite, bcp_t begin, bcp_t end) {
  bcp_t p;

  p = begin;
  p = BCoff(p, FTYPE_size(f_i));

#if defined(TRACE_REWRITE)
  fprintf(stderr, "  bytecode - 0x%lx bytes - start at 0x%08lx, det at 0x%08lx\n",
	  (long)((char *)end - (char *)p),
	  (long)FTYPE_size(f_i),
	  (long)BCOp(p, FTYPE_ctype(f_i), 0));
#endif
  while (p != NULL && p < end) {
    p = CFUN__EVAL_N(rewrite_instr, p, begin);
  }
}

void bytecode_rewrite__z(intmach_t j) {
  if (j&1) {
#if defined(TRACE_REWRITE)
    fprintf(stderr, "(unsafe)%ld", (long)Y_OffsetToIndex(j+1));
#endif
  } else {
#if defined(TRACE_REWRITE)
    fprintf(stderr, "%ld", (long)Y_OffsetToIndex(j));
#endif
  }
}

/* todo[ts]: make optional (only when alignment is required) */
void check_align(bcp_t p, intmach_t size) {
  /* TODO: move to some code verifier... this indicates a compiler bug! */
  if (((intp_t)p) % size != 0) { 
    fprintf(stderr, "Panic: corrupted bytecode (unaligned field)\n");
    exit(1);
  }
}

#define FTYPE_name(ID) ftype_name[(intmach_t)(ID)]
char *ftype_name[] = {
  [(intmach_t)FTYPE_id(f_o)]="o",
  [(intmach_t)FTYPE_id(f_e)]="e",
  [(intmach_t)FTYPE_id(f_f)]="f",
  [(intmach_t)FTYPE_id(f_i)]="i",
  [(intmach_t)FTYPE_id(f_l)]="l",
  [(intmach_t)FTYPE_id(f_g)]="g",
  [(intmach_t)FTYPE_id(f_p)]="p",
  [(intmach_t)FTYPE_id(f_t)]="t",
  [(intmach_t)FTYPE_id(f_x)]="x",
  [(intmach_t)FTYPE_id(f_y)]="y",
  [(intmach_t)FTYPE_id(f_z)]="z",
  [(intmach_t)FTYPE_id(f_C)]="C",
  [(intmach_t)FTYPE_id(f_E)]="E",
  [(intmach_t)FTYPE_id(f_Q)]="Q",
  [(intmach_t)FTYPE_id(f_Y)]="Y",
  [(intmach_t)FTYPE_id(f_Z)]="Z",
  [(intmach_t)FTYPE_id(f_b)]="b"
};

#define FuncName(Func) (SetArity((Func)->printname, 0))
#define FuncArity(Func) (Arity((Func)->printname))

CVOID__PROTO_N(display_term, tagged_t term, stream_node_t *stream, bool_t quoted);

CFUN__PROTO_N(rewrite_instr, bcp_t, bcp_t p, bcp_t begin) {
  const char *name;
  const ftype_typeid_t *format;
  definition_t *d;
  tagged_t t;
  intmach_t i, j, k;
  intmach_t arity;
  FTYPE_ctype(f_o) opcode;

#if defined(TRACE_REWRITE)
  fprintf(stderr, "  0x%08lx: ", (long)((char *)p - (char *)begin));
#endif
  check_align(p, FTYPE_size(f_o));
  opcode = BCOp(p, FTYPE_ctype(f_o), 0);

  p = BCoff(p, FTYPE_size(f_o));
  if (opcode >= INS_OPCOUNT) {
    fprintf(stderr, "Panic: corrupted bytecode (invalid opcode 0x%lx)\n", (long)opcode);
    CFUN__PROCEED(NULL);
  }
  
  name = ins_name[opcode];
  format = FTYPE_str__args(abscurr.ins_info, opcode);
  arity = FTYPE_str__arity(abscurr.ins_info, opcode);
  
#if defined(TRACE_REWRITE)
  fprintf(stderr, "%s", name);
#endif
  
  for (i = 0; i < arity; i++) {
#if defined(TRACE_REWRITE)
    fprintf(stderr, " %s(", FTYPE_name(format[i]));
#endif
    switch (format[i]) {
    case FTYPE_id(f_e):
      check_align(p, FTYPE_size(f_e));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)BCOp(p, FTYPE_ctype(f_e), 0));
#endif
      p = BCoff(p, FTYPE_size(f_e));
      break;
    case FTYPE_id(f_f):
      check_align(p, FTYPE_size(f_f));
      t = BCOp(p, FTYPE_ctype(f_f), 0);
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%s/%ld", GetString(t), (long)Arity(t));
#endif
      p = BCoff(p, FTYPE_size(f_f));
      break;
    case FTYPE_id(f_i):
      check_align(p, FTYPE_size(f_i));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)BCOp(p, FTYPE_ctype(f_i), 0));
#endif
      p = BCoff(p, FTYPE_size(f_i));
      break;
    case FTYPE_id(f_l):
      check_align(p, FTYPE_size(f_l));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)(intmach_t)BCOp(p, FTYPE_ctype(f_l), 0));
#endif
      p = BCoff(p, FTYPE_size(f_l));
      break;
    case FTYPE_id(f_g):
      /* TODO: expand in two sub-formats! use compound definitions! */
#if defined(TRACE_REWRITE)
      fprintf(stderr, "h:");
#endif
      check_align(p, FTYPE_size(f_l)); 
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)BCOp(p, FTYPE_ctype(f_l), 0));
#endif
      p = BCoff(p, FTYPE_size(f_l));
#if defined(TRACE_REWRITE)
      fprintf(stderr, ",a:");
#endif
      check_align(p, FTYPE_size(f_i));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)BCOp(p, FTYPE_ctype(f_i), 0));
#endif
      p = BCoff(p, FTYPE_size(f_i));
      break;
    case FTYPE_id(f_p):
      check_align(p, FTYPE_size(f_p));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "0x%lx", (long)(BCOp(p, FTYPE_ctype(f_p), 0) - (char *)begin));
#endif
      p = BCoff(p, FTYPE_size(f_p));
      break;
    case FTYPE_id(f_t):
      check_align(p, FTYPE_size(f_t));
      t = BCOp(p, FTYPE_ctype(f_t), 0);
#if defined(TRACE_REWRITE)
      //CVOID__CALL_N(display_term, t, Output_Stream_Ptr, TRUE);
      /* (not relocated yet) */
      fprintf(stderr, "%x", t);
#endif
      p = BCoff(p, FTYPE_size(f_t));
      break;
    case FTYPE_id(f_x):
      check_align(p, FTYPE_size(f_x));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)X_OffsetToIndex(BCOp(p, FTYPE_ctype(f_x), 0)));
#endif
      p = BCoff(p, FTYPE_size(f_x));
      break;
    case FTYPE_id(f_y):
      check_align(p, FTYPE_size(f_y));
#if defined(TRACE_REWRITE)
      fprintf(stderr, "%ld", (long)Y_OffsetToIndex(BCOp(p, FTYPE_ctype(f_y), 0)));
#endif
      p = BCoff(p, FTYPE_size(f_y));
      break;
    case FTYPE_id(f_z):
      check_align(p, FTYPE_size(f_z));
      bytecode_rewrite__z(BCOp(p, FTYPE_ctype(f_z), 0));
      p = BCoff(p, FTYPE_size(f_z));
      break;
    case FTYPE_id(f_C):
      check_align(p, FTYPE_size(f_C));
#if defined(TRACE_REWRITE)
      {
	char *ptr;
	ptr = BCOp(p, FTYPE_ctype(f_C), 0);
	fprintf(stderr, "%p", ptr);
      }
#endif
      p = BCoff(p, FTYPE_size(f_C));
      break;
    case FTYPE_id(f_E):
      check_align(p, FTYPE_size(f_E));
      d = BCOp(p, FTYPE_ctype(f_E), 0);
#if defined(TRACE_REWRITE)
      //fprintf(stderr, "%s/%ld", GetString(FuncName(d)), (long)FuncArity(d));
      /* (not relocated yet) */
      fprintf(stderr, "%p", d);
#endif
      p = BCoff(p, FTYPE_size(f_E));
      break;
    case FTYPE_id(f_Q):
      check_align(p, FTYPE_size(f_Q));
      p = BCoff(p, FTYPE_size(f_Q)); break;
    case FTYPE_id(f_Y):
      check_align(p, FTYPE_size(f_i));
      k = BCOp(p, FTYPE_ctype(f_i), 0); p = BCoff(p, FTYPE_size(f_i));
      for (j = 0; j < k; j++) {
#if defined(TRACE_REWRITE)
	if (j > 0) fprintf(stderr, " ");
#endif
	check_align(p, FTYPE_size(f_y));
#if defined(TRACE_REWRITE)
	fprintf(stderr, "%ld", (long)Y_OffsetToIndex(BCOp(p, FTYPE_ctype(f_y), 0)));
#endif
	p = BCoff(p, FTYPE_size(f_y));
      }
      break;
    case FTYPE_id(f_Z):
      check_align(p, FTYPE_size(f_i));
      k = BCOp(p, FTYPE_ctype(f_i), 0); p = BCoff(p, FTYPE_size(f_i));
      for (j = 0; j < k; j++) {
#if defined(TRACE_REWRITE)
	if (j > 0) fprintf(stderr, " ");
#endif
	check_align(p, FTYPE_size(f_z));
#if defined(TRACE_REWRITE)
	bytecode_rewrite__z(BCOp(p, FTYPE_ctype(f_z), 0));
#endif
	p = BCoff(p, FTYPE_size(f_z));
      }
      break;
    case FTYPE_id(f_b):
      /* todo[ts]: change if more blob types are added */
      check_align(p, FTYPE_size(f_i));
      t = MakeBlob((tagged_t *)p);
#if defined(TRACE_REWRITE)
      CVOID__CALL_N(display_term, t, Output_Stream_Ptr, TRUE);
#endif
      p = BCoff(p, BlobFunctorSizeAligned(*(tagged_t *)p)+sizeof(functor_t));
      break;
    default:
#if defined(TRACE_REWRITE)
      fprintf(stderr, "?");
#endif
      goto end;
    }
#if defined(TRACE_REWRITE)
    fprintf(stderr, ")");
#endif
  }
 end:
#if defined(TRACE_REWRITE)
  fprintf(stderr, "\n");
#endif
  CFUN__PROCEED(p);
}
#endif

#define USE_STR2NUM_IN_GET_BYTECODE 1

CVOID__PROTO_N(getbytecode,
	       FILE *f,
	       bcp_t insn_p,
	       int length)
{
  CIAO_REGISTER char c;
#if defined(USE_REWRITE_BYTECODE)
  bcp_t begin = insn_p;
#endif

  while ((c=GETC(f))) {
    switch (c) {
    case 'G': {
#if defined(USE_STR2NUM_IN_GET_BYTECODE)
      CBOOL__PROTO_N(string_to_number, unsigned char *AtBuf, int base, tagged_t *strnum, int arity);
      CVOID__PROTO_N(explicit_heap_overflow, int pad, int arity);
      CIAO_REGISTER int i = 0;
      CIAO_REGISTER char *ws = Atom_Buffer;
      tagged_t t, *h;
      
      while ((c = ws[i++] = GETC(f))) {
        if (i == Atom_Buffer_Length) {
            EXPAND_ATOM_BUFFER(Atom_Buffer_Length*2);
            ws = Atom_Buffer;
        }
      }

      /* TODO: This is can be improved. 

         We ensure that there is at least (length/sizeof(tagged_t)+4)
	 available words in the heap. We load the number into the
	 heap, copy it to the bytecode, and then move the heap pointer
	 back.
      */
      {
	int arity = 2;
	int cells = (length / sizeof(tagged_t)) + 4 + arity;
	if (HeapDifference(w->global_top,Heap_End)<cells) {
	  explicit_heap_overflow(Arg,cells,5);
	}
      }
      h = w->global_top;
      if (!string_to_number(w, (unsigned char *)Atom_Buffer, 10, &t, 5)) {
	SERIOUS_FAULT("$qread: wrong number!");
      }
      insn_p += compile_large(t, insn_p);
      w->global_top = h;
      break;
#else
      CIAO_REGISTER int i = 0;
      CIAO_REGISTER char *ws = Atom_Buffer;
      bool_t floatp = FALSE;
      tagged_t *wp = (tagged_t *)insn_p;
      
      while ((c = ws[i++] = GETC(f))) {
        if (c == '.') floatp = TRUE;
        if (i == Atom_Buffer_Length) {
            EXPAND_ATOM_BUFFER(Atom_Buffer_Length*2);
            ws = Atom_Buffer;
        }
      }

      if (floatp) {
        flt64_t f = atof(ws);
        tagged_t *fp = (tagged_t *)(&f);
        
        *wp++ = MakeFunctorFloat;
        *wp++ = fp[0];
        *wp++ = fp[1];
        insn_p += 6;
      } else {
        bn_from_string(ws,(Bignum *)insn_p,(Bignum *)((char *)insn_p+length), GetSmall(current_radix));
        insn_p += LargeArity(*wp)<<1;
      }
      break;
#endif
    }
      
    case '+': {
      CIAO_REGISTER char *ws = workstring;
      
      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
          SERIOUS_FAULT("workstring length exceeded in getbytecode() [+]")
#endif
        ;

      *(intmach_t *)insn_p = atol(workstring);
      insn_p += BPL;
      break;
    }
      
    case 'C': {
      CIAO_REGISTER char *ws = workstring;
      
      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
          SERIOUS_FAULT("workstring length exceeded in getbytecode() [C]")
#endif
        ;

      *(CInfo *)insn_p = builtintab[atoi(workstring)];
      insn_p += BPTP;
      break;
    }
      
    case 'F': {
      CIAO_REGISTER char *ws = workstring;
      
      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
          SERIOUS_FAULT("workstring length exceeded in getbytecode() [F]" )
#endif
        ;

      /* Was:
       *insn_p++ = (insn_t)builtintab[atoi(workstring)]);
       ant it issued a mesage since insn_t is short int and *builtintab
       is a pointer to a function */
      *insn_p++ = (insn_t)((unsigned long int)builtintab[atoi(workstring)]);
      break;
    }
    case 'D': {
      CIAO_REGISTER char *ws = workstring;
      
      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
          SERIOUS_FAULT("workstring length exceeded in getbytecode() [D]")
#endif
        ;

      *(CInfo *)insn_p = builtintab[atoi(workstring)];

      ws = workstring;

      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
          SERIOUS_FAULT("workstring length exceeded in getbytecode() [D+]")
#endif
        ;

      *(long *)insn_p += atol(workstring);
      insn_p += BPTP;
      break;
    }
      
    case 'E': {
      CIAO_REGISTER char *ws = workstring;
      
      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
          SERIOUS_FAULT("workstring length exceeded in getbytecode() [E]")
#endif
        ;

      *(unsigned long *)insn_p =
        (unsigned long)builtintab[atoi(workstring)] - (unsigned long)insn_p;
      insn_p += BPTP;
      break;
    }
      
    default: {
      CIAO_REGISTER char *ws = workstring;
      
      *ws++ = c;
      while ((*ws++ = GETC(f)))
#if defined(DEBUG)
        if ((ws - workstring) > WORKSTRINGLEN)
         SERIOUS_FAULT("workstring length exceeded in getbytecode() [default]")
#endif
        ;

      *insn_p++ = atoi(workstring);
    }
    }
  }
#if defined(USE_REWRITE_BYTECODE)
  CVOID__CALL_N(bytecode_rewrite, begin, insn_p);
#endif
}



