/*
 *  term_support.c
 *
 *  Term compiler for assert/record.
 *
 *  Copyright (C) 1996,1997,1998, 1999, 2000, 2001, 2002 UPM-CLIP
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "threads.h"
#include "datadefs.h"
#include "gcdatadefs.h"
#include "support.h"
#include "support_macros.h"
#include "wamsupport.h"
#include "instrdefs.h"
#include "task_areas.h"                 /* For register bank reallocation */
#include "objareas.h"

/* declarations for global functions accessed here */

#include "float_tostr.h"
#include "term_support.h"
#include "start.h"
#include "alloc.h"
#include "wam_alloc.h"
#include "inout.h"

/* local declarations */

static CBOOL__PROTO_N(c_term, 
		      register tagged_t t,
		      int Xreg,
		      int FreeReg,
		      int x_variables,
		      tagged_t **trail_origo,
		      bcp_t *current_insn);
static CBOOL__PROTO_N(prolog_constant_codes, 
		      bool_t atomp,
		      bool_t numberp,
		      int ci);
static CFUN__PROTO_N(emit_unify_void, bcp_t, register bcp_t P);
static CVOID__PROTO_N(c_term_size,
		      register tagged_t t,
		      int temps,
		      int *cells,
		      int *maxtemps,
		      int *insns,
		      tagged_t **trail_origo);
static CVOID__PROTO_N(c_term_trail_push,
		      tagged_t t,
		      tagged_t **trail_origo);
static CVOID__PROTO_N(copy_it, register tagged_t *loc);
static CVOID__PROTO_N(copy_it_nat, register tagged_t *loc);

#define TopOfOldHeap TagToHVA(w->global_uncond)
#define Xop(X)	(((X)+WToX0)*sizeof(tagged_t))
#define Xinv(X)	(((X)/sizeof(tagged_t))-WToX0)
#define ODDOP(Insn) \
  if (((unsigned int)P)&3) \
    EMIT(Insn-0); \
  else \
    EMIT(Insn-1), *P++ = 0;
#define EVENOP(Insn) \
  if (!(((unsigned int)P)&3)) \
    EMIT(Insn-0); \
  else \
    EMIT(Insn-1), *P++ = 0;
#define EMIT(I) Last_Insn=P, *P++ = (I)


#if defined(ANDPARALLEL)
#define GCTEST(Pad) \
{ \
  if (HeapDifference(w->global_top,Heap_End) < (Pad)) \
    heap_overflow(Arg,Pad,LOCAL,NULL);	\
  if (ChoiceDifference(w->node,w->trail_top) < (Pad)) \
    choice_overflow(Arg,Pad); \
}
#else
#define GCTEST(Pad) \
  { \
    if (HeapDifference(w->global_top,Heap_End) < (Pad)) \
      heap_overflow(Arg,Pad); \
    if (ChoiceDifference(w->node,w->trail_top) < (Pad)) \
      choice_overflow(Arg,Pad); \
  }
#endif

static CVOID__PROTO_N(copy_it, tagged_t *loc);
static CVOID__PROTO_N(copy_it_nat, tagged_t *loc);

CVOID__PROTO_N(choice_overflow, int pad);
CVOID__PROTO_N(number_to_string, tagged_t term, int base);
void checkdealloc(tagged_t *ptr, int decr);
int  bn_from_string(char *x, Bignum *z, Bignum *zmax, int base);
CVOID__PROTO_N(explicit_heap_overflow, int pad, int arity);
CVOID__PROTO_N(bn_to_string, Bignum *x, int base);

CVOID__PROTO_N(push_choicept, try_node_t *alt);
CVOID__PROTO_N(push_frame, int arity);
CVOID__PROTO(pop_frame);
CVOID__PROTO(pop_choicept);

#if defined(ANDPARALLEL) || defined(PARBACK)
CFUN__PROTO_N(heap_overflow, int, int pad, bool_t remote_reloc,worker_t *remote_worker);
#else
CFUN__PROTO_N(heap_overflow, int, int pad);
#endif

/* I believe the following must be shared and locked, because they are
   related to code compilation */


/*static insn_t*/
 /* *current_insn, */        /* Inside compile_term_aux and passed around */
 /* *last_insn;    */                          /* Inside compile_term_aux */


/*static int*/
 /* x_variables, */       /* Now inside compile_term_aux and pased around */
 /* cells,       */                         /* Inited in compile_term_aux */
 /* insns,       */                         /* Inited in compile_term_aux */
 /* maxtemps;    */                         /* Inited in compile_term_aux */

/*static tagged_t*/
 /* *trail_origo;*/                         /* Inited in compile_term_aux */


static CVOID__PROTO_N(c_term_trail_push, tagged_t t, tagged_t **trail_origo);
static CVOID__PROTO_N(c_term_size,
		      tagged_t t, int temps,
		      int *cells, int *maxtemps, int *insns, tagged_t **trail_origo);
static CFUN__PROTO_N(emit_unify_void, bcp_t, bcp_t P);
static CBOOL__PROTO_N(c_term, tagged_t t, int Xreg, int FreeReg,
		      int x_variables, tagged_t **trail_origo,
		      bcp_t *current_insn);
CBOOL__PROTO_N(compile_term, worker_t **new_worker);
CFUN__PROTO_N(compile_term_aux, instance_t *,
	      tagged_t head, tagged_t body, worker_t **new_worker);

static CVOID__PROTO_N(c_term_trail_push, tagged_t t, tagged_t **trail_origo)
{
  if (!ChoiceDifference(w->node,w->trail_top)) {
    CIAO_REGISTER tagged_t *tr = w->trail_top;
    int reloc;

    choice_overflow(Arg,-CHOICEPAD);
    reloc = (char *)w->trail_top - (char *)tr;
    *trail_origo = (tagged_t *)((char *)*trail_origo + reloc);
    tr = w->trail_top;
    while (tr[-1] & QMask)
      TrailPop(tr) += reloc;
    while (TrailYounger(tr,*trail_origo+DynamicPreserved))
      CTagToPointer(TrailPop(tr)) += reloc;
  }
  TrailPush(w->trail_top,t);
}

static CVOID__PROTO_N(c_term_size,
		      CIAO_REGISTER tagged_t t,
		      int temps,
		      int *cells, int *maxtemps, int *insns,
		      tagged_t **trail_origo)
{
  CIAO_REG_2(tagged_t, t1);
  CIAO_REGISTER int i, arity;

 start:
  DerefHeapSwitch(t,t1,{goto var_size;});
  if (!(t & TagBitComplex))
    {
      if (t&QMask && t&TagBitFunctor)
	{
	  if (!(t&3))		/* unify_variable */
	    *insns += 1,
	    CTagToPointer(CTagToPointer(t))++;
	  else			/* unify_value */
	    *insns += 1;
	}
      else if (t!=atom_nil)
	*insns += 1+BPT;	/* unify_ + pad + constant */
      return;
    }
  else if (!(t & TagBitFunctor))
    {
      *insns += 5;		/* unify_ + xi + get_list + xj + 2*u */
      *cells += 2;
      if (*maxtemps < temps)
	*maxtemps = temps;
      RefCar(t1,t);
      c_term_size(Arg, t1, temps+1, cells, maxtemps, insns, trail_origo);
      RefCdr(t,t);
      goto start;
    }
  else
    {
      if (STRIsLarge(t))
	{
	  arity = LargeInsns(TagToHeadfunctor(t));
	  *insns += 5+arity;
	  /* unify_ + xi + get_large + xj + pad + functor + largeNum */
	  *cells += 1+(arity>>1);
	  return;
	}
      arity = Arity(TagToHeadfunctor(t));
      *insns += 4+BPT+arity;	/* unify_ + xi + get_structure
				   + pad + functor + xj + arity*u */
      *cells += 1+arity;
      if (*maxtemps < temps)
	*maxtemps = temps;
      for (i=1; i<arity; i++)
	{
	  RefArg(t1,t,i);
	  c_term_size(Arg, t1, temps+arity-i, cells, maxtemps, insns, trail_origo);
	}
      RefArg(t,t,arity);
      goto start;
    }
 var_size:
  if (t & TagBitCVA)
    {
      CTagToPointer(t) = Tag(ATM,w->trail_top)|QMask|2;
      c_term_trail_push(Arg,t,trail_origo);
      *insns += 5;	/* unify_variable + xi
			   + get_constraint + xj + 2*u */
      if (*maxtemps < temps)
	*maxtemps = temps;
      t = Tag(LST,TagToGoal(t));
      goto start;
    }
  else
    {
      CTagToPointer(t) = Tag(ATM,w->trail_top)|QMask;
      c_term_trail_push(Arg,t,trail_origo);
      *insns += 1;		/* unify_variable + xi */
    }
  return;
}

static CFUN__PROTO_N(emit_unify_void, bcp_t, 
		     CIAO_REGISTER bcp_t P)
{
  switch (*Last_Insn)
    {
    case UNIFY_VOID_1:
    case UNIFY_VOID_2:
    case UNIFY_VOID_3:
      (*Last_Insn)++;
      break;
    case UNIFY_VOID_4:
      (*Last_Insn) = UNIFY_VOID;
      *P++ = 5;
      break;
    case UNIFY_VOID:
      (*(P-1))++;
      break;
    default:
      EMIT(UNIFY_VOID_1);
    }
  return P;
}

static CBOOL__PROTO_N(c_term, 
		      CIAO_REGISTER tagged_t t,
		      int Xreg, int FreeReg,
		      int x_variables,
		      tagged_t **trail_origo,
		      bcp_t *current_insn)
{
  bcp_t psave;
  tagged_t *ssave;
  int i, ar = ~0, decr, Treg;
  CIAO_REGISTER bcp_t P = *current_insn;
  CIAO_REGISTER tagged_t *s = NULL;
  CIAO_REGISTER tagged_t t1;

  /* 1:  Emit GET instruction for term's principal functor. */

  switch (TagOf(t))		/* t is already dereferenced */
    {
    case LST:
      ar = 2;
      s = TagToLST(t);
      EMIT(GET_LIST);
      *P++ = Xop(Xreg);
      break;
    case STR:
      if (STRIsLarge(t))
	{
	  EVENOP(GET_LARGE);
	  *P++ = Xop(Xreg);
	  P += compile_large(t, P);
	  *current_insn = P;
	  return TRUE;
	}
      ar = Arity(TagToHeadfunctor(t));
      s = TagToArg(t,1);
      EVENOP(GET_STRUCTURE);
      *P++ = Xop(Xreg);
      *(tagged_t *)P = TagToHeadfunctor(t);
      P += BPT;
      break;
    case ATM:
      if (t & QMask)
	goto term_is_var;
      else if (t==atom_nil)
	{
	  EMIT(GET_NIL);
	  *P++ = Xop(Xreg);
	  *current_insn = P;
	  return TRUE;
	}
    case NUM:
      EVENOP(GET_CONSTANT);
      *P++ = Xop(Xreg);
      *(tagged_t *)P = t;
      P += BPT;
      *current_insn = P;
      return TRUE;

    term_is_var:
      {
	if ((t&3)!=3)		/* get_variable */
	  {
	    EMIT(GET_X_VARIABLE);
	    *P++ = Xop(Xreg);
	    *P++ = Xop(TagToPointer(t) - *trail_origo);
	    if (t&2)		/* enqueue constraint */
	      c_term_trail_push(Arg,t,trail_origo);
	    CTagToPointer(CTagToPointer(t)) |= 3;
	  }
	else			/* get_value */
	  {
	    EMIT(GET_X_VALUE);
	    *P++ = Xop(Xreg);
	    *P++ = Xop(TagToPointer(t) - *trail_origo);
	  }
	*current_insn = P;
	return TRUE;
      }
    }

  /* 2. Emit tail-recursive UNIFY sequence for all subargs. */

  psave = P;
  ssave = s;
  Treg = reg_bank_size;

  for (i=1; i<=ar; i++)
    {
      RefHeapNext(t,s);
      DerefHeapSwitch(t,t1,goto arg_is_void;);
      switch (TagOf(t))
	{
	case LST:
	  if ((i==ar) && (Treg==reg_bank_size))
	    {
	      EMIT(UNIFY_LIST);
	      s = TagToLST(t);
	      i=0, ar=2;
	    }
	  else
	    {
	      EMIT(UNIFY_X_VARIABLE);
	      *P++ = Xop(Treg++);
	    }
	  break;
	case STR:
	  if (STRIsLarge(t))
	    {
	      if ((i==ar) && (Treg==reg_bank_size))
		{
		  ODDOP(UNIFY_LARGE);
		  P += compile_large(t, P);
		}
	      else
		{
		  EMIT(UNIFY_X_VARIABLE);
		  *P++ = Xop(Treg++);
		}
	    }
	  else if ((i==ar) && (Treg==reg_bank_size))
	    {
	      ODDOP(UNIFY_STRUCTURE);
	      *(tagged_t *)P = TagToHeadfunctor(t);
	      P += BPT;		
	      s = TagToArg(t,1);
	      i=0, ar=Arity(TagToHeadfunctor(t));
	    }
	  else
	    {
	      EMIT(UNIFY_X_VARIABLE);
	      *P++ = Xop(Treg++);
	    }
	  break;
	case ATM:
	  if (t & QMask)
	    goto arg_is_var;
	  else if (t==atom_nil)
	    {
	      EMIT(UNIFY_NIL);
	      break;
	    }
	case NUM:
	  ODDOP(UNIFY_CONSTANT);
	  *(tagged_t *)P = t;
	  P += BPT;
	  break;

	arg_is_var:
	  {
	    if ((t&3)!=3)	/* unify_variable */
	      {
		EMIT(UNIFY_X_VARIABLE);
		*P++ = Xop(TagToPointer(t) - *trail_origo);
		if (t&2)	/* enqueue constraint */
		  c_term_trail_push(Arg,t,trail_origo);
		CTagToPointer(CTagToPointer(t)) |= 3;
	      }
	    else		/* unify_value */
	      {
		EMIT(UNIFY_X_VALUE);
		*P++ = Xop(TagToPointer(t) - *trail_origo);
	      }
	    break;
	  }

        arg_is_void:
	  P = emit_unify_void(Arg, P);
	  break;
	}
    }

  /* 3. Scan emitted code and recursively emit code for nested args. */

  *current_insn = P;
  if (FreeReg < x_variables)
    return FALSE;
  if (Treg==reg_bank_size)
    return TRUE;
  decr = Treg-1-FreeReg;
  s = ssave;
  P = psave;
  psave = *current_insn;

  while (P < psave)
  {
#if defined(APOLLO_CC_BUG)
    P++;
    switch (*(P-1))
#else
    switch (*P++)
#endif
      {
      case UNIFY_LIST:
	DerefHeap(t,s);
	s = TagToLST(t);
	break;
      case UNIFY_STRUCTUREQ:
	P++;
      case UNIFY_STRUCTURE:
	P += BPT;
	DerefHeap(t,s);
	s = TagToArg(t,1);
	break;
      case UNIFY_LARGEQ:
	P++;
      case UNIFY_LARGE:
	P += LargeInsns(*(tagged_t *)P);
	(void)HeapNext(s);
	break;
      case UNIFY_CONSTANTQ:
	P++;
      case UNIFY_CONSTANT:
	P += BPT;
      case UNIFY_NIL:
	(void)HeapNext(s);
	break;
      case UNIFY_X_VARIABLE:
	i = Xinv(*P);
	if (i>=reg_bank_size)
	  {
	    *P++ = Xop(i-decr);
	    DerefHeapNext(t,s);
	    if (!c_term(Arg,t,i-decr,i-decr,x_variables,
                        trail_origo,current_insn))
	      return FALSE;
	    break;
	  }
      case UNIFY_X_VALUE:
	P++;
	(void)HeapNext(s);
	break;
      case UNIFY_VOID:
	s += *P++;
	break;
      case UNIFY_VOID_4:
	(void)HeapNext(s);
      case UNIFY_VOID_3:
	(void)HeapNext(s);
      case UNIFY_VOID_2:
	(void)HeapNext(s);
      case UNIFY_VOID_1:
	(void)HeapNext(s);
	break;
      default:
	SERIOUS_FAULT("compile_term: internal error");
      }
  }
  return TRUE;
}

/* ASSERT: X(0) is always a dereferenced list. */
/* Also, returns in the second argument a pointer to a non-null worker
   pointer if the worker has changed, or to null if it has. */

CBOOL__PROTO_N(compile_term, 
	       worker_t **new_worker)
{
  tagged_t head, body;
  instance_t *object;


  RefCar(head,X(0));
  RefCdr(body,X(0));

  /*
#if defined(DEBUG)
  display_term(Arg, head, Output_Stream_Ptr, FALSE);
  putchar('\n');
#endif
  */

  *new_worker = NULL;                             /* may be changed after */

  object = compile_term_aux(Arg, head, body, new_worker);
  Arg = *new_worker == NULL ? Arg : *new_worker;

  Unify_constant(PointerToTerm(object),X(1));
  return TRUE;
}

/* Note on memory consumption: compile_term_aux may increase the size of the
   stacks and the size of the X register bank, so the simple approach "see
   how much the total memory has increased" does not work.  Instead, we try
   to find out exactly where we allocate and deallocate memory for program
   storage.  c_term_size does not allocate program memory (instead, it may
   call choice_overflow, which expands stacks). */

CFUN__PROTO_N(compile_term_aux, instance_t *,
	      tagged_t head, tagged_t body,
	      worker_t **new_worker)
{
  int truesize;
  //  CIAO_REGISTER tagged_t t0, *pt1, *pt2;
  CIAO_REG_1(tagged_t, t0);
  CIAO_REG_2(tagged_t *, pt1);
  CIAO_REG_3(tagged_t *, pt2);
  instance_t *object = NULL;
  
  int x_variables, cells, insns, maxtemps;
  tagged_t *trail_origo;
  bcp_t current_insn /*, *last_insn */ ;

  /* TODO insns=0? */
#if FLEXIBLE_SIZE+0 == 0
  insns=1;
#else
  insns=1-FLEXIBLE_SIZE;
#endif
  cells=CONTPAD;
  maxtemps=0;
  trail_origo = Arg->trail_top-DynamicPreserved;

  DerefHeapSwitch(head,t0,{goto car_done;});
  insns += 2;			/* get + arg */
  c_term_size(Arg, head, 0, &cells, &maxtemps, &insns, &trail_origo);
 car_done:
  DerefHeapSwitch(body,t0,{goto cdr_done;});
  insns += 2;			/* get + arg */
  c_term_size(Arg, body, 0, &cells, &maxtemps, &insns, &trail_origo);
 cdr_done:

				/* allow for heapmargin_call insn */
  if (cells>=STATIC_CALLPAD) /* (was SOFT_HEAPPAD) */
    insns += 5;

				/* tidy out void vars */
  pt1 = pt2 = trail_origo+DynamicPreserved;
  while (pt1 < Arg->trail_top) {
    t0 = TrailNext(pt1);
    if (CTagToPointer(t0) & 3)
      CTagToPointer(t0) -= (char *)(pt1-1)-(char *)pt2,	TrailPush(pt2,t0);
    else
      CTagToPointer(t0) = t0;
  }
  Arg->trail_top = pt2;
  x_variables = pt2-trail_origo;

				/* ensure enough X registers */
  if (x_variables+maxtemps > reg_bank_size) {
      int size0 = reg_bank_size;
      
      if (x_variables+maxtemps > (1<<14) - WToX0 - maxtemps)
	goto sizebomb;

      reg_bank_size=x_variables+maxtemps;

      *new_worker =                                /* For local use */
	checkrealloc_FLEXIBLE(worker_t,
			      tagged_t,
			      size0,
			      reg_bank_size,
			      Arg);
#if defined(DEBUG)
      fprintf(stderr, "Reallocing WRB from %lx to %lx\n",
              (long int)Arg, (long int)*new_worker);
#endif
      Arg = *new_worker;
    }

  checkalloc_FLEXIBLE_S(instance_t,
                        objsize,
                        insn_t,
                        insns,
                        object);
  INC_MEM_PROG(object->objsize);
  current_insn = object->emulcode;
  object->pending_x2 = NULL;
  object->pending_x5 = NULL;

  if (cells>=STATIC_CALLPAD) { /* (was SOFT_HEAPPAD) */
    CIAO_REGISTER bcp_t P = current_insn;

    ODDOP(HEAPMARGIN_CALL);
    *(long *)P = cells;
    P += BPL;
    *P++ = DynamicPreserved;
    current_insn = P;
  }

  if (!IsVar(head) &&
      !c_term(Arg,head,0,reg_bank_size-1,x_variables,
              &trail_origo,&current_insn))
    goto sizebomb;
  if (!IsVar(body) &&
      !c_term(Arg,body,1,reg_bank_size-1,x_variables,
              &trail_origo,&current_insn))
    goto sizebomb;

  while (Arg->trail_top[-1] & QMask) {
    bcp_t P = current_insn;

    if (!TrailYounger(Arg->trail_top,Trail_Start))
      break;
    t0 = TrailPop(Arg->trail_top);
    if (!c_term(Arg,
                Tag(LST,TagToGoal(CTagToPointer(t0))),
                TagToPointer(t0)-trail_origo,
                  reg_bank_size-1,
                x_variables,
                &trail_origo,
                &current_insn))
      goto sizebomb;
    *P = GET_CONSTRAINT;
  }

  *current_insn++ = DYNAMIC_NECK_PROCEED;
  truesize = SIZEOF_FLEXIBLE_STRUCT(instance_t, insn_t, current_insn - object->emulcode);
  if (truesize > object->objsize) {
    DEC_MEM_PROG(object->objsize);
    checkdealloc_FLEXIBLE_S(instance_t, objsize, object);
    SERIOUS_FAULT("bug: memory overrun in assert or record");
  }

  if (IsVar(head))
    ;				/* findall record---make it fast */
  else {
    INC_MEM_PROG(truesize - object->objsize);
    object=(instance_t *)checkrealloc((tagged_t *)object, object->objsize, truesize);
    object->objsize = truesize;
  }

  pt2 = Arg->trail_top;
  Arg->trail_top = trail_origo+DynamicPreserved;
  while (TrailYounger(pt2,Arg->trail_top))
    PlainUntrail(pt2,t0,;)

  if (TagIsSTR(head))  {
    DerefArg(t0,head,1);
    if (TagIsSTR(t0))
      object->key = TagToHeadfunctor(t0);
    else if (TagIsLST(t0))
      object->key = functor_list;
    else if (!IsVar(t0))
      object->key = t0;
    else
      object->key = ERRORTAG;
  } else
    object->key = ERRORTAG;

  return object;

 sizebomb:
  DEC_MEM_PROG(object->objsize);
  checkdealloc_FLEXIBLE_S(instance_t, objsize, object);
  SERIOUS_FAULT("term too large in assert or record");
}


#if defined(OLD_DATABASE)
/* Support for current_key/2: given a '$current instance'/2 instance,
   the key of which is a large number, decode the key. */
tagged_t decode_instance_key(instance_t *inst)
{
  CIAO_REGISTER bcp_t p = inst->emulcode;
  int xreg = -1;

  for (;;)
  {
#if defined(APOLLO_CC_BUG)
    p++;
    switch (*(p-1))
#else
    switch (*p++)
#endif
      {
      case HEAPMARGIN_CALLQ:
	p++;
      case HEAPMARGIN_CALL:
	p += BPT+1;
	break;
      case GET_CONSTANTQ:
      case GET_STRUCTUREQ:
	p++;
      case GET_CONSTANT:
      case GET_STRUCTURE:
	p += 1+BPT;
	break;
      case GET_LARGEQ:
	p++;
      case GET_LARGE:
	if (xreg == Xinv(*p))
	  return MakeLarge(p+1);
	p += 1+LargeInsns(*(tagged_t *)(p+1));
	break;
      case UNIFY_CONSTANTQ:
      case UNIFY_STRUCTUREQ:
	p++;
      case UNIFY_CONSTANT:
      case UNIFY_STRUCTURE:
	p += BPT;
	break;
      case UNIFY_LARGEQ:
	p++;
      case UNIFY_LARGE:
	p += LargeInsns(*(tagged_t *)p);
	break;
      case UNIFY_X_VARIABLE:
	if (xreg == -1)
	  xreg = Xinv(*p);
      case UNIFY_X_VALUE:
      case UNIFY_VOID:
      case GET_LIST:
      case GET_CONSTRAINT:
      case GET_NIL:
	p++;
	break;
      case GET_X_VARIABLE:
      case GET_X_VALUE:
	p += 2;
      default:
	break;
      }
  }
}
#endif



static int              /* Shared, no locked --- but it should be private! */
  radixlim1,
  radixlim2,
  radixlim3;

bool_t prolog_init_radix(void)
{
  int radix = GetSmall(current_radix);

  if (radix<10)
    radixlim1 = '0'+radix,
    radixlim2 = 'a',
    radixlim3 = 'A';
  else
    radixlim1 = '0'+10,
    radixlim2 = 'a'+radix-10,
    radixlim3 = 'A'+radix-10;
  return TRUE;
}

static CBOOL__PROTO_N(prolog_constant_codes, bool_t a,bool_t n, int ci);

CBOOL__PROTO(prolog_name)
{
  // ERR__FUNCTOR("atomic_basic:name", 2);
  return prolog_constant_codes(Arg,TRUE,TRUE,1);
}

CBOOL__PROTO(prolog_atom_codes)
{
  // ERR__FUNCTOR("atomic_basic:atom_codes", 2);
  return prolog_constant_codes(Arg,TRUE,FALSE,1);
}

CBOOL__PROTO(prolog_number_codes_2)
{
  // ERR__FUNCTOR("atomic_basic:number_codes", 2);
  return prolog_constant_codes(Arg,FALSE,TRUE,1);
}

CBOOL__PROTO(prolog_number_codes_3)
{
  // ERR__FUNCTOR("atomic_basic:number_codes", 3);
  return prolog_constant_codes(Arg,FALSE,TRUE,2);
}


 /* INTEGER :== [minus]{digit} */
 /* FLOAT :== [minus]{digit}.{digit}[exp[sign]{digit}] */
 /* ATOM :== {char} */

/* string_to_number() tries to convert the string pointed to by AtBuf in a
   number contained in the tagged word *strnum.  If this cannot be done
   (e.g., AtBuf does not correspond syntactically to a number), FALSE is
   returned.  Otherwise, TRUE is returned and the conversion is done */

CBOOL__PROTO_N(string_to_number, 
	       char *AtBuf,
	       int base,
	       tagged_t *strnum,
	       int arity)
{
  bool_t sign = FALSE;
  char *s = AtBuf;
  int i, d;
  flt64_t m = 0.0;
  double num;
  int exp=0,e=0,se=1;
  i = *s++;
  if (i=='-') {
    sign = TRUE;
    i = *s++;
  }
  /* First, consider special cases */
  if ((i=='0') &&
     (s[0]==FLOAT_POINT) &&
     (s[1]=='N') &&
     (s[2]=='a') &&
     (s[3]=='n') &&
     (s[4]=='\0'))
  {
    *strnum=MakeFloat(Arg,0.0/0.0); /* Nan */
    return TRUE;
  }
  if ((i=='0') &&
     (s[0]==FLOAT_POINT) &&
     (s[1]=='I') &&
     (s[2]=='n') &&
     (s[3]=='f') &&
     (s[4]=='\0'))
  {
    if (sign) 
      num = -1.0/0.0;
    else
      num = 1.0/0.0;
    *strnum=MakeFloat(Arg,num); /* Inf */
    return TRUE;
  }
  d = char_digit[i];
  while (0 <= d && d < base) {
    i = *s++;
    d = char_digit[i];
  }
  if ((s - AtBuf) - sign > 1) {
    if (i==0) {
      /* It is an integer, either a small or a bignum */
      CIAO_REGISTER tagged_t *h = w->global_top;
      tagged_t t;
      int req = bn_from_string(AtBuf, h, Heap_End-CONTPAD, base);

      if (req) {
        explicit_heap_overflow(Arg ,req+CONTPAD, arity);
        h = w->global_top;
        if (bn_from_string(Atom_Buffer, h, Heap_End-CONTPAD, base))
          SERIOUS_FAULT("miscalculated size of bignum");
      }

      req = LargeArity(h[0]);
      if (req==2 && IntIsSmall((int)h[1]))
        t = MakeSmall(h[1]);
      else {
        w->global_top += req+1;
        h[req] = h[0];
        t = Tag(STR,h);
      }
      *strnum = t;                 /* The tagged_t word --- small or bignum */
      return TRUE;
    }

/* It is a float. Note that if is a float, after the point the digits
   must not be upper case to avoid confussion with the exponent
   indicator. This is true only if base > 10 + 'e'-'a'*/
    if (i==FLOAT_POINT) {
      s = AtBuf + (sign ? 1 : 0);
      i = *s++;
      do {
	m = m * base+char_digit[i];
	i = *s++;
      }
      while(i!=FLOAT_POINT);
      i = *s++;
      d = char_digit[i];
      if ((0 <= d) && (d < base)) {
	char exponent_lower = exponents_lower[base];
	char exponent_upper = exponents_upper[base];
        do {
	  m = m * base + char_digit[i];
	  exp++;
          i = *s++;
	  d = char_digit[i];
        } while ((0 <= d) && (d < base) && 
                 (i != exponent_lower) && (i != exponent_upper));
        if ((i==exponent_lower) || (i==exponent_upper)) {
          i = *s++;
          if ((i=='+') || (i=='-')) {
	    if(i=='-')
	      se=-1;
	    i = *s++;
	  };
	  d = char_digit[i];
          if ((0<=d) && (d<base)) {
            do {
	      e = e * base + d;
              i = *s++;
	      d = char_digit[i];
	    }
	    while ((0<=d) && (d<base));
          } else i = -1;
        };
      } else i = -1;
      if (i!=-1) {
	if(se==1) {
	  exp = -exp + e;
	} else exp = -exp - e;
	num = m * powl_int(base, exp);
	*strnum = MakeFloat(Arg, sign ? -num : num);
	return TRUE;
      }
    }
  }
  /* Could not make the conversion --- maybe a non-numeric atom */
  return FALSE;
}

/*

  Note: The ci parameter indicates where are the String argument.  If
  ci = 2, then there are 3 parameters, indicating that the second
  parameter could be the numeric base.

 */
static CBOOL__PROTO_N(prolog_constant_codes, 
		      bool_t atomp,
		      bool_t numberp,
		      int ci)
{
  /* Note: This ERR__FUNCTOR is not related to an exported predicate,
     since prolog_constant_codes is called from other places. I have
     some ideas of refining error reporting without hindering
     performance (please, ask me before changing this code). -- JFMC
   */
  ERR__FUNCTOR("atomic_basic:$constant_codes", 3);
  CIAO_REGISTER char *s;
  CIAO_REGISTER int i, base;
  CIAO_REGISTER tagged_t car, cdr;

  DEREF(X(0),X(0));
  DEREF(X(1),X(1));
  if (ci==2) DEREF(X(2),X(2));

    /* Construct a character string from the input list */
    cdr = X(ci);
    s = Atom_Buffer;
    for (i=0; cdr!=atom_nil; i++) {
      if (IsVar(cdr)) {
	goto construct_list;
      } else if (!TagIsLST(cdr)) {
        BUILTIN_ERROR(DOMAIN_ERROR(CHARACTER_CODE_LIST),X(ci),ci+1);
      } else if (i == Atom_Buffer_Length){
        EXPAND_ATOM_BUFFER(Atom_Buffer_Length*2);
        s = Atom_Buffer+i;
      }
      DerefCar(car,cdr);
      if (IsVar(car)) {
        goto construct_list;
      }
      if (!TagIsSmall(car) || (car<=TaggedZero) || (car>=MakeSmall(256))) {
	BUILTIN_ERROR(DOMAIN_ERROR(CHARACTER_CODE_LIST),X(ci),ci+1);
      }
      *s++ = GetSmall(car);
      DerefCdr(cdr,cdr);
    }
    if (i == Atom_Buffer_Length) {
      EXPAND_ATOM_BUFFER(Atom_Buffer_Length*2);
      s = Atom_Buffer+i;
    }
    *s++ = '\0';

    /* s contains now the string of character codes, and i its size */

    if (i>=MAXATOM) atomp = FALSE;  /* Unneded with dynamic atom sizes */

    if (numberp) {
      tagged_t result;
      if (ci==2) {
	if (IsInteger(X(1))) {
	  base = GetSmall(X(1));
	} else {
	  BUILTIN_ERROR(TYPE_ERROR(INTEGER),X(1),2);
	}
      } else { // if (ci==1)
	base = GetSmall(current_radix);
      }
      if ((base < 2)||(base > 36)) {
	BUILTIN_ERROR(DOMAIN_ERROR(SOURCE_SINK),X(1),2);
      }
      if (string_to_number(Arg, Atom_Buffer, base, &result, ci+1)) {
        return cunify(Arg, result, X(0));
      }
    }
    return atomp && cunify(Arg,init_atom_check(Atom_Buffer),X(0));

 construct_list:
    if (IsVar(X(0))) {
        BUILTIN_ERROR(INSTANTIATION_ERROR,atom_nil,2);
    }

    if (numberp && IsNumber(X(0))) {
      if(ci==2) {
	if (IsInteger(X(1))) {
	  base = GetSmall(X(1));
	} else {
	  BUILTIN_ERROR(TYPE_ERROR(INTEGER),X(1),2);
	}
      } else { // if (ci==1)
	base = GetSmall(current_radix);
      }
      if ((base < 2)||(base > 36)) {
	BUILTIN_ERROR(DOMAIN_ERROR(SOURCE_SINK),X(1),2);
      }
      number_to_string(Arg,X(0),base);
      s = Atom_Buffer;
    } else if (atomp && TagIsATM(X(0))) {
      s = GetString(X(0));
    } else {
      if (numberp) {
        if (atomp) {
          BUILTIN_ERROR(TYPE_ERROR(ATOMIC),X(0),1);
	} else {
	  BUILTIN_ERROR(TYPE_ERROR(NUMBER),X(0),1);
	}
      } else {
	BUILTIN_ERROR(TYPE_ERROR(STRICT_ATOM),X(0),1);
      }
    }

    s += (i = strlen(s));

    if (HeapDifference(w->global_top,Heap_End)<CONTPAD+(i<<1)) {
      explicit_heap_overflow(Arg,CONTPAD+(i<<1),ci+1);
    }

    cdr = atom_nil;
    while (i>0)	{
      i--;
      MakeLST(cdr,MakeSmall(*((unsigned char *)--s)),cdr);
    }
    return cunify(Arg,cdr,X(ci));
  
}

CBOOL__PROTO(prolog_atom_length)
{
  ERR__FUNCTOR("atomic_basic:atom_length", 2);
  DEREF(X(0),X(0));
  DEREF(X(1),X(1));

  if (!TagIsATM(X(0)))
    ERROR_IN_ARG(X(0),1,STRICT_ATOM);

  if (!IsInteger(X(1)) && !IsVar(X(1))) {
    BUILTIN_ERROR(TYPE_ERROR(INTEGER),X(1),2);
  }

#if defined(USE_ATOM_LEN)
  return cunify(Arg,MakeSmall(GetAtomLen(X(0))),X(1));
#else
  return cunify(Arg,MakeSmall(strlen(GetString(X(0)))),X(1));
#endif
}

/* sub_atom(Atom, Before, Lenght, Sub_atom) */
CBOOL__PROTO(prolog_sub_atom)
{
  ERR__FUNCTOR("atomic_basic:sub_atom", 4);
  char *s, *s1;
  int l, b, atom_length;

  DEREF(X(0),X(0));
  DEREF(X(1),X(1));
  DEREF(X(2),X(2));
  DEREF(X(3),X(3));

  if (!TagIsATM(X(0)))
    ERROR_IN_ARG(X(0),1,STRICT_ATOM);
  if (!IsInteger(X(1)))
    ERROR_IN_ARG(X(1),2,INTEGER);
  if (!IsInteger(X(2)))
    ERROR_IN_ARG(X(2),3,INTEGER);

  s = GetString(X(0));
#if defined(USE_ATOM_LEN)
  l = GetAtomLen(X(0));
#else
  l = strlen(s);
#endif

  b = GetInteger(X(1));
  if (b < 0 || b > l)
    return FALSE;

  atom_length = GetInteger(X(2));
  if (atom_length < 0 || atom_length+b > l)
    return FALSE;

  s += b;

  if (Atom_Buffer_Length <= atom_length)
    EXPAND_ATOM_BUFFER(atom_length+1);

  s1 = Atom_Buffer;

  strncpy(s1, s, atom_length);

  *(s1+atom_length) = '\0';

  return cunify(Arg,init_atom_check(Atom_Buffer),X(3));

}

CBOOL__PROTO(prolog_atom_concat)
{
  ERR__FUNCTOR("atomic_basic:atom_concat", 3);
  int new_atom_length;
  char *s, *s1, *s2;

  DEREF(X(0),X(0));
  DEREF(X(1),X(1));
  DEREF(X(2),X(2));

  if (TagIsATM(X(0))) {
    s1 = GetString(X(0));

    if (TagIsATM(X(1))) {
      if (!TagIsATM(X(2)) && !IsVar(X(2))) {
        BUILTIN_ERROR(TYPE_ERROR(STRICT_ATOM),X(2),3);
      }
/* atom_concat(+, +, ?) */
      s2 = GetString(X(1));

#if defined(USE_ATOM_LEN)
      new_atom_length = GetAtomLen(X(0)) + GetAtomLen(X(1)) + 1;
#else
      new_atom_length = strlen(s1) + strlen(s2) + 1;
#endif

#if defined(USE_DYNAMIC_ATOM_SIZE)
      if (new_atom_length >= Atom_Buffer_Length) {
	EXPAND_ATOM_BUFFER(new_atom_length);
      }
#else
      if (new_atom_length > MAXATOM) {
#if defined(DEBUG)
        SERIOUS_FAULT("atom length exceeded in prolog_atom_concat()");
#else
	return FALSE; 
#endif
      }
#endif

      /* Append the two strings in atom_buffer */
      s = Atom_Buffer;
      while (*s1)
        *s++ = *s1++;
      while (*s2)
        *s++ = *s2++;
      *s = '\0';
      return cunify(Arg,init_atom_check(Atom_Buffer),X(2));
    } else if (IsVar(X(1))) {
      if (!TagIsATM(X(2))) { ERROR_IN_ARG(X(2),3,STRICT_ATOM); }
      /* atom_concat(+, -, +) */
      s2 = GetString(X(2));

#if defined(USE_ATOM_LEN)
      new_atom_length = GetAtomLen(X(2))+1;
#else
      new_atom_length = strlen(s2) + 1;
#endif
      if (new_atom_length >= Atom_Buffer_Length)
          EXPAND_ATOM_BUFFER(new_atom_length);

      for ( ; *s1 && *s2 ; s1++, s2++)
        if (*s1 != *s2) return FALSE;

      if (*s1) return FALSE;

      s = Atom_Buffer;

      strcpy(s, s2);

      return cunify(Arg,init_atom_check(Atom_Buffer),X(1));
    } else {
      BUILTIN_ERROR(TYPE_ERROR(STRICT_ATOM),X(1),2);
    }
  } else if (IsVar(X(0))) {
    if (!TagIsATM(X(2)))
        { ERROR_IN_ARG(X(2),3,STRICT_ATOM); }

    if (TagIsATM(X(1))) {
/* atom_concat(-, +, +) */

      s1 = GetString(X(1));
      s2 = GetString(X(2));

#if defined(USE_ATOM_LEN)
      if ((new_atom_length = (GetAtomLen(X(2)) - GetAtomLen(X(1)))) < 0)
        return FALSE;
#else
      if ((new_atom_length = strlen(s2)-strlen(s1)) < 0)
        return FALSE;
#endif

      if (new_atom_length >= Atom_Buffer_Length) 
	EXPAND_ATOM_BUFFER(new_atom_length+1);

      s = s2+new_atom_length;

      if (strcmp(s1, s)) /* different */
        return FALSE;

      s = Atom_Buffer;

      strncpy(s, s2, new_atom_length);

      *(s+new_atom_length) = '\0';

      return cunify(Arg,init_atom_check(Atom_Buffer),X(0));
    } else if (IsVar(X(1))) {
/* atom_concat(-, -, +) */

      s2 = GetString(X(2));
#if defined(USE_ATOM_LEN)
      new_atom_length = GetAtomLen(X(2))+1;
#else
      new_atom_length = strlen(s2)+1;
#endif
      if (new_atom_length >= Atom_Buffer_Length)
        EXPAND_ATOM_BUFFER(new_atom_length);
      X(3) = TaggedZero;
      push_choicept(Arg,address_nd_atom_concat);
      return nd_atom_concat(Arg);
    } else {
      BUILTIN_ERROR(TYPE_ERROR(STRICT_ATOM),X(1),2);
    }
  } else {
    BUILTIN_ERROR(TYPE_ERROR(STRICT_ATOM),X(0),1);
  }
}

/* Precond: 2<=abs(base)<=36 */
CVOID__PROTO_N(number_to_string, tagged_t term, int base)
{
  if (TagIsSmall(term))
    {
      intmach_t l = GetSmall(term);
      char hibase = 'a'-10;
      bool_t sx = (l>=0);
      intmach_t digit;
      CIAO_REGISTER char *c0, *c, d;

      if (base<0)
	hibase = 'A'-10,
	base = -base;
      c = Atom_Buffer;
      if (!sx)
	*c++ = '-',
	l = -l;

      do
	{
	  digit = l % base;
	  l /= base;
	  *c++ = (digit<10 ? '0'+digit : hibase+digit);
	}
      while (l>0);

      *c++ = 0;
      for (c0=Atom_Buffer+1-sx, c-=2; c0<c; c0++, c--)
	d = *c0, *c0 = *c, *c = d;
    }
  else if (IsFloat(term))
    {
      flt64_t f;
      unsigned long *fp = (unsigned long *)(&f);
      char *cbuf;
      int eng_flt_signif = (int)((IEEE754_MANTISSA_LENGTH + 1) * invlog2[base] + 1);

      /* f = GetFloat(term); */

      fp[0] = CTagToArg(term,1);
      fp[1] = CTagToArg(term,2);

      /* Print using the default precision.  'p' is a new 'prolog' :-)
         format not considered by, e.g., the printf family, to
         implement the expected Prolog behavior */

/*       if (1024 > Atom_Buffer_Length) */
/* 	EXPAND_ATOM_BUFFER(102400); */

      cbuf = Atom_Buffer;
      cbuf = float_to_string(cbuf, eng_flt_signif, 'p', f, base);
    }
  else
    bn_to_string(Arg,TagToSTR(term),base);
}


/* copy_term(?Old,?New):
 * Algorithm:
 * If Old is a plain variable, just return.
 * Otherwise allocate a frame containing Old and New.
 * The frame slot for Old will be progressively replaced by a copy of Old.
 * Thus all relevant parts of the old and new structures are reachable from
 * the frame slot, should GC occur.
 * While copying, all old variables encountered are bound to their copies.
 * This requires a choicepoint.
 * Finally untrail but trail any new CVA:s, deallocate frame & choicept,
 * and unify copy with New.
 */
CBOOL__PROTO(prolog_copy_term)
{

  CIAO_REGISTER tagged_t t1, t2, *pt1, *pt2;

  t1 = X(0);
  SwitchOnVar(t1,t2,
	      return TRUE;,
	      ;,
	      return TRUE;,
	      ;);

  X(0) = t1;
  push_choicept(Arg,fail_alt);	/* try, arity=0 */
  push_frame(Arg,2);		/* allocate, size=2 */

  copy_it(Arg,&w->frame->term[0]); /* do the copying */

  pt1 = pt2 = TagToPointer(w->node->trail_top); /* untrail */
  while (!OffTrailtop(pt2,w->trail_top)) {
    t1 = TrailNext(pt2);	/* old var */
    CTagToPointer(t1) = t1;
  }
  w->trail_top = pt1;

  pop_frame(Arg);
  pop_choicept(Arg);		/* trust */
  return cunify(Arg,X(0),X(1));
}


static CVOID__PROTO_N(copy_it, CIAO_REGISTER tagged_t *loc)
{
  CIAO_REGISTER tagged_t t1, t2, *pt1, *pt2;
  CIAO_REGISTER int i;
  int term_so_far;		/* size of new heap before copying subterms */

 start:
  RefHeap(t1,loc);
  SwitchOnHeapVar(t1,t2,{goto copy_hva;},{goto copy_cva;},;);

  if (IsAtom(t1) || IsNumber(t1)) {                           /* NUM, ATM */
    *loc = t1;
    return;
  } else if (t1 & TagBitFunctor) {                                 /* STR */
    pt1 = TagToSTR(t1);
    pt2 = w->global_top;
    *loc = Tag(STR,pt2);
    t2 = HeapNext(pt1), HeapPush(pt2,t2);
    for (i=Arity(t2); i>0; --i) {
      RefHeapNext(t1,pt1);
      HeapPush(pt2,t1);
    }
    w->global_top = pt2;
    term_so_far = HeapDifference(TopOfOldHeap,pt2);
    GCTEST(CHOICEPAD);
    for (i=Arity(t2); i>1; --i)
      copy_it(Arg,HeapOffset(TopOfOldHeap,term_so_far-i));
  } else {				                           /* LST */
    pt1 = TagToLST(t1);
    pt2 = w->global_top;
    *loc = Tag(LST,pt2);
  copy_2_cells:
    RefHeapNext(t1,pt1);
    HeapPush(pt2,t1);
    RefHeapNext(t1,pt1);
    HeapPush(pt2,t1);
    w->global_top = pt2;
    term_so_far = HeapDifference(TopOfOldHeap,pt2);
    GCTEST(CHOICEPAD);
    copy_it(Arg,HeapOffset(w->global_top,-2));
  }
  GCTEST(CHOICEPAD);
  loc = HeapOffset(TopOfOldHeap,term_so_far-1);
  goto start;

 copy_hva:
  if (CondHVA(t1)){		                                   /* HVA */
    PreLoadHVA(*loc,loc);
    t2 = TagHVA(loc);
    BindHVA(t1,t2);
  } else *loc = t1;
  return;

 copy_cva:
  if (CondCVA(t1)){		                       /* new 3-field CVA */
    pt1 = TagToGoal(t1);
    pt2 = w->global_top;
    LoadCVA(t2,pt2);
    BindCVA_NoWake(t1,t2);
    *loc = t2;
    goto copy_2_cells;
  } else *loc = t1;
  return;
}

/* Do not copy attributes */
CBOOL__PROTO(prolog_copy_term_nat)
{
  CIAO_REGISTER tagged_t t1, t2, *pt1, *pt2;

  t1 = X(0);
  SwitchOnVar(t1,t2,
	      return TRUE;,
	      return TRUE;,
	      return TRUE;,
	      ;);

  X(0) = t1;
  push_choicept(Arg,fail_alt);	/* try, arity=0 */
  push_frame(Arg,2);		/* allocate, size=2 */

  copy_it_nat(Arg,&w->frame->term[0]); /* do the copying */

  pt1 = pt2 = TagToPointer(w->node->trail_top); /* untrail */
  while (!OffTrailtop(pt2,w->trail_top)) {
    t1 = TrailNext(pt2);	/* old var */
    CTagToPointer(t1) = t1;
  }
  w->trail_top = pt1;

  pop_frame(Arg);
  pop_choicept(Arg);		/* trust */
  return cunify(Arg,X(0),X(1));
}


static CVOID__PROTO_N(copy_it_nat, CIAO_REGISTER tagged_t *loc)
{
  CIAO_REGISTER tagged_t t1, t2, *pt1, *pt2;
  CIAO_REGISTER int i;
  int term_so_far;		/* size of new heap before copying subterms */

 start:
  RefHeap(t1,loc);
  SwitchOnHeapVar(t1,t2,{goto copy_hva;},{goto skip_cva;},;);

  if (IsAtom(t1) || IsNumber(t1)) {                           /* NUM, ATM */
    *loc = t1;
    return;
  } else if (t1 & TagBitFunctor) {                                 /* STR */
    pt1 = TagToSTR(t1);
    pt2 = w->global_top;
    *loc = Tag(STR,pt2);
    t2 = HeapNext(pt1), HeapPush(pt2,t2);
    for (i=Arity(t2); i>0; --i) {
      RefHeapNext(t1,pt1);
      HeapPush(pt2,t1);
    }
    w->global_top = pt2;
    term_so_far = HeapDifference(TopOfOldHeap,pt2);
    GCTEST(CHOICEPAD);
    for (i=Arity(t2); i>1; --i)
      copy_it_nat(Arg,HeapOffset(TopOfOldHeap,term_so_far-i));
  } else {				                           /* LST */
    pt1 = TagToLST(t1);
    pt2 = w->global_top;
    *loc = Tag(LST,pt2);
    RefHeapNext(t1,pt1);
    HeapPush(pt2,t1);
    RefHeapNext(t1,pt1);
    HeapPush(pt2,t1);
    w->global_top = pt2;
    term_so_far = HeapDifference(TopOfOldHeap,pt2);
    GCTEST(CHOICEPAD);
    copy_it_nat(Arg,HeapOffset(w->global_top,-2));
  }
  GCTEST(CHOICEPAD);
  loc = HeapOffset(TopOfOldHeap,term_so_far-1);
  goto start;

 copy_hva:
  if (CondHVA(t1)){		                                   /* HVA */
    PreLoadHVA(*loc,loc);
    t2 = TagHVA(loc);
    BindHVA(t1,t2);
  } else *loc = t1;
  return;

 skip_cva:
  if (CondCVA(t1)){  /* This code is equivalent to taking out the attribute;
                        xref bu1_detach_attribute() */
    PreLoadHVA(*loc,loc);
    t2 = TagHVA(loc);
    BindCVA_NoWake(t1,t2);
  } else  *loc = t1;
  return;
}



/* Copy a term in a remote worker to the local worker.  Returns the local
   term pointer.  It has (nontermination) problems when copying structures
   with self references. */

CFUN__PROTO_N(cross_copy_term, tagged_t, tagged_t remote_term)
{
  X(0) = remote_term;
  LoadHVA(X(1), w->global_top);
#if defined(DEBUG)
  if (!prolog_copy_term(Arg))
    fprintf(stderr, "Could not copy term in cross_copy_term!!!!\n");
#else
  prolog_copy_term(Arg);
#endif
  return X(1);
}

/* c_cyclic_term checks that the term t is cyclic.

   * Uses gc marks to mark the first argument of the compounds we
     alreardy seen terms. 
   * Must be call with a non-marked term (we take care of that for the
     recursive calls).
*/

CBOOL__PROTO_N(c_cyclic_term_and_mark, tagged_t t)
{
  tagged_t t1;
  tagged_t * ptr;
  int i;

 start:
  switch(TagOf(t)){
  case SVA:
  case HVA:
  case CVA:
    t1 = CTagToPointer(t);
    if (t1 == t) return FALSE;
    t = t1;
    goto start;
  case LST: 
    ptr = TagToLST(t);
    t = *ptr;
    if (gc_IsMarked(t)) return TRUE;
    gc_MarkM(*ptr);
    if (c_cyclic_term_and_mark(Arg, t)) return TRUE;  /* t is not marked */ 
    t = *(++ptr);
    goto start; 
  case STR:
    ptr = TagToSTR(t);
    i = Arity(*(ptr++));
    t = *ptr;
    if (gc_IsMarked(t)) return TRUE;
    gc_MarkM(*ptr);
    for (; i > 1; i--) {
      if (c_cyclic_term_and_mark(Arg, t)) return TRUE;  /* t is not marked */ 
      t = *(++ptr);
    }
    goto start;
  default:
    return FALSE;
  }
}

/* unmark_term removes marks done by c_cyclic_term_and_mark. */ 
CVOID__PROTO_N(unmark_term, tagged_t t)
{
  tagged_t t1;
  tagged_t * ptr;
  int i;

 start:
  switch(TagOf(t)){
  case SVA:
  case HVA:
  case CVA: 
    t1 = CTagToPointer(t);
    if (t1 == t) return;
    t = t1;
    goto start;
  case LST: 
    ptr = TagToLST(t);
    if (!gc_IsMarked(*ptr)) return;
    gc_UnmarkM(*ptr);
    unmark_term(Arg, *(ptr++));  
    t = *ptr;
    goto start; 
  case STR: 
    ptr = TagToSTR(t);
    i = Arity(*(ptr++));
    if (!gc_IsMarked(*ptr)) return;
    gc_UnmarkM(*ptr);
    for (; i > 1; i--) {
      unmark_term(Arg, *(ptr++));
    }
    t = *ptr;
    goto start;
  default:
    return;
  }
}

CBOOL__PROTO_N(c_cyclic_term, tagged_t t)
{
  bool_t r;

  r = c_cyclic_term_and_mark(Arg, t);
  unmark_term(Arg, t);
  return r;
}

CBOOL__PROTO(prolog_cyclic_term)
{
  return c_cyclic_term(Arg, X(0));
}

/* unifiable(Term1, Term2, Unifier)
 *
 * Algortihm: tries unify Term1 and Term2, then untrail the
 * unification while saving the unifier into a prolog list which is
 * eventually unified with Unifier. The function treats attributed 
 * variable as classical ones.
 */
CBOOL__PROTO(prolog_unifiable)
{
  tagged_t t, t1;
  tagged_t * limit, *tr;

  /* Forces trailing of bindings and saves the top of the trail. */
  push_choicept(Arg,fail_alt);	
  /* Saves the arguments in case of GC. */
  push_frame(Arg,3);

  if (!cunify(Arg, X(0), X(1))) return FALSE;  

  /* Makes sure there is enough place in the heap to construct the
     unfiers list. */
  GCTEST((w->trail_top - TagToPointer(w->node->trail_top)) * 5);

  t = atom_nil;
  tr = w->trail_top;
  limit = TagToPointer(w->node->trail_top);
   
  while (TrailYounger(tr, limit)) {
    t1 = TrailPop(tr);

    HeapPush(w->global_top, SetArity(atom_equal, 2));
    HeapPush(w->global_top, t1);
    HeapPush(w->global_top, CTagToPointer(t1));
    HeapPush(w->global_top, Tag(STR, HeapOffset(w->global_top, -3)));
    HeapPush(w->global_top, t);
    t = Tag(LST, HeapOffset(w->global_top, -2));

    CTagToPointer(t1) = t1;
  }

  /* Ignores possible wakes caused by unification of attributed
     variables */ 
  if (TestEvent) Heap_Warn_Soft = Heap_Warn;
  
  w->trail_top = limit;
  pop_frame(Arg);
  pop_choicept(Arg);	

  return cunify(Arg, X(2), t);
}  
