/*
 *  wam_macros.h
 *
 *  Emulator macros.
 *
 *  Copyright (C) 1996,1997,1998, 1999, 2000, 2001, 2002 UPM-CLIP
 */

#ifndef _CIAO_WAM_MACROS_H
#define _CIAO_WAM_MACROS_H

#include "wamsupport.h"
#include "initial.h"

/* --------------------------------------------------------------------------- */

#if defined(DEBUG_NODE)
# define AssignNodeFunctor(F, Y) \
  {F = Y;}
#else
# define AssignNodeFunctor(F, Y)
#endif

#if defined(DEBUG)
   /* stop_on_pred_calls = predtrace | profile */
# if defined(PROFILE)
#  define PredTrace(X,Y) \
{ \
  AssignNodeFunctor(w->node->functor, Y); \
  if (stop_on_pred_calls) \
  { \
    if (predtrace) wr_functor(X,Y); \
    else { \
      PROFILE__HOOK_CALL(w,Y); \
    } \
  } \
}

# else /* DEBUG & !PROFILE */
#  define PredTrace(X,Y) \
{ \
  AssignNodeFunctor(w->node->functor, Y); \
  if (predtrace) \
    wr_call(Arg,X,Y); \
}
# endif
#else
# if defined(DEBUG_NODE)
#  define PredTrace(X,Y) \
  AssignNodeFunctor(w->node->functor, Y)
# else
#  define PredTrace(X,Y)
# endif
#endif

#if defined(DEBUG)
#define TRACE_CHPT_CLEANUP(TopCChpt, TopNode) \
 if(debug_concchoicepoints) \
    fprintf(stderr, "cut: removing chains (%x to %x)\n", \
                    (int)TopCChpt, (int)TopNode);

#define TRACE_CHPT_CUT(NewNode) \
 if (debug_choicepoints)  \
   fprintf(stderr, "Cutting: new chpt = %x\n", (int)NewNode);
#else
#define TRACE_CHPT_CUT(NewNode)
#define TRACE_CHPT_CLEANUP(TopCChpt, TopNode)
#endif

/* TODO: can we reuse this to implement actions on cuts? (JFMC) */
#if defined(THREADS)
#define ConcChptCleanUp(TopCChpt, TopNode) \
    if (ChoiceYounger(TopCChpt, TopNode)) { \
       TRACE_CHPT_CLEANUP(TopCChpt, TopNode); \
       remove_link_chains(&TopCChpt, TopNode); \
    } 
#else
#define ConcChptCleanUp(TopCChpt, TopNode)
#endif

/* --------------------------------------------------------------------------- */

#define P		p
#define B		((node_t *)pt1)
#define E		((frame_t *)pt1)
#define SetE(X)         (pt1 = (tagged_t *)(X))
#define H		pt2
#define S		pt2
#define Func		((definition_t *)P)

/* --------------------------------------------------------------------------- */

#define Setfunc(X)	(P = (bcp_t)(&(X)->enter_instr))
#define SetA(Frame,Ptr) \
{ \
  w->local_top = (frame_t *)(Ptr); \
}

#define SETUP_PENDING_CALL(ADDR) \
{ \
	  ComputeE; \
	  Y(0) = PointerToTerm(Func); \
	  for(i=0; i<Func->arity; i++) Y(i+1) = X(i); \
	  E->next_insn = w->next_insn; \
	  E->frame = w->frame; \
	  w->frame = E; \
	  w->next_insn = &contcode[(1+LOffset)*(i+1)]; \
	  SetA(E,Offset(E,EToY0+i+1)); \
	  Setfunc(ADDR); \
}

/* Do not edit this defn - it's a special case of ComputeA. */
#define ComputeE \
{ \
  if (w->local_top) {                                        \
    SetE(w->local_top);                                     \
  } else {                                                   \
    SetE(NodeLocalTop(w->node));			      \
    if (!StackYounger(E,w->frame))				\
      SetE(StackCharOffset(w->frame,FrameSize(w->next_insn))); \
  } 							\
}

#endif /* _CIAO_WAM_MACROS_H */
