/*
 *  wamfunction.c
 *
 *  Routines for 'call_function' procedures.
 *
 *  Copyright (C) 1996,1997,1998, 1999, 2000, 2001, 2002 UPM-CLIP
 */

#include <limits.h>

#include "datadefs.h"
#include "support_macros.h"

/* declarations for global functions accessed here */

#include "support.h"
#include "bignum.h"
#include "wamfunction.h"
#include "wamsupport.h"

/* local declarations */

static CFUN__PROTO_N(lsh_internal, tagged_t, tagged_t t, int dist, bcp_t p);
static CFUN__PROTO_N(rsh_internal, tagged_t, tagged_t t, int dist, bcp_t p);
/* CFUN__PROTO_N(bn_call, tagged_t, int (*f)(), tagged_t x, tagged_t y, bcp_t op); */

/* Compatibility with non-clang compilers */
#ifndef __has_builtin
#define __has_builtin(x) 0
#endif

/* Signed integer multiplication with overflow checking */

/* (32 bit versions) */
/* A: signed int, B: signed int */
#if __has_builtin(__builtin_smul_overflow)
#define SMUL_OVERFLOW(A, B, P, OVERFLOW) ({ \
  if (__builtin_smul_overflow((A), (B), &P)) { OVERFLOW; } \
    })
#else
#define SMUL_OVERFLOW(A, B, P, OVERFLOW) ({ \
  int64_t R = (int64_t)(A) * (int64_t)(B); \
  if (R > INT_MAX || R < INT_MIN) { OVERFLOW; } \
  P = (int)R; \
})
#endif

CFUN__PROTO_N(fu1_minus, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$-", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  
  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t))
    {
      if (t==TaggedLow)
	return make_integer_check(Arg, GetSmall(TaggedHigh),p);
      else
	return TaggedZero-(t-TaggedZero);
    }
  else if (IsFloat(t))
    return make_float_check(Arg, -GetFloat(t),p);
  else
    return bn_call(Arg,bn_minus,t,0,p);
}

CFUN__PROTO_N(fu1_plus, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$+", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t) || p==NULL)
    return t;
  else if (IsFloat(t))
    return make_float_check(Arg, GetFloat(t),p);
  else
    return bn_call(Arg,bn_plus,t,0,p);
}

CFUN__PROTO_N(fu1_integer, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$integer", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t) || (p==NULL && !IsFloat(t)))
    return t;
  else if (IsFloat(t))
    {
      if (float_is_finite(t)) return bn_call(Arg,bn_from_float,t,0,p);
      else BUILTIN_ERROR(REPRESENTATION_ERROR(NAN_OR_INF_TO_INTEGER), t, 1);
    } 
  else
    return bn_call(Arg,bn_plus,t,0,p);
}

CFUN__PROTO_N(fu1_float, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$float", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (p==NULL && IsFloat(t))
    return t;
  else
    return make_float_check(Arg, GetFloat(t),p);
}

CFUN__PROTO_N(fu1_add1, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$++", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t))
    {
      if (t==TaggedHigh-(1<<SmallShift))
	return make_integer_check(Arg, GetSmall(TaggedHigh),p);
      else
	return t+(1<<SmallShift);
    }
  else if (IsFloat(t))
    return make_float_check(Arg, GetFloat(t) + 1.0,p);
  else
    return bn_call(Arg,bn_incr,t,0,p);
}

CFUN__PROTO_N(fu1_sub1, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$--", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);


  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t))
    {
      if (t==TaggedLow)
	return make_integer_check(Arg, GetSmall(TaggedLow)-1,p);
      else
	return t-(1<<SmallShift);
    }
  else if (IsFloat(t))
    return make_float_check(Arg, GetFloat(t) - 1.0,p);
  else
    return bn_call(Arg,bn_decr,t,0,p);
}

				/* binary functions */

CFUN__PROTO_N(fu2_plus, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$+", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  u=X1; 
  NDEREF_F(Arg, u, 1, t1);
  if (TagIsSmall(t) && TagIsSmall(u))
    {
      if (TagIsSmall(t1 = t+(u-TaggedZero)))
	return t1;
      else
	return make_integer_check(Arg, GetSmall(t1),p);
    }
  else if (IsFloat(t) || IsFloat(u))
    return make_float_check(Arg, GetFloat(t) + GetFloat(u),p);
  else
    return bn_call(Arg,bn_add,t,u,p);
}

CFUN__PROTO_N(fu2_minus, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$-", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  u=X1; 
  NDEREF_F(Arg, u, 1, t1);
  if (TagIsSmall(t) && TagIsSmall(u))
    {
      if (TagIsSmall(t1 = t-(u-TaggedZero)))
	return t1;
      else
	return make_integer_check(Arg, GetSmall(t1),p);
    }
  else if (IsFloat(t) || IsFloat(u))
    return make_float_check(Arg, GetFloat(t) - GetFloat(u),p);
  else
    return bn_call(Arg,bn_subtract,t,u,p);
}

CFUN__PROTO_N(fu2_times, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$*", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  u=X1; 
  NDEREF_F(Arg, u, 1, t1);
  if (TagIsSmall(t) && TagIsSmall(u))
    {
      int st = GetSmall(t);
      int su = (int)(u-TaggedZero);
      int stu;
      tagged_t tu;
      SMUL_OVERFLOW(st, su, stu, { goto overflow; });
      tu = ((tagged_t)stu)+TaggedZero;
      if (TagIsSmall(tu)) return tu;
overflow:
      {}
    }
  if (IsFloat(t) || IsFloat(u))
    return make_float_check(Arg, GetFloat(t) * GetFloat(u),p);
  else
    return bn_call(Arg,bn_multiply,t,u,p);
}

CFUN__PROTO_N(fu2_fdivide, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$/", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  u=X1; 
  NDEREF_F(Arg, u, 1, t1);
  return make_float_check(Arg, GetFloat(t)/GetFloat(u),p);
}

CFUN__PROTO_N(fu2_idivide, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$//", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);
  
  if (u == TaggedZero) BUILTIN_ERROR(EVALUATION_ERROR(ZERO_DIVISOR), u, 1);
  
  if (TagIsSmall(t) && TagIsSmall(u))
    return make_integer_check(Arg, (int)(t-TaggedZero)/(int)(u-TaggedZero),p);

  /*bn_quotient_wanted = TRUE;*/
  return bn_call(Arg,bn_quotient_remainder_quot_wanted,t,u,p);
}

CFUN__PROTO_N(fu2_rem, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$rem", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);

  if (u == TaggedZero) BUILTIN_ERROR(EVALUATION_ERROR(ZERO_DIVISOR), u, 1);

  if (TagIsSmall(t) && TagIsSmall(u))
    return (int)(t-TaggedZero)%(int)(u-TaggedZero)+TaggedZero;

  /*bn_quotient_wanted = FALSE;*/
  return bn_call(Arg,bn_quotient_remainder_quot_not_wanted,t,u,p);
}

CFUN__PROTO_N(fu2_mod, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$mod", 3);
  int rem, denom;
  tagged_t T_rem;

  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);

  if (u == TaggedZero) BUILTIN_ERROR(EVALUATION_ERROR(ZERO_DIVISOR), u, 1);

  if (TagIsSmall(t) && TagIsSmall(u)) {
    denom = (int)(u-TaggedZero);
    rem = (int)(t-TaggedZero)%denom;
    return ( (denom > 0 && rem < 0) || (denom < 0 && rem > 0) ?
              rem+denom : rem ) + TaggedZero;
  } else {
    /*bn_quotient_wanted = FALSE;*/
    T_rem = bn_call(Arg,bn_quotient_remainder_quot_not_wanted,t,u,p);
    return ( T_rem != TaggedZero &&
             fu1_sign(Arg,u,p) != fu1_sign(Arg,T_rem,p)
             ? bn_call(Arg,bn_add,T_rem,u,p) : T_rem );
  }
}

CFUN__PROTO_N(fu1_abs, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$abs", 2);
  flt64_t f;
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t)) {
      if (t==TaggedLow)
	return make_integer_check(Arg, GetSmall(TaggedHigh),p);
      else if (t < TaggedZero)
	return TaggedZero-(t-TaggedZero);
      else
        return t;
  } else if (IsFloat(t))
    return (((f = GetFloat(t)) < 0.0) ? make_float_check(Arg, -f, p) : t);
  else 
    return ((!bn_positive(TagToSTR(t))) ? bn_call(Arg,bn_minus,t,0,p) : t);
}

CFUN__PROTO_N(fu1_sign, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$sign", 2);
  flt64_t f;
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t))
    return ((t==TaggedZero) ? TaggedZero :
            (t < TaggedZero) ? TaggedZero-(1<<SmallShift) :
            TaggedZero+(1<<SmallShift));
  else if (IsFloat(t)) {
    f = GetFloat(t);
    return ((f == 0.0) ? t :
            (f < 0.0) ? make_float_check(Arg, -1.0, p) :
            make_float_check(Arg, 1.0, p));
  } else 
    return ((!bn_positive(TagToSTR(t))) ? TaggedZero-(1<<SmallShift) :
            TaggedZero+(1<<SmallShift));
}

CFUN__PROTO_N(fu1_not, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$\\", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t))
    return t^(QMask-(1<<SmallShift));
  else
    return bn_call(Arg,bn_not,t,0,p);
}

CFUN__PROTO_N(fu2_xor, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$#", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);
  if (TagIsSmall(t) && TagIsSmall(u))
    return t^u^TaggedZero;
  else
    return bn_call(Arg,bn_xor,t,u,p);
}

CFUN__PROTO_N(fu2_and, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$/\\", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);
  if (TagIsSmall(t) && TagIsSmall(u))
    return ((t^ZMask)&(u^ZMask))^ZMask;
  else
    return bn_call(Arg,bn_and,t,u,p);
}

CFUN__PROTO_N(fu2_or, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$\\/", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);
  if (TagIsSmall(t) && TagIsSmall(u))
    return ((t^ZMask)|(u^ZMask))^ZMask;
  else
    return bn_call(Arg,bn_or,t,u,p);
}

static CFUN__PROTO_N(lsh_internal, tagged_t, tagged_t t, int dist, bcp_t p)
{
  tagged_t u;
  
  if (TagIsSmall(t))
    {
      switch (dist)
	{
	case 0:
	  return t;
	case 1:
	  u = (t<<1) + 0x78000000;
	  break;
	case 2:
	  u = (t<<2) + 0x68000000;
	  break;
	case 3:
	  u = (t<<3) + 0x48000000;
	  break;
	case 4:
	  u = (t<<4) + 0x08000000;
	  break;
	default:
	  u = 0;
	}
      if (TagIsSmall(u))
	return u;
      /*
      int value = GetSmall(t);

      if (dist<32 &&
          value>=0 && value < (unsigned long)(1<<31)>>dist ||
	  value<0 && value >= (long)(-1<<31)>>dist)
	return make_integer_check(value<<dist,p);
	*/
    }

  return bn_call(Arg, bn_lshift, t, MakeInteger(Arg,dist), p);
}

static CFUN__PROTO_N(rsh_internal, tagged_t, tagged_t t, int dist, bcp_t p)
{
  if (TagIsSmall(t))
    {
      if (dist>=WORDSIZE)
	return MakeSmall((t>=TaggedZero)-1);
      else
	return ((int)((t>>dist)-(TaggedZero>>dist)) & -4) + TaggedZero;
    }

  return bn_call(Arg, bn_rshift, t, MakeInteger(Arg,dist), p);
}

/* 
   For shiffting operations we assume it is not possible to allocate
   enough memory to represent any number bigger than
   2^(2^INTMACH_MAX).  Therefore shifting to the left any number by
   more than 2^INTMACH_MAX throws a memory overflow exception and
   shifting to the right any number by more than 2^INTMACH_MAX returns
   0.
*/
CFUN__PROTO_N(fu2_lsh, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$<<", 3);
  int dist;
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);
 
  if (IsIntegerFix(u)) {
    /* 2^INTMACH_MIN =< u =< 2^INTMACH_MAX */
    dist = GetInteger(u);
    
    if (dist == INTMACH_MIN) {
      /* u = 2^INTMACH_MIN, i.e. -u > 2^INTMACH_MAX */
      return TaggedZero;
    } else {
      /* abs(u) =< 2^INTMACH_MIN */
      return (dist<0 ? rsh_internal(Arg,t,-dist,p) : lsh_internal(Arg,t,dist,p));
    }
  } else if (bn_positive(TagToSTR(u)) && t != TaggedZero) {
    /* u > 2^INTMACH_MAX */
    BUILTIN_ERROR(RESOURCE_ERROR(R_STACK), u, 2); 
  } else {
    /* -u > 2^INTMACH_MAX */
    return TaggedZero;
  }  
}

CFUN__PROTO_N(fu2_rsh, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$>>", 3);
  int dist;
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_I(Arg, t, 0, t1);
  u=X1; 
  NDEREF_I(Arg, u, 1, t1);
  
  if (IsIntegerFix(u)) { 
    /* 2^INTMACH_MIN =< u =< 2^INTMACH_MAX */
    dist = GetInteger(u);
    
    if (dist == INTMACH_MIN) {
      /* u = 2^INTMACH_MIN, i.e. -u > 2^INTMACH_MAX */
      BUILTIN_ERROR(RESOURCE_ERROR(R_STACK), u, 2); 
    } else {
      /* abs(u) =< 2^INTMACH_MIN */
      return (dist<0 ? lsh_internal(Arg,t,-dist,p) : rsh_internal(Arg,t,dist,p));
    }
  } else if (bn_positive(TagToSTR(u)) || t == TaggedZero) {
    /* u > 2^INTMACH_MAX */
    return TaggedZero;
  } else {
    /* -u > 2^INTMACH_MAX */
    BUILTIN_ERROR(RESOURCE_ERROR(R_STACK), u, 2); 
  }
}


/*  GCD for rat arithm., ch feb 92
    
    This works through the following mechanism:
		
    - The arithm. functions (via is/2,</2,...) apply NDEREF to their
      args which calls evaluate [wamsupport.c] which, when it sees
      terms, calls the corresponding function ( max arity = 2) Eval
      calls this functions with p = NULL which means that numstack_end
      must not be NULL (make_float_check, make_integer_check)
*/

CFUN__PROTO_N(fu2_gcd, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$gcd", 3);
  CIAO_REG_1(tagged_t, u);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, v);

  CIAO_REGISTER int type = 3;			/* big x big */ 
  
  u=X0; 
  NDEREF_I(Arg, u, 0, t1);
  if (TagIsSmall(u)) {
    type -= 2;
    if (u<=TaggedZero)
      u = (u==TaggedLow ? make_integer_check(Arg, GetSmall(TaggedHigh),p)
	                : TaggedZero-(u-TaggedZero));
  } 
  else if (!bn_positive(TagToSTR(u)))
    u = bn_call(Arg,bn_minus,u,0,p);

  v=X1; 
  NDEREF_I(Arg, v, 1, t1);
  if (TagIsSmall(v)) {
    type -= 1;
    if (v<=TaggedZero)
      v = (v==TaggedLow ? make_integer_check(Arg, GetSmall(TaggedHigh),p)
	                : TaggedZero-(v-TaggedZero));
  } 
  else if (!bn_positive(TagToSTR(v)))
    v = bn_call(Arg,bn_minus,v,0,p);
                                
  if ( u==TaggedZero ) return v;
  if ( v==TaggedZero ) return u;
  /*bn_quotient_wanted = FALSE;*/

  for (;;) {
    switch (type) {  			/*     u x v     */

      case 0:				/* small x small */
  small_x_small:
      	{ CIAO_REGISTER unsigned long int x = GetSmall(u), y = GetSmall(v);
      	  for (;;) {
	    x = x % y; if ( x==0 ) return MakeSmall(y);
	    y = y % x; if ( y==0 ) return MakeSmall(x);
      	  }
	}

      case 1:				/* small x big   */
	v = bn_call(Arg,bn_quotient_remainder_quot_not_wanted,v,u,p);
	if ( v==TaggedZero ) return u;
	goto small_x_small;

      case 2:                           /*   big x small */
 	u = bn_call(Arg,bn_quotient_remainder_quot_not_wanted,u,v,p);
	if ( u==TaggedZero ) return v;
	goto small_x_small;

      case 3:				/*   big x big   */
	u = bn_call(Arg,bn_quotient_remainder_quot_not_wanted,u,v,p); 
	if ( u==TaggedZero ) return v;
	if ( TagIsSmall(u) ) type -= 2;
	v = bn_call(Arg,bn_quotient_remainder_quot_not_wanted,v,u,p); 
	if ( v==TaggedZero ) return u;
	if ( TagIsSmall(v) ) type -= 1;
    }
  }
} 




#include <math.h>

#if defined(Solaris) || defined(LINUX) || defined(DARWIN) || defined(Win32) || defined(IRIX) || defined(BSD)
# define aint(f) (f>=0.0 ? floor(f) : ceil(f))
#endif

CFUN__PROTO_N(fu1_intpart, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$float_integer_part", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  flt64_t f;

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  f = GetFloat(t);
  return make_float_check(Arg, aint(f),p);
}

CFUN__PROTO_N(fu1_fractpart, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$float_fractional_part", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  flt64_t f;

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  f = GetFloat(t);
  return make_float_check(Arg, f-aint(f),p);
}

CFUN__PROTO_N(fu1_floor, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$floor", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  tagged_t f;

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t) || (p==NULL && !IsFloat(t)))
    return t;
  else if (IsFloat(t)) {
    if (float_is_finite(t)) {
      f = make_float_check(Arg, floor(GetFloat(t)),p);
      return bn_call(Arg,bn_from_float,f,0,p);
    }
    else 
      BUILTIN_ERROR(REPRESENTATION_ERROR(NAN_OR_INF_TO_INTEGER), t, 1);
  } else
    return bn_call(Arg,bn_plus,t,0,p);
}

CFUN__PROTO_N(fu1_round, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$round", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  tagged_t f;

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t) || (p==NULL && !IsFloat(t)))
    return t;
  else if (IsFloat(t)) {
    if (float_is_finite(t)) {
      f = make_float_check(Arg, floor(GetFloat(t)+0.5),p);
      return bn_call(Arg,bn_from_float,f,0,p);
    }
    else 
      BUILTIN_ERROR(REPRESENTATION_ERROR(NAN_OR_INF_TO_INTEGER), t, 1);
  } else
    return bn_call(Arg,bn_plus,t,0,p);
}

CFUN__PROTO_N(fu1_ceil, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$ceiling", 2);
  CIAO_REGISTER tagged_t t,t1;

  tagged_t f;

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  if (TagIsSmall(t) || (p==NULL && !IsFloat(t)))
    return t;
  else if (IsFloat(t)) {
    if (float_is_finite(t)) {
      f = make_float_check(Arg, ceil(GetFloat(t)),p);
      return bn_call(Arg,bn_from_float,f,0,p);
    }
    else 
      BUILTIN_ERROR(REPRESENTATION_ERROR(NAN_OR_INF_TO_INTEGER), t, 1);
  } else
    return bn_call(Arg,bn_plus,t,0,p);
}

CFUN__PROTO_N(fu2_pow, tagged_t, tagged_t X0, tagged_t X1, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$**", 3);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);
  CIAO_REG_3(tagged_t, u);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);
  u=X1; 
  NDEREF_F(Arg, u, 1, t1);
  return make_float_check(Arg, pow(GetFloat(t),GetFloat(u)),p);
}

CFUN__PROTO_N(fu1_exp, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$exp", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  return make_float_check(Arg, exp(GetFloat(t)),p);
}

CFUN__PROTO_N(fu1_log, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$log", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  return make_float_check(Arg, log(GetFloat(t)),p);
}

CFUN__PROTO_N(fu1_sqrt, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$sqrt", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);


  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  return make_float_check(Arg, sqrt(GetFloat(t)),p);
}

CFUN__PROTO_N(fu1_sin, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$sin", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);


  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  return make_float_check(Arg, sin(GetFloat(t)),p);
}

CFUN__PROTO_N(fu1_cos, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$cos", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  return make_float_check(Arg, cos(GetFloat(t)),p);
}

CFUN__PROTO_N(fu1_atan, tagged_t, tagged_t X0, bcp_t p)
{
  ERR__FUNCTOR("arithmetic:$atan", 2);
  CIAO_REG_1(tagged_t, t);
  CIAO_REG_2(tagged_t, t1);

  t=X0; 
  NDEREF_F(Arg, t, 0, t1);

  return make_float_check(Arg, atan(GetFloat(t)),p);
}

/* ---------------------------------------------------------------- */

/*
CFUN__PROTO_N(fu2_arg, tagged_t,
              CIAO_REGISTER tagged_t number,
	      CIAO_REGISTER tagged_t complex,
	      bcp_t p)
{  
  ERR__FUNCTOR("term_basic:arg", 3);
  CIAO_REG_1(tagged_t, t0);
  CIAO_REGISTER int i;

  DerefSwitch(number,t0, BUILTIN_ERROR(INSTANTIATION_ERROR, number, 1););
  DerefSwitch(complex,t0, BUILTIN_ERROR(INSTANTIATION_ERROR, complex, 2););

  if (TagIsSmall(number)) i = GetSmall(number);
  else if (TagIsLarge(number) && !LargeIsFloat(number)) return FALSE;
  else BUILTIN_ERROR(TYPE_ERROR(INTEGER), number, 1);

  if (i < 0)
    BUILTIN_ERROR(DOMAIN_ERROR(NOT_LESS_THAN_ZERO), number, 1);

  if (TagIsSTR(complex))
    {
      CIAO_REGISTER tagged_t f = TagToHeadfunctor(complex);

      if (i == 0 || i > Arity(f) || f&QMask) return FALSE;

      RefArg(t0,complex,i);
      return t0;
    }
  else if (IsComplex(complex))	/\* i.e. list *\/
    {
      if (i == 1)
	{
	  RefCar(t0,complex);
	  return t0;
	}
      else if (i == 2)
	{
	  RefCdr(t0,complex);
	  return t0;
	}
      else return FALSE;
    }
  // comment next line for full ISO compliance
  else if (IsAtom(complex)) return FALSE;      
  else  BUILTIN_ERROR(TYPE_ERROR(COMPOUND), complex, 2);
 

}
*/

// Old code without exception 
CFUN__PROTO_N(fu2_arg, tagged_t,
	      CIAO_REGISTER tagged_t number,
	      CIAO_REGISTER tagged_t complex,
	      bcp_t p)
{
  CIAO_REG_1(tagged_t, t0);
  
  DerefSwitch(number,t0,{goto barf1;});
  DerefSwitch(complex,t0,{goto barf2;});


  if (TagIsSTR(complex))
    {
      CIAO_REGISTER int i = GetSmall(number);
      CIAO_REGISTER tagged_t f = TagToHeadfunctor(complex);

      if (i<=0 || i>Arity(f) || f&QMask)
	goto barf1;
      
      RefArg(t0,complex,i);
      return t0;
    }
  else if (IsComplex(complex))	// i.e. list 
    {
      if (number==MakeSmall(1))
	{
	  RefCar(t0,complex);
	  return t0;
	}
      else if (number==MakeSmall(2))
	{
	  RefCdr(t0,complex);
	  return t0;
	}
      else
	goto barf1;
    }
  else
    goto barf2;

 barf1:
  MINOR_FAULT("arg/3: incorrect 1st argument");

 barf2:
  MINOR_FAULT("arg/3: incorrect 2nd argument");
}

