#!/bin/sh
# Grep on Ciao sources
#
# Author: Jose F. Morales
#
#   This command performs a grep on Ciao source files (avoid .svn and
#   .git files and automatically generated files like .po, .itf, files
#   under build/).
#
# Usage:
#   grep-source.sh [-d DIR] REGEXP
#
# TODO: 
#   - Nothing is cached or precomputed (so search is really slow)
#   - Search is not semantic nor syntactic.
# ---------------------------------------------------------------------------

# Physical directory where the script is located
_base=$(e=$0;while test -L "$e";do d=$(dirname "$e");e=$(readlink "$e");\
        cd "$d";done;cd "$(dirname "$e")";pwd -P)

RELBUILDDIR=build
BOOTRELBUILDDIR=build-boot

# Exit immediately if a simple command exits with a non-zero status
set -e

if [ x"$1" == x"-d" ]; then
    shift
    DIR=$1
    shift
else
    DIR=${_base}/../..
fi

cd "${DIR}"
find . -name '.svn' -prune -o \
       -name '.git' -prune -o \
       -path './'"${RELBUILDDIR}" -prune -o \
       -path './regr-db' -prune -o \
       -path './'"${BOOTRELBUILDDIR}" -prune -o \
       \( '!' -type d \
          '!' -name '*~' -a \
          '!' -name '*.html*' -a \
          '!' -name '*.ps' -a \
          '!' -name '*.po' -a \
          '!' -name '*.itf' -a \
          '!' -name '*.asr' -a \
          '!' -name '*.ast' -a \
          '!' -name '*.so' -a \
          '!' -name '*.texi*' -a \
          '!' -name '*.ascii' -a \
          '!' -name '*.info' \) \
       -exec grep -I -nH -e "$*" {} \; 

