#!/bin/sh
# ===========================================================================
# clean_aux.sh: auxiliary functions to clean results of compilation 
#
# Author: Jose F. Morales (based on previous code)
#
# Note: This code is implemented in SH as a safeguard against
#       undesirable accidents. If this code were implemented in
#       Prolog, then it would not be possible to recover from some
#       unstable states (e.g., if the engine or libraries failed to
#       compile and a recompilation from a clean state is necessary).
# ===========================================================================
#
# TODO:
#
#  - It should be possible to generate part of this file automatically
#    (the structure is really simple).
#
#  - The list of suffixes should be given by the system (the previous
#    point would help).
#
#  - CIAOCACHEDIR may simplify some code
#
#  - Cleaning may clean things we do not want to clean... (e.g.,
#    *_auto.pl?)
#
# ---------------------------------------------------------------------------

# Delete one file
do_delete_file() { # FILE
    test x"${1}" = x"" && return
    /bin/rm -f "${1}"
}

# Delete a directory and its contents recursively
do_remove_dir() { # DIR
    test x"${1}" = x"" && return
    /bin/rm -r -f "${1}"
}

# TODO: Remove! Used only for cleaning of 'ilciao'
# TODO: Write in Prolog or use build.xml
do_delete_example_pl_files() { # DIR
    test x"${1}" = x"" && return
    /bin/rm -f "${1}/examples.*.pl"
}

# Clean the compilation output for just one module
do_clean_mod() { # MOD
    test x"${1}" = x"" && return
    /bin/rm -f \
           "${1}.asr" \
           "${1}.ast" \
           "${1}.itf" \
           "${1}.po" \
	   "${1}_${CIAOARCH}.o" \
	   "${1}_${CIAOARCH}.so" \
           "${1}_${CIAOARCH}.dll" \
	   "${1}_${CIAOARCH}.dylib" \
           "${1}_${CIAOARCH}_glue.c" \
           "${1}_${CIAOARCH}_glue.o" \
	   "${1}_${CIAOARCH}_inline.c"
}

# Clean the compilation output for all files, recursively
# TODO: split in backups, ciaoc output, etc.
do_clean_dir_rec() {
    # NOTE: this code has been optimized (usually it should take less
    #   than 10 seconds)
    #
    # We cannot use -print0 | xargs -0 because -print0 is not compatible
    # with Solaris, nor -print | xargs because it is buggy
    test x"${1}" = x"" && return
    find "${1}"/. -name ".svn" -prune -o \( \
	-name "*.po" \
	-o -name "*.itf" \
	-o -name "*.cpx" \
	-o -name "*.wam" \
	-o -name "*.asr" \
	-o -name "*.ast" \
	-o -name "*.o" \
	-o -name "*.so" \
	-o -name "*.dll" \
	-o -name "*.dylib" \
	-o -name "*_glue.c" \
	-o -name "*_inline.c" \
	-o -name "*_auto.pl" \
	-o -name "*.ascii" \
	-o -name "*.class" \
	\
	-o -name "*.aux" \
	-o -name "*.log" \
	-o -name "*.err" \
	-o -name "tmpciao*" \
	\
	-o -name "*_co.pl" \
	-o -name "*_co.java" \
	-o -name "*.iss" \
	\
	-o -name "*~" \
	-o -name "#*" \) \
	-exec /bin/rm -f {} \;
}

