# ---------------------------------------------------------------------------
# Extract bootstrap options (before a working Prolog engine is
# compiled). This includes parameters such as the C compiler, linker,
# compile and link options, and options for ciaoc.sta for compilation
# of bootbundletool.
#
# Author: Jose F. Morales (based on previous code)
# ---------------------------------------------------------------------------

# Input: arguments and options
# Output: configuration variables for builddir_aux.sh:builddir_configure_boot
#   CUSTOM_CC
#   CUSTOM_LD
#   EXTRA_CFLAGS
#   EXTRA_LDFLAGS
#   CIAOCOPTS
scan_bootstrap_opts() {
    ciao__CUSTOM_CC=""
    ciao__CUSTOM_LD=""
    ciao__EXTRA_CFLAGS=""
    ciao__EXTRA_LDFLAGS=""
    CIAOCOPTS=""
    #
    if [ ! $# = 0 ] ; then
	for ARG in "$@" ; do
	    if expr "${ARG}" : '\(--[^=][^=]*=..*\)' >/dev/null  ; then
		OPT=`expr "${ARG}" : '--\([^=][^=]*\)=..*'|sed -e s:_:-:g`
		VAL=`expr "${ARG}" : '--[^=][^=]*=\(..*\)'`
                if [ x"${OPT}" = x"ciao:custom-cc" ] ; then
		    ciao__CUSTOM_CC="${VAL}"
                elif [ x"${OPT}" = x"ciao:custom-ld" ] ; then
		    ciao__CUSTOM_LD="${VAL}"
                elif [ x"${OPT}" = x"ciao:extra-cflags" ] ; then
		    ciao__EXTRA_CFLAGS="${VAL}"
		elif [ x"${OPT}" = x"ciao:extra-ldflags" ] ; then
		    ciao__EXTRA_LDFLAGS="${VAL}"
		elif [ x"${OPT}" = x"ciao:unused-pred-warnings" ] ; then
		    # (option for ciaoc)
		    test x"${VAL}" = x"yes" && add_ciaoc_opt "--unused-pred-warnings"
		fi
	    fi
	done
    fi
}

add_ciaoc_opt() {
    CIAOCOPTS="${CIAOCOPTS} ${1}"
}

# TODO: 
#   - We need an option to enable system-wide rtchecks, etc.  The
#     current way to enable rtchecks is with this configuration
#     option:
#
#	--ciaosh-commands="\"set_prolog_flag(runtime_checks,yes)\""
#
#     However, it is just a hack that enables the flag in the toplevel.
# TODO: Synchronize with definitions at the Prolog side


