:- module(bundleconfig_tr, [sent/3], [assertions]).

:- data hook_decl/2.

sent(end_of_file, Clauses, M) :-
	get_multifile_preds(Clauses, M).
sent((:- bundle_param(Name, Opts)), Clauses, Mod) :- !,
	get_bundle_name(Mod, M),
	Clauses = [('$bundleconfig_entry'(Name, M, Opts))].
sent(Sent, [], _Mod) :- \+ allowed_sent(Sent), !,
 	message(warning, ['Invalid sentence for bundle configuration: ', Sent]).

allowed_sent((:- Decl)) :- nonvar(Decl), !,
	allowed_decl(Decl).
allowed_sent((?- _)) :- fail.
allowed_sent(_). % TODO: not all rules should be allowed

allowed_decl(doc(_,_)).
allowed_decl(discontiguous(_)).
allowed_decl(multifile(_)).
allowed_decl(include(M)) :- allowed_import(M).
allowed_decl(use_module(M)) :- allowed_import(M).
allowed_decl(use_module(M,_)) :- allowed_import(M).

% TODO: be more precise
allowed_import(library(_)).

get_bundle_name(M1, M) :-
	( atom_concat('bundle_config__', M, M1) -> true
	; M = M1
	).

get_multifile_preds(Clauses, M1) :-
	get_bundle_name(M1, M),
	Clauses = [
          (m_bundle_config_call(M,G) :- call(G)),
	  (m_bundle_config_entry(M,Name,Props) :- '$bundleconfig_entry'(Name,M,Props))
          |Clauses1],
	Clauses1 = [end_of_file].

