:- module(_, [], [dcg, fsyntax, hiord, assertions, regtypes, isomodes]).

:- doc(title,  "Bundle Actions Aux").
:- doc(author, "Jose F. Morales").

:- use_module(library(bundletool/paths_extra), [fsR/2, bundle_metasrc/3]).
:- use_module(library(system), [cd/1, working_directory/2]).
:- use_module(library(glob), [glob/3]).

:- use_module(library(bundletool/config_common), [
    instype/1,
    installdirbase_lib/2,
    with_docs/1
]).
:- use_module(library(bundletool/bundle_aux), [
	ensure_builddir/0,
        ensure_build_doc_dirs/0,
        rl_install_copy_and_link/2,
        rl_install_link_as/2,
        rl_uninstall_link/1,
        rl_uninstall_copy_and_link/1,
        r_install_as/2,
        r_abslink_as/2,
        b_install_dir/1,
        b_install_dir_rec/2,
        b_installsrc_dir_rec/2,
        b_install_copy_and_link/3,
        b_install_file_noexec/2,
        b_install_link_as/4,
        b_uninstall_link/2,
        b_uninstall_copy_and_link/3,
        b_uninstall_file/1,
        b_uninstall_dir_rec/1,
        b_uninstallsrc_dir_rec/1,
        b_uninstall_dir/1,
        b_uninstall_dir_if_empty/1,
        install_standalone_list/2,
        uninstall_standalone_list/2
	]).
:- use_module(library(bundletool/ciaoc_aux), [
        build_standalone_list/3
	]).

% ============================================================================

:- include(library(bundletool/bundle_actions_hookdefs)).

% ============================================================================
% Load bundleconfig and bundlehook definitions

% TODO: move to other module?

% TODO: add reference counting?
% TODO: reused for bundleconfig too, change name?
:- use_module(library(bundletool/bundlehooks_holder)).

:- export(load_bundle_metasrc/2).
load_bundle_metasrc(Bundle, Metasrc) :-
	( '$bundle_id'(Bundle), BundleDir = ~fsR(bundle_src(Bundle)) ->
	    true
	; throw(unregistered_or_unknown_bundle(Bundle))
	),
	% TODO: make it optional (some bundles may not need customized hooks)
	bundle_metasrc(Bundle, Metasrc, BundleMetasrcFile0),
	fsR(BundleMetasrcFile0, BundleMetasrcFile),
	( file_exists(BundleMetasrcFile) ->
	    working_directory(PWD, BundleDir), % TODO: Needed here?
	    ( catch(bundlehooks_holder:do_use_module(BundleMetasrcFile), 
	            E, R = exception(E)) ->
	        R = ok
	    ; R = fail
	    ),
	    cd(PWD),
	    ( R = exception(E) -> throw(E)
	    ; R = fail -> fail
	    ; true
	    )
	; optional_metasrc(Metasrc) ->
	    true
	; % TODO: write handler?
	  throw(no_bundle_metasrc(Bundle, Metasrc))
	).

% TODO: use optional_metasrc/1 (do not unload if it was optional)
:- export(unload_bundle_metasrc/2).
unload_bundle_metasrc(Bundle, Metasrc) :-
	bundle_metasrc(Bundle, Metasrc, BundleMetasrcFile0),
	fsR(BundleMetasrcFile0, BundleMetasrcFile),
	bundlehooks_holder:do_unload(BundleMetasrcFile).

optional_metasrc(bundle_config).
	
% ============================================================================
% Bundle configuration rules

:- data bundleconfig_rules_loaded/0.

:- export(ensure_load_bundleconfig_rules/0).
% TODO: make it fine-grained (one bundle at a time)
ensure_load_bundleconfig_rules :-
	bundleconfig_rules_loaded, !.
ensure_load_bundleconfig_rules :-
	assertz_fact(bundleconfig_rules_loaded),
	( % (failure-driven loop)
	  '$bundle_id'(Bundle),
	    load_bundle_metasrc(Bundle, bundle_config),
	    fail
	; true
	).

% ============================================================================
% Operations on bundle and sub-bundles

% TODO: Simplify!

:- use_module(engine(internals), ['$bundle_id'/1, '$bundle_prop'/2]).
:- use_module(library(bundle/bundle_info), [
	enum_sub_bundles/2,
	enumrev_sub_bundles/2
	]).

:- use_module(library(terms), [atom_concat/2]).

:- use_module(library(bundletool/messages_aux),
	[bold_message/2, normal_message/2]).

:- export(bundle_do_hookcmd/2).
% Invoke a command on Bundle that needs bundle hooks. Bundle hooks for
% the bundle are loaded automatically. The bundle Manifest must be
% already registered.
bundle_do_hookcmd(Bundle, Target) :-
	ensure_load_bundleconfig_rules, % TODO: make it more fine grained (not always needed, not all bundles needed)
	%
%	display(user_error, 'LOADINGMETASRC'(Bundle, Target)), nl(user_error),
	load_bundle_metasrc(Bundle, bundle_hooks),
	%
	% TODO: is cd/1 really necessary here?
	BundleDir = ~fsR(bundle_src(Bundle)),
	working_directory(PWD, BundleDir),
	catch(bundlehook_call(Bundle, Target), E, OK = no(E)),
	( var(OK) -> OK = yes ; true ),
	cd(PWD),
	%
%	display(user_error, 'UNLOADINGMETASRC'(Bundle, Target)), nl(user_error),
	unload_bundle_metasrc(Bundle, bundle_hooks),
	%
	( OK = no(E) ->
	    throw(bundle_do_failed(Bundle, Target, E))
	; true
	).

bundlehook_call(Bundle, Target) :-
	ll_bundlehook_decl(Target),
	!,
	( ll_bundlehook_do(Target, Bundle) ->
	    true
	; throw(bundlehook_failed)
	).
bundlehook_call(Bundle, Target) :-
	BundleHooksMod = ~atom_concat(bundle_actions__, Bundle),
	( m_bundlehook_decl(BundleHooksMod, Target) ->
	    ( m_bundlehook_do(BundleHooksMod, Target) ->
	        true
	    ; throw(bundlehook_failed)
	    )
	; throw(bundlehook_undefined)
	).

:- use_module(library(bundletool/bundle_configure),
	[config_list_options/1, config_describe_option/1]).

ll_bundlehook_decl(build_nolibs).
ll_bundlehook_decl(build_platdep).
ll_bundlehook_decl(build_libraries).
ll_bundlehook_decl(prebuild_libraries). % TODO: used rule?
ll_bundlehook_decl(install).
ll_bundlehook_decl(uninstall).
ll_bundlehook_decl(build_nodocs).
ll_bundlehook_decl(build_docs).
ll_bundlehook_decl(build_docs_readmes).
ll_bundlehook_decl(build_docs_manuals).
ll_bundlehook_decl(install_docs).
ll_bundlehook_decl(uninstall_docs).
ll_bundlehook_decl(create_tags).
ll_bundlehook_decl(delete_tags).
ll_bundlehook_decl(config_list_options).
ll_bundlehook_decl(config_describe_option).
ll_bundlehook_decl(runtests).
ll_bundlehook_decl(runbenchmarks).
ll_bundlehook_decl(calibrateciaopp).
ll_bundlehook_decl(save_calibrateciaopp).
ll_bundlehook_decl(gen_timingmodel).

ll_bundlehook_do(install, Bundle) :- !,
	bundle_lldo(Bundle, install),
	bundle_lldo(Bundle, bundle_activate).
ll_bundlehook_do(uninstall, Bundle) :- !,
	bundle_lldo(Bundle, bundle_deactivate),
	bundle_lldo(Bundle, uninstall).
%
ll_bundlehook_do(config_list_options, Bundle) :- !,
	config_list_options(Bundle).
ll_bundlehook_do(config_describe_option, Bundle) :- !,
	% (parameter in DESCRIBE_OPTION)
	config_describe_option(Bundle).
%
ll_bundlehook_do(Target, Bundle) :-
	bundle_lldo(Bundle, Target).

:- export(bundle_lldo/2).
% (like bundle_do, do not load BundleHooks)
% TODO: this should be unified with the previous code
bundle_lldo(Bundle, Target) :-
	bundle_lldo_(Target, Bundle).

bundle_lldo_(prebuild, Bundle) :- bundle_prebuild_hook(Bundle).
bundle_lldo_(prebuild_libraries, Bundle) :- bundle_prebuild_libraries_hook(Bundle).
bundle_lldo_(build_nodocs, Bundle) :-
	% TODO: see big_group
	bundle_dname(Bundle, BundleDesc),
	bold_message("Building ~w (no docs)", [BundleDesc]),
	bundle_build_nodocs_hook(Bundle),
	bold_message("~w build (no docs) completed", [BundleDesc]).
bundle_lldo_(build_nolibs, Bundle) :-
	bundle_build_nolibs_hook(Bundle).
bundle_lldo_(build_platdep, Bundle) :-
	bundle_build_platdep_hook(Bundle).
bundle_lldo_(build_libraries, Bundle) :- bundle_build_libraries_hook(Bundle).
bundle_lldo_(build_docs, Bundle) :-
	( with_docs(yes) ->
	    % TODO: see big_group
	    bundle_dname(Bundle, BundleDesc),
	    bold_message("Building (docs) ~w", [BundleDesc]),
	    %
	    ensure_builddir, % TODO: needed?
	    ensure_build_doc_dirs, % TODO: needed?
	    bundle_prepare_docs_hook(Bundle),
	    bundle_docs_readmes(Bundle),
	    bundle_docs_manuals(Bundle),
	    sub_bundles_build_docs_hook(Bundle),
	    %
	    bold_message("~w build (docs) completed", [BundleDesc])
	; true
	).
bundle_lldo_(build_docs_readmes, Bundle) :-
	( with_docs(yes) ->
	    ensure_builddir, % TODO: needed?
	    ensure_build_doc_dirs, % TODO: needed?
	    bundle_docs_readmes(Bundle)
	; true
	).
bundle_lldo_(build_docs_manuals, Bundle) :-
	( with_docs(yes) ->
	    ensure_builddir, % TODO: needed?
	    ensure_build_doc_dirs, % TODO: needed?
	    bundle_docs_manuals(Bundle)
	; true
	).
bundle_lldo_(install, Bundle) :-
	bundle_install_hook(Bundle),
	rec_bundles_do(Bundle, all, install).
bundle_lldo_(uninstall, Bundle) :-
	revrec_bundles_do(Bundle, all, uninstall),
	bundle_uninstall_hook(Bundle).
bundle_lldo_(install_docs, Bundle) :- % (no hooks)
	install_item(Bundle, docs(Bundle)),
	rec_bundles_do(Bundle, all, install_docs).
bundle_lldo_(uninstall_docs, Bundle) :- % (no hooks)
	revrec_bundles_do(Bundle, all, uninstall_docs),
	uninstall_item(Bundle, docs(Bundle)).
bundle_lldo_(register, Bundle) :- bundle_register_hook(Bundle).
bundle_lldo_(unregister, Bundle) :- bundle_unregister_hook(Bundle).
bundle_lldo_(bundle_activate, Bundle) :-
	install_bundlereg(Bundle),
	bundle_lldo(Bundle, register).
bundle_lldo_(bundle_deactivate, Bundle) :-
	bundle_lldo(Bundle, unregister),
	uninstall_bundlereg(Bundle).
% Creation of TAGS for use with the find-tag command (ESC-.) in emacs
bundle_lldo_(create_tags, Bundle) :-
        create_tags_file(Bundle).
% Deletion of TAGS file
bundle_lldo_(delete_tags, Bundle) :-
	delete_tags_file(Bundle).
% Run tests
bundle_lldo_(runtests, Bundle) :-
	bundle_custom_hook(Bundle, runtests).
% Run benchmarks
bundle_lldo_(runbenchmarks, Bundle) :-
	bundle_custom_hook(Bundle, runbenchmarks).
% (custom)
bundle_lldo_(calibrateciaopp, Bundle) :-
	bundle_custom_hook(Bundle, calibrateciaopp).
% (custom)
bundle_lldo_(gen_timingmodel, Bundle) :-
	bundle_custom_hook(Bundle, gen_timingmodel).

sub_bundles_build_docs_hook(Bundle) :- rec_bundles_do(Bundle, all, build_docs).

:- export(rec_bundles_do/3).
% Invoke bundle_do on sub-bundles of ParentBundle
rec_bundles_do(ParentBundle, extra, Target) :- !, % (extra sub-bundles)
	( % (failure-driven loop)
	  enum_sub_bundles(ParentBundle, P),
          '$bundle_prop'(P, kind(extra)), % TODO: Generalize filter
	  bundle_do_hookcmd(P, Target),
	    fail
	; true
	).
rec_bundles_do(ParentBundle, all, Target) :- !, % (all sub-bundles)
	( % (failure-driven loop)
	  enum_sub_bundles(ParentBundle, P),
	    bundle_do_hookcmd(P, Target),
	    fail
	; true
	).

:- export(revrec_bundles_do/3).
% Invoke bundle_do on sub-bundles of ParentBundle (in reverse order)
revrec_bundles_do(ParentBundle, all, Target) :- !, % (all sub-bundles)
	( % (failure-driven loop)
	  enumrev_sub_bundles(ParentBundle, P),
	    bundle_do_hookcmd(P, Target),
	    fail
	; true
	).

% ============================================================================
% TODO: include 'groups'
% TODO: add dependencies

action_revlist(uninstall). % this action uses reverse order 

:- export(bundletree/1).
:- regtype bundletree(X) # "@var{X} is the symbolic description of the
   contents of a bundle".
% bundletree is the input of the generic install/uninstall operations
bundletree(_).

is_list([]).
is_list([_|_]). % wrong... but faster

:- export(prebuild_item/2).
:- pred prebuild_item(+Bundle, +X) :: atm * bundletree # "Pre-build bundle-item @var{X}".
prebuild_item(Bundle, X) :- bundleitem_do(X, Bundle, prebuild).

:- export(build_item/2).
:- pred build_item(+Bundle, +X) :: atm * bundletree # "Build bundle-item @var{X}".
build_item(Bundle, X) :- bundleitem_do(X, Bundle, build_nodocs).

:- export(install_item/2).
:- pred install_item(+Bundle, +X) :: atm * bundletree # "Install bundle-item @var{X}".
install_item(Bundle, X) :- bundleitem_do(X, Bundle, install).

:- export(uninstall_item/2).
:- pred uninstall_item(+Bundle, +X) :: atm * bundletree # "Uninstall bundle-item @var{X}".
uninstall_item(Bundle, X) :- bundleitem_do(X, Bundle, uninstall).

% ----

action_message(prebuild, "Prebuilding", "prebuild").
action_message(install, "Installing", "installation").
action_message(uninstall, "Uninstalling", "uninstallation").
action_message(build_nodocs, "Building (no docs)", "build (no docs)").

bundleitem_do(X, Bundle, Action) :- is_list(X), !,
	( action_revlist(Action) ->
	    bundleitem_revlist_do(X, Bundle, Action)
	; bundleitem_list_do(X, Bundle, Action)
	).
bundleitem_do(X, Bundle, Action) :- bundleitem_def_decl_hook(X), !,
	bundleitem_def_hook(X, Y), bundleitem_do(Y, Bundle, Action).
bundleitem_do(X, _Bundle, prebuild) :- bundleitem_prebuild_hook_decl(X), !, bundleitem_prebuild_hook(X).
bundleitem_do(X, _Bundle, build_nodocs) :- bundleitem_build_hook_decl(X), !, bundleitem_build_hook(X).
bundleitem_do(X, _Bundle, install) :- bundleitem_install_hook_decl(X), !, bundleitem_install_hook(X).
bundleitem_do(X, _Bundle, uninstall) :- bundleitem_uninstall_hook_decl(X), !, bundleitem_uninstall_hook(X).
bundleitem_do(small_group(Name, X), Bundle, Action) :- !,
	action_message(Action, ActionMsg0, _ActionMsg1),
	normal_message("~s ~s", [ActionMsg0, Name]),
	bundleitem_do(X, Bundle, Action).
bundleitem_do(group(Name, X), Bundle, Action) :- !,
	action_message(Action, ActionMsg0, _ActionMsg1),
	bold_message("~s ~s", [ActionMsg0, Name]),
	bundleitem_do(X, Bundle, Action).
bundleitem_do(big_group(Name, X), Bundle, Action) :- !,
	action_message(Action, ActionMsg0, ActionMsg1),
	bold_message("~s ~s", [ActionMsg0, Name]),
	bundleitem_do(X, Bundle, Action),
	bold_message("~s ~s completed", [Name, ActionMsg1]).
%
bundleitem_do(only_global_ins(X), Bundle, Action) :- !,
	% consider X only if installation kind is global
	( ~instype = global ->
	    bundleitem_do(X, Bundle, Action)
	; true
	).
bundleitem_do(dir(Path), Bundle, Action) :- !, bundleitem_do(dir(Path, []), Bundle, Action).
bundleitem_do(dir(Path, Props), _Bundle, install) :- !,
	b_install_dir(~fsR(Path)), % perms?
	( member(files_from(SrcDir), Props) ->
	    % copy contents from Dir
	    b_install_dir_rec(SrcDir, ~fsR(Path))
	; member(files_from(SrcDir, Pattern), Props) ->
	    % copy contents from Dir (as specified by Pattern)
	    ( % (failure-driven loop)
		member(File, ~glob(~fsR(SrcDir), Pattern)),
		b_install_file_noexec(~fsR(SrcDir/(File)),
	                              ~fsR(Path/(File))),
	        fail
	    ; true
	    )
	; true
	).
bundleitem_do(dir(Path, Props), _Bundle, uninstall) :- !,
	( member(del_rec, Props) ->
	    % on uninstall, remove contents recursively
	    % TODO: remove also the directory?
	    b_uninstall_dir_rec(~fsR(Path)) % TODO: use bundle
	; member(del_if_empty, Props) ->
	    % delete if the directory is empty
	    % TODO: only if it is empty? what dirs should have this prop?
%	    b_uninstall_dir_if_empty(~installdirbase_lib(ciao)) fsR(path)??
	    b_uninstall_dir_if_empty(~installdirbase_lib(ciao)) % TODO: use bundle
	; member(do_not_del, Props) ->
	    % do not delete on uninstall
	    true
	; b_uninstall_dir(~fsR(Path))
	).
bundleitem_do(lib(Bundle, DirName), _Bundle, install) :- !, % (only instype=global) % TODO: use _Bundle
	% Install the module collection under DirName (along compiled files)
	bold_message("Installing (~w) '~w' libraries", [Bundle, DirName]),
	( DirName = '.' ->
	    % TODO: simplify
	    b_install_dir_rec(bundle_src(Bundle), ~fsR(installdir_lib(Bundle)))
	; b_install_dir_rec(DirName, ~fsR(installdir_lib(Bundle)/(DirName)))
	).
bundleitem_do(lib(Bundle, DirName), _Bundle, uninstall) :- !, % (only instype=global)
	% Uninstall the previously installed module collection DirName
	( DirName = '.' ->
	    % TODO: simplify
	    b_uninstall_dir_rec(~fsR(installdir_lib(Bundle)))
	; b_uninstall_dir_rec(~fsR(installdir_lib(Bundle)/(DirName)))
	).
bundleitem_do(src(Bundle, DirName), _Bundle, install) :- !, % (only instype=global) % TODO: use _Bundle
	% Install the module collection under DirName (just source, e.g., for examples)
	bold_message("Installing (~w) '~w' source", [Bundle, DirName]),
	b_installsrc_dir_rec(DirName, ~fsR(installdir_lib(Bundle)/(DirName))).
bundleitem_do(src(Bundle, DirName), _Bundle, uninstall) :- !, % (only instype=global)
	% Uninstall the previously installed source-only module collection DirName
	b_uninstallsrc_dir_rec(~fsR(installdir_lib(Bundle)/(DirName))).
bundleitem_do(docs(Bundle), _Bundle, install) :- !, % (for instype=global and instype=local) % TODO: use _Bundle
	bundle_install_docs(Bundle).
bundleitem_do(docs(Bundle), _Bundle, uninstall) :- !, % (for instype=global and instype=local)
	bundle_uninstall_docs(Bundle).
bundleitem_do(standalone_list(Bundle, Path, List), _Bundle, build_nodocs) :- !, % TODO: use _Bundle
	build_standalone_list(Bundle, Path, List).
bundleitem_do(standalone_list(Bundle, _, List), _Bundle, install) :- !, % TODO: use _Bundle
	install_standalone_list(Bundle, List).
bundleitem_do(standalone_list(Bundle, _, List), _Bundle, uninstall) :- !,
 	uninstall_standalone_list(Bundle, List).
bundleitem_do(lib_file_list(Path, List), _Bundle, Action) :- !, % TODO: use _Bundle
	lib_file_list_do(List, Path, Action).
bundleitem_do(bin_copy_and_link(K, Bundle, File, Props), _Bundle, install) :- !, % TODO: use _Bundle
	b_install_copy_and_link(K, Bundle, File),
	( member(link_as(Link), Props) ->
	    b_install_link_as(K, Bundle, File, Link)
	; true
	).
bundleitem_do(bin_copy_and_link(K, Bundle, File, Props), _Bundle, uninstall) :- !,
	( member(link_as(Link), Props) ->
	    b_uninstall_link(K, Link)
	; true
	),
	b_uninstall_copy_and_link(K, Bundle, File).
bundleitem_do(file(Path), _Bundle, uninstall) :- !,
	b_uninstall_file(Path).

% TODO: we would need dependencies here
% (uninstall is done in reverse order)
bundleitem_revlist_do(Xs, Bundle, Action) :-
	bundleitem_list_do(~reverse(Xs), Bundle, Action).

bundleitem_list_do([], _Bundle, _Action).
bundleitem_list_do([X|Xs], Bundle, Action) :- bundleitem_do(X, Bundle, Action), bundleitem_list_do(Xs, Bundle, Action).

lib_file_list_do([], _Path, _Action).
lib_file_list_do([X|Xs], Path, Action) :-
	lib_file_item_do(X, Path, Action),
	lib_file_list_do(Xs, Path, Action).

lib_file_item_do(File-Props, Path, install) :- !,
	( member(copy_and_link, Props) ->
	    rl_install_copy_and_link(Path, File)
	; true
	),
	( member(link_as(Link), Props) ->
	    rl_install_link_as(File, Link)
	; true
	),
	( member(mid(Mid), Props) -> % TODO: why?
	    % TODO: incompatible with any other prop
	    r_install_as(Path/File, Mid),
	    r_abslink_as(Mid, ~installdirbase_lib(ciao)/(File)) % TODO: use bundle
	; true
	).
lib_file_item_do(File-Props, _Path, uninstall) :- !,
	( member(link_as(Link), Props) ->
	    rl_uninstall_link(Link)
	; true
	),
	( member(copy_and_link, Props) ->
	    rl_uninstall_copy_and_link(File)
	; true
	),
	( member(mid(Mid), Props) -> % TODO: why?
	    % TODO: incompatible with any other prop
	    b_uninstall_file(Mid),
	    rl_uninstall_copy_and_link(File)
	; true
	).

:- use_module(library(lists), [reverse/2]).

% ---------------------------------------------------------------------------
% Bundle registry installation (after bundle scan!)
% TODO: rename those predicates

:- use_module(library(bundle/bundle_scan),
	[create_bundlereg/2, remove_bundlereg/2]).

install_bundlereg(Bundle) :-
	( ~instype = global ->
	    BundleDir = ~fsR(bundle_src(Bundle)),
	    create_bundlereg(BundleDir, global)
	; true % (done in bundle scan)
	).

uninstall_bundlereg(Bundle) :-
	( ~instype = global ->
	    remove_bundlereg(Bundle, global)
	; true % (nothing needed for local installations)
	).

% ===========================================================================

%:- doc(section, "Generation of TAGS (for find-tag emacs command)").
% TODO: It should not generate tags for the 'whole' bundle

% TODO: Make sure that it works

:- use_module(library(bundletool/bundle_aux), [
        create_tags_file/1,
        delete_tags_file/1]).

% ===========================================================================

:- doc(subsection, "Documentation of a Bundle").

:- use_module(library(bundletool/config_common),
	[ % (lpdoc bundle)
          docformat/1, docformatdir/2]).
:- use_module(library(bundle/bundle_info), [bundle_manual_base/2]).

:- use_module(library(pathnames), [path_concat/3, path_split/3]).
:- use_module(library(system), [file_exists/1]).
:- use_module(library(system_extra), [mkpath/1]).
:- use_module(library(source_tree),
	[copy_file_or_dir/2, remove_file_or_dir/1]).

:- use_module(library(aggregates), [findall/3]).

:- use_module(library(bundletool/ciaoc_aux),
	[invoke_lpdoc/2, invoke_lpdoc/1]).

% Creation of README files (from .lpdoc to ascii)
% Output is moved to the bundle root directory.
bundle_docs_readmes(Bundle) :- with_docs(yes), !,
	( % (failure-driven loop)
	  bundle_readme_path(Bundle, Readme),
	    ( Readme = as(SrcPath, OutName) ->
	        true
	    ; SrcPath = Readme,
	      OutName = Name
	    ),
	    path_split(SrcPath, _, Name),
	    BundleDir = ~fsR(bundle_src(Bundle)),
	    path_concat(BundleDir, SrcPath, SrcPath2),
	    path_split(SrcPath2, FilePath, _),
	    Ascii = ~atom_concat(Name, '.ascii'),
	    invoke_lpdoc(['-d', ~atom_concat('filepath=', FilePath),
	                  '-d', 'doc_mainopts=no_versioned_output',
	                  '-c', Ascii]),
	    DocSrc = ~fsR(builddir_doc(build)/Ascii),
	    insert_autogen_warning(DocSrc, OutName),
	    fail
	; true
	).
bundle_docs_readmes(_Bundle).

:- use_module(library(lists), [append/3]).
:- use_module(library(file_utils), [file_to_string/2, string_to_file/2]).

insert_autogen_warning(Src, Target) :-
	file_to_string(Src, SrcStr),
	append(~warning_autogen, SrcStr, TargetStr),
	string_to_file(TargetStr, Target).

%               ...........................................................................
warning_autogen("[Please, do not edit. File automatically formatted by LPdoc]\n\n").

% Creates the manuals
bundle_docs_manuals(Bundle) :- with_docs(yes), !,
	( % (failure-driven loop)
	  bundle_manual_dir(Bundle, Manual),
	    ( Manual = as(SrcDir, _) ->
	        true
	    ; SrcDir = Manual
	    ),
	    BundleDir = ~fsR(bundle_src(Bundle)),
	    path_concat(BundleDir, SrcDir, R0),
	    path_concat(R0, 'SETTINGS', Settings),
	    invoke_lpdoc(Settings, [all]),
	    fail
	; true
	).
bundle_docs_manuals(_Bundle).

bundle_install_docs(Bundle) :- with_docs(yes), !,
	bold_message("Installing documentation for '~w'", [Bundle]),
	( % (failure-driven loop)
	  docformat(DocFormat),
	    docformatdir(DocFormat, TargetDir),
	    FileName = ~atom_concat([~bundle_manual_base(Bundle), '.', DocFormat]),
	    Source = ~fsR(builddir_doc(build)/(FileName)),
	    ( file_exists(Source) ->
		Target = ~atom_concat(TargetDir, FileName),
		( Source == Target ->
		    % TODO: Wrong note message (we are not really installing then!)
		    note(['Skipping copy of ', Target])
		; % mkpath(TargetDir), % TODO: perms and owner?
		  copy_file_or_dir(Source, TargetDir)
		),
		bundle_install_docs_format_hook(DocFormat, Target)
	    ; warning(['File ', Source, ' not generated yet. Skipping copy'])
	    ),
	    fail
	; true
	).
bundle_install_docs(_Bundle).

% TODO: skip it if documentation was not installed
bundle_uninstall_docs(Bundle) :- with_docs(yes), !,
	bold_message("Uninstalling documentation for '~w'", [Bundle]),
	( % (failure-driven loop)
	  docformat(DocFormat),
	    docformatdir(DocFormat, TargetDir),
	    FileName = ~atom_concat([~bundle_manual_base(Bundle), '.', DocFormat]),
	    Target = ~atom_concat(TargetDir, FileName),
	    ( bundle_uninstall_docs_format_hook(DocFormat, Target),
	      remove_file_or_dir(Target) ->
	        true
	    ; warning(['Could not uninstall documentation in ', 
	               DocFormat, ' format for ',
		       Bundle, ' (was it generated?)'])
	    ),
	    fail
	; true
	).
bundle_uninstall_docs(_Bundle).

% These predicates install the 'info' files in info dir.

:- use_module(library(bundletool/info_installer)).

bundle_install_docs_format_hook(info, Target) :- !,
	dirfile_install_info(~fsR(builddir_doc(build)), Target).
bundle_install_docs_format_hook(_, _).

bundle_uninstall_docs_format_hook(info, Target) :- !,
	dirfile_uninstall_info(~fsR(builddir_doc(build)), Target).
bundle_uninstall_docs_format_hook(_, _).


