% ============================================================================
% Multifile definitions for bundles
% TODO: Most of this should be part of the bundle registry

% this_bundle_id(Id) # "@var{Id} is the identifier of this bundle".

% bundle_dname(Id, X) # "@var{X} is the textual name of the bundle".
:- discontiguous bundle_dname/2.
:- multifile bundle_dname/2.
% bundle_readme_path(Id, D) # "@var{D} is the (bundle relative) path
%   for a README file (in lpdoc format)".
:- discontiguous bundle_readme_path/2.
:- multifile bundle_readme_path/2.
% bundle_manual_dir(Id, D) # "@var{D} is the (bundle relative)
%   subdirectory where each manual is found (containing a SETTINGS.pl
%   file)".
:- discontiguous bundle_manual_dir/2.
:- multifile bundle_manual_dir/2.

% ============================================================================
% Hooks for operations on @regtype{bundletree/1}
%
% OP_hook_delc(X): target @var{X} accepts operation @tt{OP}
% OP_hook(X,...): implementation of operation @tt{OP} for target @var{X}
%
% where OP is one of:
%
%   - 'def': definition as a named node in a bundletree (incompatible
%            with the rest of OP)
%   - 'prebuild': preparation before build
%   - 'build_nodocs': build (no docs)
%   - 'install': installation
%   - 'uninstall': uninstallation

:- multifile m_bundlehook_decl/2.
:- multifile m_bundlehook_do/2.

% TODO: (new) check again
:- multifile bundle_prebuild_hook/1.
:- discontiguous bundle_prebuild_hook/1.
:- multifile bundle_prebuild_libraries_hook/1.
:- discontiguous bundle_prebuild_libraries_hook/1.
:- multifile bundle_build_nodocs_hook/1.
:- discontiguous bundle_build_nodocs_hook/1.
:- multifile bundle_build_nolibs_hook/1.
:- discontiguous bundle_build_nolibs_hook/1.
:- multifile bundle_build_platdep_hook/1.
:- discontiguous bundle_build_platdep_hook/1.
:- multifile bundle_build_libraries_hook/1.
:- discontiguous bundle_build_libraries_hook/1.

:- discontiguous bundleitem_prebuild_hook_decl/1.
:- multifile bundleitem_prebuild_hook_decl/1.
:- discontiguous bundleitem_prebuild_hook/1.
:- multifile bundleitem_prebuild_hook/1.

:- discontiguous bundleitem_build_hook_decl/1.
:- multifile bundleitem_build_hook_decl/1.
:- discontiguous bundleitem_build_hook/1.
:- multifile bundleitem_build_hook/1.
%bundleitem_build_hook_decl(_) :- fail.
%bundleitem_build_hook(_) :- fail.

% hooks for bundle (un)installation
:- discontiguous bundle_install_hook/1.
:- multifile bundle_install_hook/1.
:- discontiguous bundle_uninstall_hook/1.
:- multifile bundle_uninstall_hook/1.

:- discontiguous bundle_prepare_docs_hook/1.
:- multifile bundle_prepare_docs_hook/1.

% hooks for bundle (un)registry
:- discontiguous bundle_register_hook/1.
:- multifile bundle_register_hook/1.
:- discontiguous bundle_unregister_hook/1.
:- multifile bundle_unregister_hook/1.

% hooks for tests, benchmarks, and other actions
:- multifile bundle_custom_hook/2.
:- discontiguous bundle_custom_hook/2.

% hooks for bundleitem definition
:- discontiguous bundleitem_def_decl_hook/1.
:- multifile bundleitem_def_decl_hook/1.
:- discontiguous bundleitem_def_hook/2. % bundleitem_def_hook(Node, Def)
:- multifile bundleitem_def_hook/2.

% hooks for bundleitem (un)installation
:- discontiguous bundleitem_install_hook_decl/1.
:- multifile bundleitem_install_hook_decl/1.
:- discontiguous bundleitem_install_hook/1.
:- multifile bundleitem_install_hook/1.
:- discontiguous bundleitem_uninstall_hook_decl/1.
:- multifile bundleitem_uninstall_hook_decl/1.
:- discontiguous bundleitem_uninstall_hook/1.
:- multifile bundleitem_uninstall_hook/1.

% hooks for bundleitem (un)registry
:- discontiguous bundleitem_register_hook/1.
:- multifile bundleitem_register_hook/1.
:- discontiguous bundleitem_unregister_hook/1.
:- multifile bundleitem_unregister_hook/1.

% hooks for third-party components
:- discontiguous m_third_party_name/1.
:- multifile m_third_party_name/1.
:- discontiguous m_third_party_version/2.
:- multifile m_third_party_version/2.
:- discontiguous m_third_party_source_url/2.
:- multifile m_third_party_source_url/2.
:- discontiguous m_third_party_source_md5/2.
:- multifile m_third_party_source_md5/2.
:- discontiguous m_third_party_build_system/2.
:- multifile m_third_party_build_system/2.
:- discontiguous m_third_party_option1/2.
:- multifile m_third_party_option1/2.
:- discontiguous m_third_party_option2/3.
:- multifile m_third_party_option2/3.
