:- module(_, [], [assertions, fsyntax]).

:- doc(title,  "Emacs batch calls").

:- doc(author, "Jos@'{e} F. Morales (redesign, new code)").
:- doc(author, "Ciao Development Team").

:- use_module(library(process), [process_call/3]).
:- use_module(library(bundletool/paths_extra), [fsR/2]).

% ---------------------------------------------------------------------------

% TODO: bundle_param_value/3 should not be used here
:- use_module(library(bundletool/bundle_params),
	[root_bundle_name/1,
	 bundle_param_value/3]).
:- use_module(library(system), [winpath/2]).

:- use_module(library(bundle/bundle_config), [bundle_config_value/3]).

:- export(emacs_type/1).
emacs_type(EmacsType) :-
	% Note: set from win32/CiaoDE.iss.skel or
	%       environment_and_windows_bats action
	bundle_param_value(ciao, 'EMACS_TYPE', EmacsType),
	!.
emacs_type(posix).

:- export(emacs_path/1).
emacs_path(Emacs) :-
	% Note: set from win32/CiaoDE.iss.skel or
	%       environment_and_windows_bats action
	root_bundle_name(Root),
	bundle_param_value(Root, 'EMACS_PATH', Emacs0),
	winpath(Emacs, Emacs0),
	!.
emacs_path(Emacs) :-
	% find_emacs(Emacs),
	emacs_for_ciao(Emacs),
	!.
emacs_path(XEmacs) :-
	% find_xemacs(XEmacs).
	xemacs_for_ciao(XEmacs).

emacs_for_ciao := ~bundle_param_or_config_value(ciao, 'EMACS_FOR_CIAO').
xemacs_for_ciao := ~bundle_param_or_config_value(ciao, 'XEMACS_FOR_CIAO').

bundle_param_or_config_value(Bundle, Name, Value) :-
	bundle_param_value(Bundle, Name, Value),
	!.
bundle_param_or_config_value(Bundle, Name, Value) :-
	bundle_config_value(Bundle, Name, Value).

% ---------------------------------------------------------------------------

:- use_module(library(system), [cyg2win/3]).

:- export(emacsstylepath/3).
emacsstylepath('Win32', Dir, Path) :- !,
	atom_codes(Dir, SDir),
	cyg2win(SDir, Path, noswap).
emacsstylepath(_, Dir, Path) :- !,
	atom_codes(Dir, Path).

% ---------------------------------------------------------------------------

% TODO: Use 'append' creation mode for logs? (use a single log)
:- export(emacs_batch_call/3).
emacs_batch_call(Dir, LogName, Args) :-
	% Environment variables unset for this call
	NoEnv = ['SHELL', 'EMACSLOADPATH', 'EMACSDOC'],
	%
	Out = ~fsR(concat_k(ext('.log'), Dir/LogName)),
	Err = ~fsR(concat_k(ext('.err'), Dir/LogName)),
	%
	process_call(~emacs_path, ['-batch'|Args],
	    [cwd(~fsR(Dir)),
	     noenv(NoEnv),
	     stdout(file(Out)),
	     stderr(file(Err)),
	     show_error_on_error, show_output_on_error,
	     verbose_error, status(_)]).

% ---------------------------------------------------------------------------

:- use_module(library(system_extra), [del_file_nofail/1]).

:- export(clean_log/2).
% Clean log files
clean_log(Dir, LogName) :-
	del_file_nofail(~fsR(concat_k(ext('.log'), Dir/LogName))),
	del_file_nofail(~fsR(concat_k(ext('.err'), Dir/LogName))).

% ---------------------------------------------------------------------------

:- use_module(library(terms), [atom_concat/2]).

:- export(emacs_update_autoloads/3).
:- pred emacs_update_autoloads(Dir, Log, AutoloadEL)
   # "Invoke the Emacs @tt{'batch-update-autoloads'} function to generate
      the autoload file @var{AutoloadEL}.".
 
emacs_update_autoloads(Dir, Log, AutoloadEL) :-
	emacs_type(EmacsType),
	AutoloadEL2S = ~emacsstylepath(EmacsType, AutoloadEL),
	% TODO: espape AutoloadEL2
	atom_codes(AutoloadEL2, AutoloadEL2S),
	%
	emacs_batch_call(Dir, Log, 
          ['--eval', ~atom_concat(['(setq generated-autoload-file "', AutoloadEL2, '")']),
	   '-f', 'batch-update-autoloads', '.']).

:- export(emacs_batch_byte_compile/3).
:- pred emacs_batch_byte_compile(Dir, Log, EL_Files)
   # "Invoke the Emacs '@tt{'batch-byte-compile'} function to byte compile 
      the specified @var{EL_Files} elisp files.".

emacs_batch_byte_compile(Dir, Log, EL_Files) :-
	emacs_batch_call(Dir, Log,
	  ['--eval', '(setq load-path (cons "." load-path))',
	   '-f', 'batch-byte-compile'|
           EL_Files]).

