:- module(messages_aux, [
	bold_message/1,
	bold_message/2,
	normal_message/2,
	bolder_message/1,
	bolder_message/2
	], [assertions, regtypes, hiord]).

:- doc(title, "Messages for the build process").

:- use_module(library(format), [format/3]).
:- use_module(library(lists), [append/3]).
:- use_module(library(messages)).

bold_message(Mess) :-
	bold_message(Mess, []).

bold_message(Mess, Args) :-
	prefix_lines(Mess, ":: ", NMess),
	format(user_output, NMess, Args),
	format(user_output, "~n",  []).

normal_message(Mess, Args) :-
	prefix_lines(Mess, "   ", NMess),
	format(user_output, NMess, Args),
	format(user_output, "~n",  []).

bolder_message(Mess) :-
	bolder_message(Mess, []).

bolder_message(Mess, Args) :-
	prefix_lines(Mess, "== ", NMess),
	format(user_output, NMess, Args),
	format(user_output, "~n",  []).

% Add @var{Prefix} to each line in @var{String0}, put result in @var{String}
prefix_lines(String0, Prefix, String) :-
	append(Prefix, String1, String),
	prefix_lines_(String0, Prefix, String1).

prefix_lines_([],       _,      []).
prefix_lines_([0'\n|R], Prefix, NR) :- !,
	NR = [0'\n|NR1],
	append(Prefix, NR0, NR1),
	prefix_lines_(R, Prefix, NR0).
prefix_lines_([C|R], Prefix, [C|NR]) :- !,
	prefix_lines_(R, Prefix, NR).

