:- module(pbundle_gen_common, [], [assertions, fsyntax]).

:- doc(title,  "Common Definitions for Installer Generation").
:- doc(author, "Edison Mera").
:- doc(author, "Jose F. Morales").
:- doc(module, "This file is part of the CiaoDE installation system.").

:- use_module(library(aggregates)).
:- use_module(library(file_utils)).
:- use_module(library(lists)).
:- use_module(library(system)).
:- use_module(library(system_extra)).
:- use_module(library(terms), [atom_concat/2]).
:- use_module(library(pathnames), [path_get_relative/3]).

:- use_module(library(bundle/bundle_info), [
	enum_bundle/2,
	bundle_versioned_packname/2]).
:- use_module(library(bundletool/paths_extra), [fsR/2]).

:- use_module(library(bundletool/config_common), [perms/1]).

:- use_module(library(bundletool/ciaoc_aux), [enum_platdep_noarch_file/1]).
:- use_module(library(bundletool/pbundle_generator)).
:- use_module(library(bundletool/messages_aux),
	[bolder_message/1, bold_message/2]).

% (hooks for gen_pbundle)
:- include(library(bundletool/pbundle_gen_hookdefs)).

% ===========================================================================

% The directory where bundletool files are placed
:- export(bundletool_dir/1).
bundletool_dir := bundle_src(ciao)/'library'/'bundletool'.

% ===========================================================================

:- export(pbundle_output_dir/1).
% TODO: The definition of directory is repeated in ciaobot/SHARED
%       (PBUNDLE_BUILD_DIR). Share the definition.
pbundle_output_dir := ~fsR(builddir(build)/'pbundle').

:- export(create_pbundle_output_dir/0).
:- pred create_pbundle_output_dir # "Make sure that the directory
   where generated pbundle files are placed exists and has the
   NODISTRIBUTE mark.".

create_pbundle_output_dir :-
	TargetDir = ~pbundle_output_dir,
	mkpath(TargetDir, ~perms), % TODO: owner?
	% TODO: use 'directory mark' preds
	string_to_file("", ~fsR(TargetDir/'NODISTRIBUTE')).

% TODO: rename by pbundle_create_tarball?
:- export(gen_pbundle_common/3).
gen_pbundle_common(Bundle, PackType, Descs) :-
	SourceDir = ~atom_concat(~fsR(bundle_src(Bundle)), '/'),
	create_pbundle_output_dir,
	TargetDir = ~pbundle_output_dir,
	VersionedPackName = ~bundle_versioned_packname(Bundle),
	exclude_files(PackType, ExcludeFiles),
	bold_message("Creating ~w tarball for ~w ~w, please be patient ...",
	    [PackType, VersionedPackName, Descs]),
	TargetDir2 = ~atom_concat(TargetDir, '/'), % TODO: remove '/'
	build_pbundle_codeitem_contents(PackType, SourceDir, TargetDir2,
	    VersionedPackName, ExcludeFiles, FileList),
	list(Descs, build_pbundle_codeitem(SourceDir, TargetDir2,
		VersionedPackName, PackType, FileList)).

exclude_files(PackType) := ~findall(File, exclude_file(PackType, File)).

% Note: CIAOBUILDDIR directory is not fully excluded
% TODO: obtain this list from other place?
:- export(exclude_file/2).
exclude_file(PrecompLevel) := F :-
	F = ~relciaodir(~exclude_file_(PrecompLevel)).

exclude_file_(_) := ~enum_config_file.
exclude_file_(noa) := ~enum_platdep_noarch_file.

% Configuration files
enum_config_file := ~fsR(builddir(build)/'CONFIG_input'). % TODO: necessary?
enum_config_file := ~fsR(builddir(build)/'CONFIG_mkf').
enum_config_file := ~fsR(builddir(build)/'CONFIG_sh').
enum_config_file := ~fsR(builddir(build)/'CONFIG_saved').

:- export(relciaodir/2).
relciaodir(S) := Dir :-
	R = ~fsR(bundle_src(ciaode)),
	path_get_relative(R, S, Dir).

:- export(gen_pbundle_descfile/1).
% Generate the desc.tmpl file (see pbundle_meta)
gen_pbundle_descfile(Bundle) :-
	gen_pbundle_hook(descfile, Bundle, []).

% (hook)
% TODO: a real target?
gen_pbundle_hook(descfile, Bundle, _Options) :- !,
	TargetDir = ~pbundle_output_dir,
	mkpath(TargetDir, ~perms), % TODO: owner?
	DescFile = ~fsR(TargetDir/'desc.tmpl'),
	pbundle_generate_meta(Bundle, DescFile).

