;;; flymake-ciao.el --- A flymake handler for ciao-mode files
;;
;;; Author: Manuel Hermenegildo <manuel@hermenegildo.org>
;;; URL: https://.../flymake-ciao
;;; Version: DEV
;;; Package-Requires: ((flymake-easy "0.1"))
;;;
;;; Commentary:
;; Usage:
;; ciaoc ciao-check-syntax.pl; make sure in path
;; (require 'flymake-ciao)
;; (add-hook 'ciao-mode-hook 'flymake-ciao-load)
;;
;; Uses flymake-easy, from https://github.com/purcell/flymake-easy
;; 
;; Just a quick hack to test flymake for now!
;; Current bugs:
;; - Only begin line marked in source file (very imprecise for ciao)
;; - Does not understand when there are several errors reported for
;;   one path, i.e., only finds first error in:
;; {Reading /Users/herme/clip/Systems/ciao-devel/emacs-mode/flymake/foo.pl
;; ERROR: (lns 6-7) syntax error: operator expected after expression
;; ...
;; ERROR: (lns 18-19) syntax error: operator expected after expression
;; ...
;; }
;; 
;;; Code:
(require 'flymake-easy)
;; (REGEXP FILE-IDX LINE-IDX COL-IDX ERR-TEXT-IDX).
(defvar flymake-ciao-err-line-patterns
  '(("^{Reading \\(.*\\)\n\\(ERROR\\|WARNING\\|NOTE\\): (lns \\([0-9]*\\)-\\([0-9]*\\)) \\([^}]*\n\\)*}" 1 3 80 5)))
;; 1=file, 2=error type, 3=begin line, 4=end line, 5=error text 
;; Alt:
;; (defvar flymake-ciao-err-line-patterns
;;   '(("^{Reading \\(.*\\)\n\\(ERROR\\|WARNING\\|NOTE\\): (lns \\([0-9]*-[0-9]*\\)) \\([^}]*\n\\)*}" 1 3 nil 4)))
;; 1=file, 2=error type, 3=line numbers, 5=error text 


;; Should really be ciaoc/ciaopp (with appropriate options). Special
;; executable for now:
(defvar flymake-ciao-executable "/Users/herme/GNU/lisp/flymake-easy/ciao-check-syntax"
  "The ciao executable to use for syntax checking.")
(defun flymake-ciao-command (filename)
  "Construct a command that flymake can use to check ciao source."
  ;; (list flymake-ciao-executable "-w" "-c" filename))
  (list flymake-ciao-executable filename))
;;;###autoload
(defun flymake-ciao-load ()
"Configure flymake mode to check the current buffer's ciao syntax."
(interactive)
(flymake-easy-load 
 'flymake-ciao-command
 flymake-ciao-err-line-patterns
 'tempdir
 "pl"))
(provide 'flymake-ciao)
;;; flymake-ciao.el ends here

