:- module(_, _, [bundlehooks]).

:- doc(title,  "Bundle Hooks for LPdoc").
:- doc(author, "Jose F. Morales").
:- doc(author, "Edison Mera (first version)").

% ===========================================================================
% Bundle description

this_bundle_id(lpdoc).
bundle_dname(lpdoc, 'LPdoc').
bundle_manual_dir(lpdoc) := as('doc', 'lpdoc').
bundle_readme_path(lpdoc) := as('readmes/INSTALLATION_LPDOC', 'INSTALLATION').
bundle_readme_path(lpdoc) := as('readmes/README_LPDOC', 'README').

% ============================================================================

% (options used in this bundle)
:- use_module(library(bundletool/config_common), [
    instype/1,
    % (lpdoc bundle)
    docdir/1
]).

:- use_module(library(bundletool/paths_extra), [fsR/2]).

% ===========================================================================
% TODO: Refine and prune

:- use_module(library(system), [copy_file/3]).
:- use_module(library(system_extra), ['-'/1, '--'/1]).

:- use_module(library(bundletool/bundle_aux), [
        b_install_dir/1,
        b_install_copy_and_link/3,
        b_uninstall_copy_and_link/3,
        wr_template/4,
	generate_version_auto/2
	]).

:- use_module(library(bundletool/ciaoc_aux), [
	% TODO: use bundle defs instead
        invoke_ciaosh_batch/1]).

% ============================================================================
% COMPILATION
% ============================================================================

% (hook)
bundle_build_nodocs_hook(Bundle) :- Bundle = lpdoc, !,
	bundle_lldo(Bundle, prebuild_libraries),
	bundle_bin.

% (invoked from 'bundle_actions__*', for each bundle)
bundle_build_nolibs_hook(Bundle) :- Bundle = lpdoc, !,
	bundle_lldo(Bundle, prebuild_libraries),
	bundle_bin.

% (invoked from 'bundle_actions__*', for each bundle)
bundle_build_platdep_hook(lpdoc) :-
	bundle_bin.

% (invoked from 'bundle_actions__*', for each bundle)
bundle_build_libraries_hook(lpdoc).

% Prepare the code
bundle_prebuild_libraries_hook(lpdoc) :-
	generate_version_auto_lpdoc.

%% ---------------------------------------------------------------------------

% TODO: generalize for all bundles; change modiftime only if there are changes
%       move, include other options (for runtime)?
generate_version_auto_lpdoc :-
	Bundle = lpdoc,
	File = ~fsR(bundle_src(Bundle)/'src'/'version_auto.pl'),
	% normal_message("Creating 'version_auto.pl' for ~w", [Bundle]),
	generate_version_auto(Bundle, File).

bundleitem_def_decl_hook(lpdoccl).
bundleitem_def_hook(lpdoccl) :=
    standalone_list(lpdoc, bundle_src(lpdoc)/'src', [
        'lpdoc'-[
          name="command line",
	  plexe,
	  final_ciaoc
	]
    ]).

bundle_bin :-
        build_item(lpdoc, lpdoccl).

% ============================================================================
:- doc(section, "Documentation").
% ============================================================================

bundle_prepare_docs_hook(lpdoc).

% =============================================================================
% REGISTER
% =============================================================================

bundle_register_hook(lpdoc).
bundle_unregister_hook(lpdoc).

% =============================================================================
% INSTALLATION                                                              
% =============================================================================

bundle_install_hook(lpdoc) :-
	bundle_install_(~instype).

bundle_install_(local) :-
	Bundle = lpdoc,
	bold_message("Skipping copy of LPdoc files."),
	bundle_install_lib_hook(Bundle),
	bundle_lldo(Bundle, install_docs).
bundle_install_(global) :-
	Bundle = lpdoc,
	bold_message("Installing LPdoc."),
	bundle_install_lib_hook(Bundle),
	bundle_install_bin_hook(Bundle),
	bundle_lldo(Bundle, install_docs),
	bold_message("LPdoc installation completed").

bundle_uninstall_hook(lpdoc) :-
	bundle_uninstall_(~instype).

bundle_uninstall_(local) :-
	Bundle = lpdoc,
	bold_message("Skipping deletion of LPdoc files."),
	bundle_uninstall_lib_hook(Bundle).
bundle_uninstall_(global) :-
	Bundle = lpdoc,
	bold_message("Uninstalling LPdoc"),
	bundle_lldo(Bundle, uninstall_docs),
	bundle_uninstall_lib_hook(Bundle),
	bundle_uninstall_bin_hook(Bundle),
	bold_message("LPdoc uninstallation completed").

% Installation of LPdoc executable
bundle_install_bin_hook(lpdoc) :-
	install_item(lpdoc, lpdoccl).

bundle_uninstall_bin_hook(lpdoc) :-
	uninstall_item(lpdoc, lpdoccl).

% ===========================================================================

bundle_install_lib_hook(lpdoc) :- install_lib_(~instype).

install_lib_(local).
install_lib_(global) :-
	Bundle = lpdoc,
	b_install_dir(~fsR(installdir_lib(Bundle))),
	install_item(Bundle, lib(Bundle, 'src')),
	install_item(Bundle, lib(Bundle, 'lib')),
	postinstall_lib_hook.

bundle_uninstall_lib_hook(lpdoc) :- uninstall_lib_(~instype).

uninstall_lib_(local).
uninstall_lib_(global) :-
	Bundle = lpdoc,
	% TODO: wrong, uninstall all versions
	%  b_uninstall_dir_rec(~fsR(installdirbase(Bundle))).
	uninstall_item(Bundle, lib(Bundle, 'lib')),
	uninstall_item(Bundle, lib(Bundle, 'src')).

postinstall_lib_hook :-
	gen_DOTcshrc,
	gen_DOTprofile,
	gen_lpdoc_lib_link.
%	make([~'word-help-setup.el']).

% (only ins)
gen_DOTcshrc :-
	wr_template(inslib(lpdoc), bundle_src(lpdoc)/lib, 'DOTcshrc', [
	    'binary_directory' = ~fsR(installdir_bin(build)),
	    'documentation_directory' = ~docdir
        ]).

% (only ins)
gen_DOTprofile :-
	wr_template(inslib(lpdoc), bundle_src(lpdoc)/lib, 'DOTprofile', [
	    'binary_directory' = ~fsR(installdir_bin(build)),
	    'documentation_directory' = ~docdir
	]).

% TODO: This just seems a symbolic link to the active lpdoc version! Generalize for all bundles (if necessary)
% TODO: call directly? (see install_lib/0)
gen_lpdoc_lib_link :-
	Bundle = lpdoc,
	% TODO: when is this link used? is this link repeated?
	--copy_file(~fsR(concat_ver(Bundle, Bundle)),
	            ~fsR(rootprefix(installdirbase(Bundle))),
		    [overwrite, symlink]).

% ===========================================================================

:- doc(section, "Tests and Benchmarks").
% TODO: Add bundle defs for unit tests, integration tests, regression
%   tests, acceptance tests, etc.

bundle_custom_hook(lpdoc, runtests) :- !,
	bold_message("Running LPdoc tests."),
	invoke_run_test_dir(~fsR(bundle_src(lpdoc))).

invoke_run_test_dir(Dir) :-
	invoke_ciaosh_batch([
          use_module(library(unittest), [run_test_dir/2]),
	  run_test_dir(Dir, [rtc_entry])
	]).

bundle_custom_hook(lpdoc, runbenchmarks) :- !.

