:- module(bundle_config__lpdoc, [], [bundleconfig]).

:- doc(title, "Configuration rules for LPdoc").
:- doc(author, "Ciao Development Team").

:- use_module(library(terms), [atom_concat/2]).
:- use_module(library(system), [get_pwnam/1]).

:- use_module(library(bundletool/paths_extra), [fsR/2]).

% ===========================================================================

% Default paths for LPdoc output
build_doc_dir := ~fsR(builddir_doc(build)).

:- bundle_param('DOCDIR', [
    rule_set_value(Value, (
      param(ciaode:'INSTYPE'(InsType)),
      InsType == 'local', build_doc_dir(Value))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      param(ciaode:'INSTALLDIR_PREFIX'(Prefix)),
      get_docdir(SysregType, Prefix, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define this to be the directory in which you wish the documentation\n"||
      "to be installed."),
    comment("Installation directory for documentation")
]).
get_docdir(all,  Prefix) := ~fsR(Prefix/'share'/'doc'/'ciao').

% ~atom_concat(~rootprefix, ~bundle_config_value(lpdoc, 'HTMLURL')).
:- bundle_param('HTMLDIR', [
    rule_set_value(Value, (
      param(ciaode:'INSTYPE'(InsType)),
      InsType == 'local', build_doc_dir(Value))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_htmldir(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define this to be the directory in which you wish the documentation\n"||
      "in HTML format to be installed."),
    comment("Installation directory for HTML manuals")
]).
% TODO: trailing /?
get_htmldir(all) := '/var/www/html/ciao'.
% get_htmldir(user) := ~atom_concat(~fsR(~get_home/'public_html'/'CiaoDE'), '/').

:- bundle_param('HTMLURL', [
    rule_set_value(Value, (
      param(ciaode:'INSTYPE'(InsType)),
      InsType == 'local', build_doc_dir(Value))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      get_htmlurl(SysregType, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define the URL to access the previous directory via WWW."),
    comment("URL for installed HTML documents")
]).
% TODO: trailing /?
get_htmlurl(all) := '/ciao/'.
get_htmlurl(user) := ~atom_concat(['/~', ~get_pwnam, '/CiaoDE/']).

:- bundle_param('MANDIR', [
    rule_set_value(Value, (
      param(ciaode:'INSTYPE'(InsType)),
      InsType == 'local', build_doc_dir(Value))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      param(ciaode:'INSTALLDIR_PREFIX'(Prefix)),
      get_mandir(SysregType, Prefix, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define this to be the directory in which you wish the man (unix manual\n"||
      "entry) file to be installed.\n"),
    comment("Installation directory for 'man' pages")
]).
get_mandir(all,  Prefix) := ~fsR(Prefix/'share'/'man').

:- bundle_param('INFODIR', [
    rule_set_value(Value, (
      param(ciaode:'INSTYPE'(InsType)),
      InsType == 'local', build_doc_dir(Value))),
    rule_default(DefValue, (
      param(ciaode:'REGISTRATION_TYPE'(SysregType)),
      param(ciaode:'INSTALLDIR_PREFIX'(Prefix)),
      get_infodir(SysregType, Prefix, DefValue))),
    %
    interactive([minimum, extended],
      % .....................................................................
      "Define this to be the directory in which you wish the info file\n"||
      "installed.  Ideally, this directory should be accesible via emacs.\n"),
    comment("Installation directory for 'info' files")
]).
get_infodir(all,  Prefix) := ~fsR(Prefix/'share'/'info').

% ---------------------------------------------------------------------------

%% The .bib files are in the repository (i.e. as SVN external),
%% it makes no sense configuring this. (JFMC)
% :- bundle_param('BIBFILES', [
%     interactive([],
%       % .....................................................................
%       "Specifies the bibtex files used to create the bibliography of the\n" ||
%       "documentation."),
%     rule_default('/home/clip/bibtex/clip/clip,/home/clip/bibtex/clip/general')
% ]).

% ---------------------------------------------------------------------------





