:- module(autodoc_aux, [], [assertions, regtypes, basicmodes, fsyntax]).

:- doc(title, "Auxiliary Definitions").
:- doc(author, "Manuel Hermenegildo").
:- doc(author, "Jose F. Morales").

:- use_module(library(messages)).
:- use_module(library(system), [file_exists/1]).

:- use_module(lpdoc(autodoc_settings)).

% ---------------------------------------------------------------------------

:- export(all_vars/1).
% All elements of the list are variables
all_vars([]).
all_vars([H|T]) :- var(H), all_vars(T).

% ---------------------------------------------------------------------------

% TODO: replace by a predicate that opens a file, and closes it if
%       fails? call_cleanup?

:- export(read_file/2).
read_file(File, Content) :-
	file_exists(File),
	!,
	try_finally(
	    open(File, read, IS),
	    read_stream(IS, Content),
	    close(IS)
	).
read_file(File, []) :-
	error_message("file ~w not found", [File]).

read_stream(IS, Content) :-
	get_code(IS, N),
	( N = -1 ->
	    Content = []
	; Content = [N|Rest],
	  read_stream(IS, Rest)
	).

:- use_module(library(system_extra), [try_finally/3]).

% ---------------------------------------------------------------------------

:- export(ascii_blank_lines/2).
ascii_blank_lines(0,"") :- !.
ascii_blank_lines(N,[0'\n | R]) :-
	N1 is N-1,
	ascii_blank_lines(N1,R).

% ---------------------------------------------------------------------------

:- use_module(library(process), [process_call/3]).
:- use_module(library(lists), [append/3, select/3]).
:- use_module(library(make/make_rt), [get_value/2, make_option/1]).
:- use_module(library(bundletool/bundle_aux), [stop_command_options/2]).

% TODO: check
% Options for logging external commands (controlled by verbosity options)
logopts(LogOpts, Opts) :-
	make_rt:get_value(stop_if_error, V),
	stop_command_options(V, Opts0),
	!,
	append(LogOpts, Opts0, Opts).
logopts(LogOpts, A) :-
	( current_fact(make_option('-v')) ->
	    A = [] % Do not store logs
	; % Store logs and dump them in case of error
	  A = [show_output_on_error, show_error_on_error|LogOpts]
	).

:- export(logged_process_call/3).
% Like process_call/3, but expands logbase(Base,Suffix) option to
% stdin and stdout redirections.
logged_process_call(Cmd, Args, Opts) :-
	select(logbase(Base, Suffix), Opts, Opts1),
	!,
	% TODO: Option for (no)append logs?
	% TODO: use stderr_to_stdout?
	atom_concat(Base, Suffix, Base2),
	atom_concat(Base2, '.log', LogFile),
	atom_concat(Base2, '.err', ErrFile),
	LogOpts = ~logopts([stdout(file(LogFile)), stderr(file(ErrFile))]),
	%
	append(LogOpts, Opts1, Opts2),
	%
	process_call(Cmd, Args, Opts2).
logged_process_call(Cmd, Args, Opts) :-
	process_call(Cmd, Args, Opts).

