% Test cases from the ISO standard

:- discontiguous((stc/4, etc/4, atc/4, program/2, macro/2)).

% Legend:
% stc --- test case in the ISO standard
% etc --- test case in A Ed-Dbali's test suite
% atc --- additional test case added at SICS

test_case(Name, iso_doc, Goal, Result, Others) :-
        stc(Name, Goal, Result, Others).
test_case(Name, eddbali, Goal, Result, Others) :-
        etc(Name, Goal, Result, Others).
test_case(Name, sics, Goal, Result, Others) :-
        atc(Name, Goal, Result, Others).

% stc(Label,    Goal,                           Result,                         Prerequisites and side
%                                                                               effects).

% 6.3.3.1
stc(term_sx: 1, read(_),                        success,                        [txtin('f(x,y).')]).
stc(term_sx: 2, read(_),                        success,                        [txtin('f(:-, ;, [:-, :-|:-]).')]).
stc(term_sx: 3, read(_),                        syntax,                         [txtin('f(,,a).')]).
stc(term_sx: 4, read(_),                        syntax,                         [txtin('[a,,|v].')]).
stc(term_sx: 5, read(_),                        syntax,                         [txtin('[a,b|,].')]).
stc(term_sx: 6, read(_),                        success,                        [txtin("f(',',a).")]).
stc(term_sx: 7, read(_),                        success,                        [txtin("[a,','|v].")]).
stc(term_sx: 8, read(_),                        success,                        [txtin("[a,b|','].")]).


% 6.3.4
stc(ops: 1,     read(_),	    		syntax,				[prog(fxops), txtin('fx fx 1.')]).
stc(ops: 2,     read(_),	    		success,			[prog(fxops), txtin('fx (fx 1).')]).
stc(ops: 3,     read(_),	    		syntax,				[prog(fxops), txtin('1 xf xf.')]).
stc(ops: 4,     read(_),	    		success,			[prog(fxops), txtin('(1 xf) xf.')]).
stc(ops: 5,     read(_),	    		syntax,				[prog(fxops),
	  			    						 txtin('1 xfx 2 xfx 3.')]).
stc(ops: 6,     read(_),	    		success,			[prog(fxops),
	  			    						 txtin('(1 xfx 2) xfx 3.')]).
stc(ops: 7,     read(_),	    		success,			[prog(fxops),
	  			    						 txtin('1 xfx (2 xfx 3).')]).
stc(ops: 8,     (read(TU), read(TB)),		[[TU == TB]], 			[prog(fyops),
	  	                     			     			 txtin('fy fy 1. fy (fy 1).')]).
stc(ops: 9,     (read(TU), read(TB)),		[[TU == TB]],			[prog(fyops),
	  	                     			     			 txtin(['1 xfy  2 xfy 3. ',
	  	                     			     			       '1 xfy (2 xfy 3).'])]).
stc(ops:10,     (read(TU), read(TB)),		[[TU == TB]],			[prog(fyops),
	  	                     			     			 txtin(['1 xfy  2 yfx 3. ',
	  	                     			     			       '1 xfy (2 yfx 3).'])]).
stc(ops:11,     (read(TU), read(TB)),		[[TU == TB]],			[prog(fyops),
	  	                     			     			 txtin('fy 2 yf. fy (2 yf).')]).
stc(ops:12,     (read(TU), read(TB)),		[[TU == TB]],  			[prog(fyops),
	  	                     			     			 txtin('1 yf yf. (1 yf) yf.')]).
stc(ops:13,     (read(TU), read(TB)),		[[TU == TB]],			[prog(fyops),
				    			    			 txtin([' 1 yfx 2  yfx 3. ',
										       '(1 yfx 2) yfx 3.'])]).

/* % pts: start of skipped tests
*/ % pts: end of skipped tests
% 6.3.4.3
stc(ops:14,     read(T),	    		[[T <-- -(1,2)]],		[txtin('-(1,2).')]).
stc(ops:15,     read(T),	    		[[T <-- -((1,2))]],		[txtin('- (1,2).')]).

% 6.3.5.1
stc(list: 1,    read(T),                        [[T <-- .(a, [])]],             [txtin('[a].')]).
stc(list: 2,    read(T),                        [[T <-- .(a, .(b, []))]],	[txtin('[a,b].')]).
stc(list: 3,    read(T),                        [[T <-- .(a, b)]],              [txtin('[a|b].')]).

% 6.3.6.1
stc(curly: 1,   read(T),                        [[T <-- '{}'(a)]],             	[txtin('{a}.')]).
stc(curly: 2,   read(T),                        [[T <-- '{}'(','(a,b))]],	[txtin('{a,b}.')]).

% 6.3.7.1
stc(dbl_qts: 1, (read(Goal), call(Goal)),	success,			[double_quotes_ex1,
										 set_double_quotes(chars)]).
stc(dbl_qts: 2, (read(Goal), call(Goal)),	success,			[double_quotes_ex1,
										 set_double_quotes(codes)]).
stc(dbl_qts: 3, (read(Goal), call(Goal)),	success,			[double_quotes_ex1,
										 set_double_quotes(atom)]).
stc(dbl_qts: 4, (read(Goal), call(Goal)),	success,			[double_quotes_ex2,
										 set_double_quotes(chars)]).
stc(dbl_qts: 5, (read(Goal), call(Goal)),	success,			[double_quotes_ex2,
										 set_double_quotes(codes)]).
stc(dbl_qts: 6, (read(Goal), call(Goal)),	success,			[double_quotes_ex2,
										 set_double_quotes(atom)]).
atc(dbl_qts: 7, read(T),			[[atom_chars('jim', T)]],	[txtin('"jim".'),
										 set_double_quotes(chars)]).
atc(dbl_qts: 8, read(T),			[[atom_codes('jim', T)]],	[txtin('"jim".'),
										 set_double_quotes(codes)]).
atc(dbl_qts: 9, read(T),			[['jim'== T]],			[txtin('"jim".'),
										 set_double_quotes(atom)]).
atc(dbl_qts:10, read(T),			[[ T <-- [] ]],			[txtin('"".'),
										 set_double_quotes(chars)]).
atc(dbl_qts:11, read(T),			[[ T <-- [] ]],			[txtin('"".'),
										 set_double_quotes(codes)]).
atc(dbl_qts:12, read(T),			[[ T <-- '' ]],			[txtin('"".'),
										 set_double_quotes(atom)]).


% 7.8.1.4
stc(true: 1,    true,                           success,                        []).

% 7.8.2.4
stc(fail: 1,    fail,                           failure,                        []).

% 7.8.3.4
stc(call: 1,    call(!),                        success,                        []).
stc(call: 2,    call(fail),                     failure,                        []).
stc(call: 3,    call((fail, _X)),               failure,                        []).
stc(call: 4,    call((fail, call(1))),          failure,                        []).
stc(call: 5,    call(b(_)),                     inst,                           [prog(call),txtout(rx('_.*'))]).
stc(call: 6,    call(b(3)),                     type(callable,(write(3),3)),    [prog(call),
										 variant(literally_iso,[
										   no -txtout(''),
										   yes-txtout('3')])
										]).
                                                                                % ERROR in the std:
                                                                                % prescribes: txtout('3').
                                                                                % Cf. call:14
stc(call: 7,    (Z=!, call((Z=!, a(X), Z))),    [[Z <-- !, X <-- 1]],           [prog(call)]).
stc(call: 8,    (call((Z=!, a(X), Z))),         [[Z <-- !, X <-- 1],
                                                 [Z <-- !, X <-- 2]],           [prog(call)]).
stc(call: 9,    call((write(3), _X)),           inst,                           [txtout('3')]).
stc(call:10,    call((write(3), call(1))),      type(callable,1),               [txtout('3')]).
stc(call:11,    call(_X),                       inst,                           []).
stc(call:12,    call(1),                        type(callable,1),               []).
stc(call:13,    call((fail, 1)),                type(callable,(fail,1)),        []).
stc(call:14,    call((write(3), 1)),            type(callable,(write(3),1)),    [txtout('')]).
stc(call:15,    call((1; true)),                type(callable,(1;true)),        []).

% 7.8.4.4
stc(cut: 1,     !,                              success,                        []).
stc(cut: 2,     (!,fail;true),                  failure,                        []).
stc(cut: 3,     (call(!),fail;true),            success,                        []).
stc(cut: 4,     (twice(_), !,
                 write('Forwards '), fail),     failure,                        [prog(cut),txtout('C Forwards ')]).
stc(cut: 5,     ((!; write('No ')),
                 write('Cut disjunction '),
                 fail),                         failure,                        [txtout('Cut disjunction ')]).
stc(cut: 6,     (twice(_), (write('No '); !),
                 write('Cut '), fail),          failure,                        [prog(cut),
                                                                                 txtout('C No Cut Cut ')]).
stc(cut: 7,     (twice(_),
                 (!, fail ;write('No '))),      failure,                        [prog(cut),txtout('C ')]).
stc(cut: 8,     (twice(X), call(X),
                 write('Forwards '), fail),     failure,                        [prog(cut),txtout(
                                                                                  'C Forwards Moss Forwards ')]).
stc(cut: 9,     (goal(X), call(X),
                 write('Forwards '), fail),     failure,                        [prog(cut),txtout(
                                                                                  'C Forwards Three Forwards ')]).
stc(cut:10,     (twice(_), \+(\+(!)),
                 write('Forwards '), fail),     failure,                        [prog(cut),txtout(
                                                                                  'C Forwards Moss Forwards ')]).
stc(cut:11,     (twice(_), once(!),
                 write('Forwards '), fail),     failure,                        [prog(cut),txtout(
                                                                                  'C Forwards Moss Forwards ')]).
stc(cut:12,     (twice(_), call(!),
                 write('Forwards '), fail),     failure,                        [prog(cut),txtout(
                                                                                  'C Forwards Moss Forwards ')]).

% 7.8.5.4
stc(and:1,      ','(X=1, var(X)),               failure,                        []).
stc(and:2,      ','(var(X), X=1),               [[X <-- 1]],                    []).
stc(and:3,      ','(X = true, call(X)),         [[X <-- true]],                 []).

% 7.8.6.4
stc(or:1,       ';'(true, fail),                success,                        []).
stc(or:2,       ';'((!, fail), true),           failure,                        []).
stc(or:3,       ';'(!, call(3)),                success,                        []).
stc(or:4,       ';'((X=1, !), X=2),             [[X <-- 1]],                    []).
stc(or:5,       ';'(X=1, X=2),                  [[X <-- 1],
                                                 [X <-- 2]],                    []).

% 7.8.7.4
stc(if_then:1,  '->'(true, true),               success,                        []).
stc(if_then:2,  '->'(true, fail),               failure,                        []).
stc(if_then:3,  '->'(fail, true),               failure,                        []).
stc(if_then:4,  '->'(true, X=1),                [[X <-- 1]],                    []).
stc(if_then:5,  '->'(';'(X=1, X=2), true),       [[X <-- 1]],                   []).
stc(if_then:6,  '->'(true, ';'(X=1, X=2)),      [[X <-- 1],
                                                 [X <-- 2]],                    []).

% 7.8.8.4
stc(ifthels:1,  ';'('->'(true, true), fail),    success,                        []).
stc(ifthels:2,  ';'('->'(fail, true), true),    success,                        []).
stc(ifthels:3,  ';'('->'(true, fail), fail),    failure,                        []).
stc(ifthels:4,  ';'('->'(fail, true), fail),    failure,                        []).
stc(ifthels:5,  ';'('->'(true, X=1), X=2),      [[X <-- 1]],                    []).
stc(ifthels:6,  ';'('->'(fail, X=1), X=2),      [[X <-- 2]],                    []).
stc(ifthels:7,  ';'('->'(true, ';'(X=1, X=2)),
                    true),                      [[X <-- 1],
                                                 [X <-- 2]],                    []).
stc(ifthels:8,  ';'('->'(';'(X=1, X=2), true),
                    true),                      [[X <-- 1]],                    []).
stc(ifthels:9,  ';'('->'(','(!, fail), true),
                    true),                      success,                        []).

% 7.8.9.4
stc(catch:1,    catch(foo(5),test(Y), true),    [[Y <-- 10]],                   [prog(catch)]).
stc(catch:2,    catch(bar(3),Z,true),           [[Z <-- 3]],                    [prog(catch)]).
stc(catch:3,    catch(true,_,3),                success,                        []).
stc(catch:4,    (catch(true,_C,write(demoen)),
                 throw(bla)),                   variant(literally_iso,[
						  no- user_error(bla),
						  yes-system]),	                [txtout('')]).
                                                  % system error ?!
stc(catch:5,    catch(car(_X),Y,true),          [[Y <-- 1]],                    [prog(catch)]).
stc(catch:6,    catch(
                 number_chars(_X,['1',a,'0']),
                 error(syntax_error(_),_),
                 fail),                         failure,                        []).
stc(catch:7,    catch(g, C, write(h1)),         [[C <-- c]],                    [prog(catch),txtout('h1')]).
stc(catch:8,    catch(coo(_X), Y, true),        [[Y <-- error(
                                                   instantiation_error,@@)]],   [prog(catch)]).

% 8.2.1.4
stc(unify: 1,   '='(1,1),                       success,                        []).
stc(unify: 2,   '='(X,1),                       [[X <-- 1]],                    []).
stc(unify: 3,   '='(X,Y),                       [[X <-- Y]],                    []).
stc(unify: 4,   '='(_,_),                       success,                        []).
stc(unify: 5,   ('='(X,Y),'='(X,abc)),          [[X <-- abc,
                                                  Y <-- abc]],                  []).
stc(unify: 6,   '='(f(X,def),f(def,Y)),         [[X <-- def,
                                                  Y <-- def]],                  []).
stc(unify: 7,   '='(1,2),                       failure,                        []).
stc(unify: 8,   '='(1,1.0),                     failure,                        []).
stc(unify: 9,   '='(g(X),f(f(X))),              failure,                        []).
stc(unify:10,   '='(f(X,1),f(a(X))),            failure,                        []).
stc(unify:11,   '='(f(X,Y,X),f(a(X),a(Y),Y,2)), failure,                        []).
stc(unify:12,   '='(X,a(X)),                    sto([[X <-- Y@a(Y)]]),          []).
stc(unify:13,   '='(f(X,1),f(a(X),2)),          sto(failure),                   []).
stc(unify:14,   '='(f(1,X,1),f(2,a(X),2)),      sto(failure),                   []).
stc(unify:15,   '='(f(1,X),f(2,a(X))),          sto(failure),                   []).
stc(unify:16,   '='(f(X,Y,X,1),
                    f(a(X),a(Y),Y,2)),          sto(failure),                   []).

% 8.2.2.4
stc(uni_woc: 1, unify_with_occurs_check(1, 1),  success,                        []).
stc(uni_woc: 2, unify_with_occurs_check(X, 1),  [[X <-- 1]],                    []).
stc(uni_woc: 3, unify_with_occurs_check(X, Y),  [[X <-- Y]],                    []).
stc(uni_woc: 4, unify_with_occurs_check(_,_),   success,                        []).
stc(uni_woc: 5, (unify_with_occurs_check(X,Y),
                 unify_with_occurs_check(
                   X, abc)),                    [[X <-- abc,
                                                  Y <-- abc]],                  []).
stc(uni_woc: 6, unify_with_occurs_check(
                  f(X,def),f(def,Y)),           [[X <-- def,
                                                  Y <-- def]],                  []).
stc(uni_woc: 7, unify_with_occurs_check(1,2),   failure,                        []).
stc(uni_woc: 8, unify_with_occurs_check(1,1.0), failure,                        []).
stc(uni_woc: 9, unify_with_occurs_check(
                  g(X),f(f(X))),                failure,                        []).
stc(uni_woc:10, unify_with_occurs_check(
                  f(X,1),f(a(X))),              failure,                        []).
stc(uni_woc:11, unify_with_occurs_check(
                  f(X,Y,X),f(a(X),a(Y),Y,2)),   failure,                        []).
stc(uni_woc:12, unify_with_occurs_check(
                  X,a(X)),                      failure,                        []).
stc(uni_woc:13, unify_with_occurs_check(
                  f(X,1),f(a(X),2)),            failure,                        []).
stc(uni_woc:14, unify_with_occurs_check(
                  f(1,X,1),f(2,a(X),2)),        failure,                        []).
stc(uni_woc:15, unify_with_occurs_check(
                  f(1,X),f(2,a(X))),            failure,                        []).
stc(uni_woc:16, unify_with_occurs_check(
                  f(X,Y,X,1),f(a(X),a(Y),Y,2)), failure,                        []).

% 8.2.3.4
stc(not_uni: 1, '\\='(1,1),                     failure,                        []).
stc(not_uni: 2, \=(_X,1),                       failure,                        []).
stc(not_uni: 3, '\\='(_X,_Y),                   failure,                        []).
stc(not_uni: 4, \=(_,_),                        failure,                        []).
stc(not_uni: 5, \=(f(_X,def),f(def,_Y)),        failure,                        []).
stc(not_uni: 6, '\\='(1,2),                     success,                        []).
stc(not_uni: 7, \=(1,1.0),                      success,                        []).
stc(not_uni: 8, '\\='(g(X),f(f(X))),            success,                        []).
stc(not_uni: 9, \=(f(X,1),f(a(X))),             success,                        []).
stc(not_uni:10, '\\='(f(X,Y,X),
                      f(a(X),a(Y),Y,2)),        success,                        []).
stc(not_uni:11, \=(X,a(X)),                     sto(failure),                   []).
stc(not_uni:12, '\\='(f(X,1),f(a(X),2)),        sto(success),                   []).
stc(not_uni:13, '\\='(f(1,X,1),f(2,a(X),2)),    sto(success),                   []).
stc(not_uni:14, \=(f(1,X),f(2,a(X))),           sto(success),                   []).
stc(not_uni:15, '\\='(f(X,Y,X,1),
                      f(a(X),a(Y),Y,2)),        sto(success),                   []).

% 8.3.1.4
stc(var:1,      var(foo),                       failure,                        []).
stc(var:2,      var(_Foo),                      success,                        []).
stc(var:3,      (foo = Foo, var(Foo)),          failure,                        []).
stc(var:4,      var(_),                         success,                        []).

% 8.3.2.4
stc(atom:1,     atom(atom),                     success,                        []).
stc(atom:2,     atom('string'),                 success,                        []).
stc(atom:3,     atom(a(b)),                     failure,                        []).
stc(atom:4,     atom(_Var),                     failure,                        []).
stc(atom:5,     atom([]),                       success,                        []).
stc(atom:6,     atom(6),                        failure,                        []).
stc(atom:7,     atom(3.3),                      failure,                        []).

% 8.3.3.4
stc(integer:1,  integer(3),                     success,                        []).
stc(integer:2,  integer(-3),                    success,                        []).
stc(integer:3,  integer(3.3),                   failure,                        []).
stc(integer:4,  integer(_X),                    failure,                        []).
stc(integer:5,  integer(atom),                  failure,                        []).

% 8.3.4.4
stc(float:1,    float(3.3),                     success,                        []).
stc(float:2,    float(-3.3),                    success,                        []).
stc(float:3,    float(3),                       failure,                        []).
stc(float:4,    float(atom),                    failure,                        []).
stc(float:5,    float(_X),                      failure,                        []).

% 8.3.5.4
stc(atomic:1,   atomic(atom),                   success,                        []).
stc(atomic:2,   atomic(a(b)),                   failure,                        []).
stc(atomic:3,   atomic(_Var),                   failure,                        []).
stc(atomic:4,   atomic(6),                      success,                        []).
stc(atomic:5,   atomic(3.3),                    success,                        []).

% 8.3.6.4
stc(compound:1, compound(33.3),                 failure,                        []).
stc(compound:2, compound(-33.3),                failure,                        []).
stc(compound:3, compound(-a),                   success,                        []).
stc(compound:4, compound(_),                    failure,                        []).
stc(compound:5, compound(a),                    failure,                        []).
stc(compound:6, compound(a(b)),                 success,                        []).
stc(compound:7, compound([]),                   failure,                        []).
stc(compound:8, compound([a]),                  success,                        []).

% 8.3.7.4
stc(nonvar:1,   nonvar(33.3),                   success,                        []).
stc(nonvar:2,   nonvar(foo),                    success,                        []).
stc(nonvar:3,   nonvar(_Foo),                   failure,                        []).
stc(nonvar:4,   (foo = Foo, nonvar(Foo)),       success,                        []).
stc(nonvar:5,   nonvar(_),                      failure,                        []).
stc(nonvar:6,   nonvar(a(b)),                   success,                        []).

% 8.3.8.4
stc(number:1,   number(3),                      success,                        []).
stc(number:2,   number(3.3),                    success,                        []).
stc(number:3,   number(-3),                     success,                        []).
stc(number:4,   number(a),                      failure,                        []).
stc(number:5,   number(_X),                     failure,                        []).

% 8.4.1.4
stc(trm_cmp: 1, '@=<'(1.0, 1),                  success,                        []).
stc(trm_cmp: 2, '@<'(1.0, 1),                   success,                        []).
stc(trm_cmp: 3, '\\=='(1, 1),                   failure,                        []).
stc(trm_cmp: 4, '@=<'(aardvark, zebra),         success,                        []).
stc(trm_cmp: 5, '@=<'(short, short),            success,                        []).
stc(trm_cmp: 6, '@=<'(short, shorter),          success,                        []).
stc(trm_cmp: 7, '@>='(short, shorter),          failure,                        []).
stc(trm_cmp: 8, '@<'(foo(a,b), north(a)),       failure,                        []).
stc(trm_cmp: 9, '@>'(foo(b), foo(a)),           success,                        []).
stc(trm_cmp:10, '@<'(foo(a, _X), foo(b, _Y)),   success,                        []).
stc(trm_cmp:11, '@<'(foo(_X, a), foo(_Y, b)),   succfail,                       []).
stc(trm_cmp:12, '@=<'(X, X),                    success,                        []).
stc(trm_cmp:13, '=='(X, X),                     success,                        []).
stc(trm_cmp:14, '@=<'(_X, _Y),                  succfail,                       []).
stc(trm_cmp:15, '=='(_X, _Y),                   failure,                        []).
stc(trm_cmp:16, \==(_, _),                      success,                        []).
stc(trm_cmp:17, '=='(_, _),                     failure,                        []).
stc(trm_cmp:18, '@=<'(_, _),                    succfail,                       []).
stc(trm_cmp:19, '@=<'(foo(_X, a), foo(_Y, b)),  succfail,                       []).

% 8.5.1.4
stc(functor: 1, functor(foo(a,b,c), foo, 3),    success,                        []).
stc(functor: 2, functor(foo(a,b,c),X,Y),        [[X <-- foo, Y <-- 3]],         []).
stc(functor: 3, functor(X,foo,3),               [[X <-- foo(_A,_B,_C)]],        []).
stc(functor: 4, functor(X,foo,0),               [[X <-- foo]],                  []).
stc(functor: 5, functor(mats(A,B),A,B),         [[A <-- mats,B <-- 2]],         []).
stc(functor: 6, functor(foo(a),foo,2),          failure,                        []).
stc(functor: 7, functor(foo(a),fo,1),           failure,                        []).
stc(functor: 8, functor(1,X,Y),                 [[X <-- 1,Y <-- 0]],            []).
stc(functor: 9, functor(X,1.1,0),               [[X <-- 1.1]],                  []).
stc(functor:10, functor([_|_],'.',2),           success,                        []).
stc(functor:11, functor([],[],0),               success,                        []).
stc(functor:12, functor(_X, _Y, 3),             inst,                           []).
stc(functor:13, functor(_X, foo, _N),           inst,                           []).
stc(functor:14, functor(_X, foo, a),            type(integer,a),                []).
stc(functor:15, functor(_F, 1.5, 1),            type(atom,1.5),                 []).
stc(functor:16, functor(_F,foo(a),1),           type(atomic,foo(a)),            []).
stc(functor:17, (current_prolog_flag(
                   max_arity, A), X is A+1,
                 functor(_T, foo, X)),          repr(max_arity),                []).
stc(functor:18, functor(_T, foo, -1),           dom(not_less_than_zero,-1),     []).

% 8.5.2.4
stc(arg: 1,     arg(1, foo(a,b), a),            success,                        []).
stc(arg: 2,     arg(1,foo(a,b),X),              [[X <-- a]],                    []).
stc(arg: 3,     arg(1,foo(X,b),a),              [[X <-- a]],                    []).
stc(arg: 4,     arg(1,foo(X,b),Y),              [[Y <-- X]],                    []).
stc(arg: 5,     arg(1,foo(a,b),b),              failure,                        []).
stc(arg: 6,     arg(0,foo(a,b),foo),            failure,                        []).
stc(arg: 7,     arg(3,foo(3,4),_N),             failure,                        []).
stc(arg: 8,     arg(_X,foo(a,b),a),             inst,                           []).
stc(arg: 9,     arg(1,_X,a),                    inst,                           []).
stc(arg:10,     arg(0,atom,_A),                 type(compound, atom),           []).
stc(arg:11,     arg(0,3,_A),                    type(compound, 3),              []).
stc(arg:12,     arg(1,foo(X),u(X)),             sto([[X<--Y@u(Y)]]),            []).

etc(arg:13,     arg(-3,foo(a,b),_A),            dom(not_less_than_zero,-3),     []).
etc(arg:14,     arg(a,foo(a,b),_X),             type(integer, a),               []).
etc(arg:15,     arg(2,foo(a,f(X,b),c), f(a,Y)), [[X <-- a, Y <-- b]],           []).
atc(arg:16,     arg(1,3,_A),                    type(compound, 3),              []).

% 8.5.3.4
stc(univ: 1,    '=..'(foo(a,b), [foo,a,b]),     success,                        []).
stc(univ: 2,    '=..'(X, [foo,a,b]),            [[X <-- foo(a,b)]],             []).
stc(univ: 3,    '=..'(foo(a,b), L),             [[L <-- [foo,a,b]]],            []).
stc(univ: 4,    '=..'(foo(X,b), [foo,a,Y]),     [[X <-- a,Y <-- b]],            []).
stc(univ: 5,    '=..'(1, [1]),                  success,                        []).
stc(univ: 6,    '=..'(foo(a,b), [foo,b,a]),     failure,                        []).
stc(univ: 7,    '=..'(_X, _Y),                  inst,                           []).
stc(univ: 8,    '=..'(_X, [foo,a|_Y]),          inst,                           []).
stc(univ: 9,    '=..'(_X, [foo|bar]),           type(list,[foo|bar]),           []).
stc(univ:10,    '=..'(_X, [_Foo,bar]),          inst,                           []).
stc(univ:11,    '=..'(_X, [3,1]),               type(atom,3),                   []).
stc(univ:12,    '=..'(_X, [1.1,foo]),           type(atom,1.1),                 []).
stc(univ:13,    '=..'(_X, [a(b),1]),            type(atom,a(b)),                []).
stc(univ:14,    '=..'(_X, 4),                   type(list,4),                   []).
stc(univ:15,    '=..'(f(X), [f,u(X)]),          sto([[X <-- Y@u(Y)]]),          []).

atc(univ:16,    '=..'(_X, [f(a)]),              type(atomic, f(a)),             []).
atc(univ:17,    '=..'(_X, []),                  dom(non_empty_list, []),        []).
atc(univ:18,    '=..'(_X, [f|L]),               repr(max_arity),                [list_of(N, 1, L)]) :-
	context_info(max_arity, Max), N is Max+1.


% 8.5.4.4
stc(cp_term:1,  copy_term(_X,_Y),               success,                        []).
stc(cp_term:2,  copy_term(_X,3),                success,                        []).
stc(cp_term:3,  copy_term(_,a),                 success,                        []).
stc(cp_term:4,  copy_term(a+X,X+b),             [[X <-- a]],                    []).
stc(cp_term:5,  copy_term(_,_),                 success,                        []).
stc(cp_term:6,  copy_term(X+X+_Y,A+B+B),        [[A <-- B]],                    []).
stc(cp_term:7,  copy_term(a,b),                 failure,                        []).
stc(cp_term:8,  (copy_term(a+X,X+b),
                 copy_term(a+X,X+b)),           failure,                        []).
stc(cp_term:9,  copy_term(demoen(X,X),
                          demoen(Y,f(Y))),      sto([[Y<--Z@f(Z)]]),            []).

% 8.6.1.4
stc(is: 1,      'is'(Result,3 + 11.0),          [[Result <-- 14.0]],            []).
stc(is: 2,      (X = 1+2, 'is'(Y, X*3)),        [[X <-- 1+2, Y <-- 9]],         []).
stc(is: 3,      'is'(3, 3),                     success,                        []).
stc(is: 4,      'is'(3, 3.0),                   failure,                        []).
stc(is: 5,      'is'(foo, 77),                  failure,                        []).
stc(is: 6,      'is'(77, _N),                   inst,                           []).

% 8.7.1.4
stc(ari_cmp: 1, '=:='(0,1),                     failure,                        []).
stc(ari_cmp: 2, '=\\='(0, 1),                   success,                        []).
stc(ari_cmp: 3, '<'(0, 1),                      success,                        []).
stc(ari_cmp: 4, '>'(0, 1),                      failure,                        []).
stc(ari_cmp: 5, '>='(0, 1),                     failure,                        []).
stc(ari_cmp: 6, '=<'(0, 1),                     success,                        []).
stc(ari_cmp: 7, '=:='(1.0, 1),                  success,                        []).
stc(ari_cmp: 8, =\=(1.0, 1),                    failure,                        []).
stc(ari_cmp: 9, '<'(1.0, 1),                    failure,                        []).
stc(ari_cmp:10, '>'(1.0, 1),                    failure,                        []).
stc(ari_cmp:11, '>='(1.0, 1),                   success,                        []).
stc(ari_cmp:12, '=<'(1.0, 1),                   success,                        []).
stc(ari_cmp:13, '=:='(3*2, 7-1),                success,                        []).
stc(ari_cmp:14, '=\\='(3*2, 7-1),               failure,                        []).
stc(ari_cmp:15, '<'(3*2, 7-1),                  failure,                        []).
stc(ari_cmp:16, '>'(3*2, 7-1),                  failure,                        []).
stc(ari_cmp:17, '>='(3*2, 7-1),                 success,                        []).
stc(ari_cmp:18, '=<'(3*2, 7-1),                 success,                        []).
stc(ari_cmp:19, '=:='(_X, 5),                   inst,                           []).
stc(ari_cmp:20, =\=(_X, 5),                     inst,                           []).
stc(ari_cmp:21, '<'(_X, 5),                     inst,                           []).
stc(ari_cmp:22, '>'(_X, 5),                     inst,                           []).
stc(ari_cmp:23, '>='(_X, 5),                    inst,                           []).
stc(ari_cmp:24, '=<'(_X, 5),                    inst,                           []).

% 8.8.1.4
stc(clause: 1,  clause(cat,true),               success,                        [prog(clause)]).
stc(clause: 2,  clause(dog,true),               success,                        [prog(clause)]).
stc(clause: 3,  clause(legs(I,6),Body),         [[Body <-- insect(I)]],         [prog(clause)]).
stc(clause: 4,  clause(legs(C,7),Body),         [(D= @C)^
                                                 [Body <-- (call(D),call(D))]], [prog(clause)]).
stc(clause: 5,  clause(insect(I),T),            [[I <-- ant, T <-- true],
                                                 [I <-- bee, T <-- true]],      [prog(clause)]).
stc(clause: 6,  clause(x,_Body),                failure,                        []).
stc(clause: 7,  clause(_,_B),                   inst,                           []).
stc(clause: 8,  clause(4,_B),                   type(callable,4),               []).
stc(clause: 9,  clause(elk(_N),_Body),          perm(access,private_procedure,
                                                     @elk/1),                   [prog(clause)]).
stc(clause:10,  clause(atom(_),_Body),          perm(access,private_procedure,
                                                     @atom/1),                  []).
stc(clause:11,  clause(legs(A,6),insect(f(A))), sto([[A <-- Y@f(Y)]]),          [prog(clause)]).

etc(clause:12,  clause(f(_),5),                 type(callable,5),               []).


% 8.8.2.4
stc(cur_prd: 1, current_predicate(dog/0),       success,                        [prog(clause)]).
stc(cur_prd: 2, current_predicate(
                     current_predicate/1),      failure,                        []).
stc(cur_prd: 3, current_predicate(elk/Arity),   [[Arity <-- 1]],                [prog(clause)]).
stc(cur_prd: 4, current_predicate(foo/_A),      failure,                        [prog(clause)]).
stc(cur_prd: 5, current_predicate(Name/1),      variant(modules, [
						 yes - {[Name <-- elk],
							[Name <-- insect]},
						 no  - {[Name <-- elk],
							[Name <-- insect], ...}
								 ]),            [prog(clause)]).
stc(cur_prd: 6, current_predicate(4),           type(predicate_indicator,@4),   []).

etc(cur_prd: 7, current_predicate(dog),         type(predicate_indicator,@dog), []).
etc(cur_prd: 8, current_predicate(0/dog),       type(predicate_indicator,
                                                          @0/dog),              []).
etc(cur_prd: 9, current_predicate(P),           variant(modules, [
                                                  yes - {[P<--cat/0],
							 [P<--dog/0],
							 [P<--elk/1],
							 [P<--legs/2],
							 [P<--insect/1]},
						  no  - {[P<--cat/0],
							 [P<--dog/0],
							 [P<--elk/1],
							 [P<--legs/2],
							 [P<--insect/1], ...}
								 ]),            [prog(clause)]).

% 8.9.1.4
stc(asserta: 1, asserta(legs(octopus, 8)),      success,                        [prog(dynpreds1),dynp(dynpreds2)]).
stc(asserta: 2, asserta(
                (legs(A,4):-animal(A))),        success,                        [prog(dynpreds2),dynp(dynpreds3)]).
stc(asserta: 3, asserta((foo(X) :- X,call(X))), success,                        [prog(dynpreds3),dynp(dynpreds4)]).
stc(asserta: 4, asserta(_),                     inst,                           []).
stc(asserta: 5, asserta(4),                     type(callable,4),               []).
stc(asserta: 6, asserta((foo :- 4)),            type(callable,4),               []).
stc(asserta: 7, asserta((atom(_) :- true)),     perm(modify,static_procedure,
                                                     @atom/1),                  []).

etc(asserta: 8, (asserta(insct(bee)), insct(X),
                 asserta(insct(ant)),insct(Y)), [[X <-- bee, Y <-- ant],
                                                 [X <-- bee, Y <-- bee]],       []).

% 8.9.2.4
stc(assertz: 1, assertz(legs(spider, 8)),       success,                        [prog(dynpreds4),dynp(dynpreds5)]).
stc(assertz: 2, assertz((legs(B, 2):-bird(B))), success,                        [prog(dynpreds5),dynp(dynpreds6)]).
stc(assertz: 3, assertz((foo(X):-X->call(X))),  success,                        [prog(dynpreds6),dynp(dynpreds7)]).
stc(assertz: 4, assertz(_),                     inst,                           []).
stc(assertz: 5, assertz(4),                     type(callable,4),               []).
stc(assertz: 6, assertz((foo :- 4)),            type(callable,4),               []).
stc(assertz: 7, assertz((atom(_) :- true)),     perm(modify,static_procedure,
                                                     @atom/1),                  []).
% 8.9.3.4
stc(retract: 1, retract(legs(octopus, 8)),      success,                        [prog(dynpreds7),dynp(dynpreds8)]).
stc(retract: 2, retract(legs(spider, 6)),       failure,                        [prog(dynpreds8),dynp(dynpreds8)]).
stc(retract: 3, retract((legs(X, 2) :- T)),     [[T <-- bird(X)]],              [prog(dynpreds8),dynp(dynpreds9)]).
stc(retract: 4, retract((legs(X, Y) :- Z)),     [[Y <-- 4, Z <-- animal(X)],
                                                 [Y <-- 6, Z <-- insect(X)],
                                                 [X <-- spider, Y <-- 8,
                                                  Z <-- true]],                 [prog(dynpreds9),dynp(dynpreds10)]).
stc(retract: 5, retract((legs(_X, _Y) :- _Z)),  failure,                        [prog(dynpreds10)]).
stc(retract: 6, (retract(insect(I)), write(I),
                 retract(insect(bee))),         [[I<--ant]],                    [prog(dynpreds10),txtout('antbee'),
                                                                                 dynp(dynpreds11)]).
stc(retract: 7, retract((foo(A) :- A,call(A))), sto(undefined),			[prog(dynpreds11)]).
                                                % too complicated to describe
                                                % the effect of cyclic terms
                                                % plus module qualification...
stc(retract: 8, retract((foo(C) :- A -> B)),    [(D = @C)^
                                                 [A<--call(D), B<--call(D)]],   [prog(dynpreds11),
                                                                                 dynp(dynpreds12)]).
stc(retract: 9, retract((_X :- in_eec(_Y))),    inst,                           []).
stc(retract:10, retract((4 :- _X)),             type(callable,4),               []).
stc(retract:11, retract((atom(X) :- X =='[]')), perm(modify,static_procedure,
                                                     @atom/1),                  []).

% 8.9.4.4
stc(abolish: 1, abolish(foo/2),                 success,                        []).
stc(abolish: 2, abolish(foo/_),                 inst,                           []).
stc(abolish: 3, abolish(foo),                   type(predicate_indicator,@foo), []).
stc(abolish: 4, abolish(foo(X)),                type(predicate_indicator,
                                                     @foo(X)),                  []).
stc(abolish: 5, abolish(abolish/1),             perm(modify,static_procedure,
                                                     @abolish/1),               []).

etc(abolish: 6, abolish(foo/1),                 success,                        [prog(abolish)]).
etc(abolish: 7, (insect(X), abolish(insect/1)), [[X <-- ant],
                                                 [X <-- bee]],                  [prog(abolish)]).
etc(abolish: 8, abolish(foo/_),                 inst,                           [prog(abolish)]).
etc(abolish: 9, abolish(bar/1),                 perm(modify,static_procedure,
                                                     @bar/1),                   [prog(abolish)]).
etc(abolish:10, abolish(foo/a),                 type(integer,a),                []).
etc(abolish:11, abolish(foo/(-1)),              dom(not_less_than_zero,-1),     []).
etc(abolish:12, abolish(foo/X),                 repr(max_arity),                [over_arity(X)]).
etc(abolish:13, abolish(5/2),                   type(atom,5),                   []).
etc(abolish:14, abolish(insect),                type(predicate_indicator,
                                                     @insect),                  [prog(abolish)]).

% 8.10.1.4
stc(findall: 1, findall(X,(X=1;X=2),S),         [[S <-- [1,2]]],                []).
stc(findall: 2, findall(X+_Y,(X=1),S),          [[S <-- [1+_]]],                []).
stc(findall: 3, findall(_X,fail,L),             [[L <-- []]],                   []).
stc(findall: 4, findall(X,(X=1;X=1),S),         [[S <-- [1,1]]],                []).
stc(findall: 5, findall(X,(X=2;X=1),[1,2]),     failure,                        []).
stc(findall: 6, findall(X,(X=1;X=2),[X,Y]),     [[X <-- 1, Y <-- 2]],           []).
stc(findall: 7, findall(_X,_Goal,_S),           inst,                           []).
stc(findall: 8, findall(_X,4,_S),               type(callable,4),               []).
atc(findall: 9, findall(X,X=1,[A|1]),           type(list,[A|1]),               []).

% 8.10.2.4
stc(bagof: 1,   bagof(X,(X=1;X=2),S),           [[S <-- [1, 2]]],               []).
stc(bagof: 2,   bagof(X,(X=1;X=2),X),           [[X <-- [1, 2]]],               []).
stc(bagof: 3,   bagof(X,(X=Y;X=Z),S),           [[S <-- [Y, Z]]],               []).
stc(bagof: 4,   bagof(_X,fail,_S),              failure,                        []).
stc(bagof: 5,   bagof(1,(Y=1;Y=2),L),           {[L <-- [1], Y <-- 2],
                                                 [L <-- [1], Y <-- 1]},         []).
stc(bagof: 6,   bagof(f(X,Y),(X=a;Y=b),L),      [[L <-- [f(a, _), f(_, b)]]],   []).
stc(bagof: 7,   bagof(X,Y^((X=1,Y=1)
                          ;(X=2,Y=2)),S),       [[S <-- [1, 2]]],               []).
stc(bagof: 8,   bagof(X,Y^((X=1;Y=1)
                          ;(X=2,Y=2)),S),       [[S <-- [1, _, 2]]],            []).
stc(bagof: 9,   bagof(X,(Y^(X=1;Y=1);X=3),S),
                                                variant(caret, [undef-
						[[S <-- [3], Y <-- _]], % no ^/2
								def-
						{[S <-- [1,3] ],
						 [S <-- [_], Y <-- 1]}  % ^ /2 exists
							       ]),     		[pre(set_prolog_flag(unknown, fail))]).
stc(bagof:10,   bagof(X,(X=Y;X=Z;Y=1),S),       {[S <-- [Y, Z]],
                                                 [S <-- [_], Y <-- 1]},         []).
stc(bagof:11,   bagof(X,a(X,Y),L),              [[L <-- [1, 2], Y <-- f(_)]],   [prog([a(1,f(_)),a(2,f(_))])]).
stc(bagof:12,   bagof(X,b(X,Y),L),              {[L <-- [1, 1, 2], Y <-- 1],
                                                 [L <-- [1, 2, 2], Y <-- 2]},   [prog([b(1,1), b(1,1), b(1,2),
                                                                                       b(2,1), b(2,2), b(2,2)])]).
stc(bagof:13,   bagof(_X,_Y^_Z,_L),             inst,                           []).
stc(bagof:14,   bagof(_X,1,_L),                 type(callable, 1),              []).

% 8.10.3.4
stc(setof: 1,   setof(X,(X=1;X=2),S),           [[S <-- [1, 2]]],               []).
stc(setof: 2,   setof(X,(X=1;X=2),X),           [[X <-- [1, 2]]],               []).
stc(setof: 3,   setof(X,(X=2;X=1),S),           [[S <-- [1, 2]]],               []).
stc(setof: 4,   setof(X,(X=2;X=2),S),           [[S <-- [2]]],                  []).
stc(setof: 5,   setof(X,(X=Y;X=Z),S),           ([[S <-- [Z, Y]]];
                                                 [[S <-- [Y, Z]]]),             []).
stc(setof: 6,   setof(_X,fail,_S),              failure,                        []).
stc(setof: 7,   setof(1,(Y=1;Y=2),L),           {[L <-- [1], Y <-- 1],
                                                 [L <-- [1], Y <-- 2]},         []).
stc(setof: 8,   setof(f(X,Y),(X=a;Y=b),L),      [[L <-- [f(_, b),f(a, _)]]],    []).
stc(setof: 9,   setof(X,Y^((X=1,Y=1)
                          ;(X=2,Y=2)),S),       [[S <-- [1, 2]]],               []).
stc(setof:10,   setof(X,Y^((X=1;Y=1)
                          ;(X=2,Y=2)),S),       [[S <-- [_, 1, 2]]],            []).
stc(setof:11,   setof(X,(Y^(X=1;Y=1);X=3),S),   variant(caret, [undef-
						[[S <-- [3], Y <-- _]], % no ^
								def-
						{[S <-- [1,3] ],
						 [S <-- [_], Y <-- 1]}  % ^ /2 exists
								]),     	[pre(set_prolog_flag(unknown, fail))]).
stc(setof:12,   setof(X,(X=Y;X=Z;Y=1),S),       [[S <-- ([Y, Z];[Z, Y])],
                                                 [S <-- [_], Y <-- 1]],         []).
stc(setof:13,   setof(X,a(X,Y),L),              [[L <-- [1, 2], Y <-- f(_)]],   [prog([(a(1,f(_)) :- true),
                                                                                       (a(2,f(_)) :- true)])]).
stc(setof:14,   setof(X,my_member(X,[f(U,b),
                                  f(V,c)]),L),  [[L <-- ([f(U,b),f(V,c)]
                                                        ;[f(V,c),f(U,b)])]],    [prog(member)]).
stc(setof:15,   setof(X,my_member(X,[f(U,b),
                                  f(V,c)]),
                      [f(a,c),f(a,b)]),         ([[U <-- a, V <-- a]]
                                                ;failure),                      [prog(member)]).
stc(setof:16,   setof(X,my_member(X,[f(b,U),
                                  f(c,V)]),
                      [f(b,a),f(c,a)]),         [[U <-- a, V <-- a]],           [prog(member)]).
stc(setof:17,   setof(X,
                  my_member(X,[V,U,f(U),f(V)]),L), [[ L <-- ([U,V,f(U),f(V)]
                                                         ;[V,U,f(V),f(U)])]],   [prog(member)]).
stc(setof:18,   setof(X,
                  my_member(X,[V,U,f(U),f(V)]),
                  [a,b,f(a),f(b)]),             ([[ U <-- a, V <-- b]]
                                                ;[[ U <-- b, V <-- a]]),        [prog(member)]).
stc(setof:19,   setof(X,
                  my_member(X,[V,U,f(U),f(V)]),
                  [a,b,f(b),f(a)]),             failure,                        [prog(member)]).
stc(setof:20,   setof(X,
                  exists(U,V)^my_member(X,[V,U,f(U),f(V)]),
                  [a,b,f(a),f(b)]),             success,                        [prog(member)]).
stc(setof:21,   setof(X,b(X,Y),L),              {[L <-- [1, 2], Y <-- 1],
                                                 [L <-- [1, 2], Y <-- 2]},      [prog([b(1,1), b(1,1), b(1,2),
                                                                                       b(2,1), b(2,2), b(2,2)])]).
stc(setof:22,   setof(X-Xs,
                  Y^setof(Y,b(X,Y),Xs),L),      [[L <-- [1-[1,2],2-[1,2]]]],    [prog([b(1,1), b(1,1), b(1,2),
                                                                                       b(2,1), b(2,2), b(2,2)])]).
stc(setof:23,   setof(X-Xs,
                  setof(Y,b(X,Y),Xs),L),        [[Y <-- _,
                                                  L <-- [1-[1,2],2-[1,2]]]],    [prog([b(1,1), b(1,1), b(1,2),
                                                                                       b(2,1), b(2,2), b(2,2)])]).
stc(setof:24,   setof(X-Xs,
                  bagof(Y,d(X,Y),Xs),L),        [[Y <-- _,
                                                  L <-- [1-[1,2,1],
                                                         2-[2,1,2]]]],          [prog([d(1,1), d(1,2), d(1,1),
                                                                                       d(2,2), d(2,1), d(2,2)])]).

etc(setof:25,   setof(f(X,Y),X=Y,[f(g(Z),Z)]),  sto(undefined),                 []).
etc(setof:26,   setof(X, X^(true; 4),_L),       type(callable,(true;4)),   []).
atc(setof:27,   setof(_X,A^A^1,_L),             type(callable,1),               []).
atc(setof:28,   setof(X,X=1,[1|A]),             [[A <-- []]],                   []).
atc(setof:29,   setof(X,X=1,[A|1]),             type(list,[A|1]),               []).

% 8.11.1
atc(currinp: 1, current_input(_S),              success,                        []).
atc(currinp: 2, current_input(foo),             dom(stream,foo),                []).
atc(currinp: 3, current_input(S),               failure,                        [pre(current_output(S))]).
atc(currinp: 4, current_input(S),               dom(stream,S),                  [closed_instream(S,[])]).
atc(currinp: 5, current_input(S),               success,                        [pre(current_input(S))]).

% 8.11.2
atc(currout: 1, current_output(_S),             success,                        []).
atc(currout: 2, current_output(foo),            dom(stream,foo),                []).
atc(currout: 3, current_output(S),              failure,                        [pre(current_input(S))]).
atc(currout: 4, current_output(S),              dom(stream,S),                  [closed_outstream(S,[])]).
atc(currout: 5, current_output(S),              success,                        [pre(current_output(S))]).

% 8.11.3
atc(set_in: 1,  set_input(S),                   success,                        [pre(current_input(S))]).
atc(set_in: 2,  set_input(_S),                  inst,                           []).
atc(set_in: 3,  set_input(foo),                 dom(stream_or_alias,foo),       []).
atc(set_in: 4,  set_input(S),                   exist(stream,S),                [closed_instream(S,[])]).
atc(set_in: 5,  set_input(S),                   perm(input,stream,S),           [pre(current_output(S))]).

% 8.11.4
atc(set_out: 1, set_output(S),                  success,                        [pre(current_output(S))]).
atc(set_out: 2, set_output(_S),                 inst,                           []).
atc(set_out: 3, set_output(foo),                dom(stream_or_alias,foo),       []).
atc(set_out: 4, set_output(S),                  exist(stream,S),                [closed_outstream(S,[])]).
atc(set_out: 5, set_output(S),                  perm(output,stream,S),          [pre(current_input(S))]).

% 8.11.5.4
stc(open: 1,    open('roger_data', read, D,
                     [type(binary)]),           [[at_end_of_stream(D)]],        [makebin(roger_data, [])]).
stc(open: 2,    open('scowen', write, D,
                     [alias(editor)]),          [[stream_property(D,
                                                        alias(editor)) ]],      [checktxt(scowen,'')]).
stc(open: 3,    open('dave', read, DD, []),     [[read(DD, foo),
                                                  at_end_of_stream(DD)]],	[maketxt(dave, 'foo.')]).

atc(open: 4,    open(_,read,_),                 inst,                           []).
atc(open: 5,    open(f,_,_),                    inst,                           []).
atc(open: 6,    open(f,write,_,_),              inst,                           []).
atc(open: 7,    open(f,write,_,[type(text)|_]), inst,                           []).
atc(open: 8,    open(f,write,_,[type(text),_]), inst,                           []).
atc(open: 9,    open(f,1,_),                    type(atom,1),                   []).
atc(open:10,    open(f,write,_,type(text)),     type(list,type(text)),          []).
atc(open:11,    open(f,write, bar),             type(variable,bar),             []).
atc(open:12,    open(foo(1,2),write,_),         dom(source_sink,foo(1,2)),      []).
atc(open:13,    open('foo',red,_),              dom(io_mode,red),               []).
atc(open:14,    open(foo,write,_,[bar]),        dom(stream_option,bar),         []).
atc(open:15,    open('nonexistent',read,_),     exist(source_sink,@@),          []).
atc(open:16,    open(bar,write,_,[alias(a)]),   perm(open,source_sink,
                                                     alias(a)),                 [pre(open(foo,write,_,[alias(a)])),
                                                                                 clean(foo)]).
/* [PM] 3.8.4 Was
atc(open:17,    open('/dev/tty',read,_,
                     [reposition(true)]),       perm(open,source_sink,
                                                     reposition(true)),[]).
*/
/* [PM] 3.8.4 /dev/tty does not exist on Windows and not when building
   via rsh. Expect therefore existence error as an alternative. */
atc(open:17,    open('/dev/tty',read,_,
                     [reposition(true)]),       ( perm(open,source_sink,
                                                       reposition(true))
                                                ; exist(source_sink,_)
                                                ),                              []).

% 8.11.6
atc(close: 1,   close(S),                       success,                        [pre(open('foo',write,S)),
                                                                                 checktxt(foo,'')]).
atc(close: 2,   close(_),                       inst,                           []).
atc(close: 3,   close(S,_),                     inst,                           [pre(current_input(S))]).
atc(close: 4,   close(S,[force(true)|_]),       inst,                           [pre(current_input(S))]).
atc(close: 5,   close(S,[force(true),_]),       inst,                           [pre(current_input(S))]).
atc(close: 6,   close(S,foo),                   type(list,foo),                 [pre(current_input(S))]).
atc(close: 7,   close(S,[foo]),                 dom(close_option,foo),          [pre(current_input(S))]).
atc(close: 8,   close(foo),                     dom(stream_or_alias,foo),       []).
atc(close: 9,   close(S),                       exist(stream,S),                [closed_outstream(S, [])]).

% 8.11.7
atc(flush_o: 1, flush_output(S),                success,                        [pre((open('foo', write, S),
                                                                                      write(S, foo))),
                                                                                 checktxt(foo,'foo')]).
atc(flush_o: 2, flush_output(foo),              dom(stream_or_alias,foo),       []).
atc(flush_o: 3, flush_output(_S),               inst,                           []).
atc(flush_o: 4, flush_output(S),                exist(stream,S),                [closed_outstream(S, [])]).
atc(flush_o: 5, flush_output(S),           	perm(output,stream,S),          [maketxt(foo,''),
                                                                                 pre(open(foo,read,S))]).
atc(flush_o: 6, flush_output(st_o),             success,                        [txtout(st_o,'','')]).

% 8.11.8.4
stc(strprop: 1, stream_property(S,
				file_name(F)),	{[S <-- S1, F <-- @@],
                                                 [S <-- S2, F <-- @@]}, 	[maketxt(foo,''),
										 pre((open(foo, read, S1),
										      open(bar, write, S2))),
										 clean(bar)]).
stc(strprop: 2,  stream_property(S, output), 	{[S <-- FOut ],
                                                 [S <-- COut], ...}, 		[pre((open(bar, write, FOut),
										      current_output(COut))),
										 clean(bar)]).

atc(strprop: 3, stream_property(foo,_S), 	dom(stream,foo),		[]).
atc(strprop: 4, stream_property(_S,foo), 	dom(stream_property,foo),      	[]).
atc(strprop: 5, stream_property(S, P),          {[P <-- input],
						 [P <-- alias(user_input)],
						 [P <-- eof_action(reset)],
						 [P <-- mode(read)],
						 [P <-- reposition(false)],
						 [P <-- type(text)], ...},    	[pre(current_input(S))]).
						% 7.10.2.13, Table 40
atc(strprop: 6, stream_property(S, P),          {[P <-- output],
						 [P <-- alias(user_output)],
						 [P <-- eof_action(reset)],
						 [P <-- mode(append)],
						 [P <-- reposition(false)],
						 [P <-- type(text)], ...},     	[pre(current_output(S))]).
atc(strprop: 7, stream_property(_S,
				type(binary)),	failure,			[]).

atc(atendst: 8, at_end_of_stream(_S),  		inst,                   	[]).
atc(atendst: 9, at_end_of_stream(foo), 		dom(stream_or_alias,foo),	[]).
atc(atendst:10, at_end_of_stream(S), 		exist(stream,S),     		[closed_outstream(S,[])]).
atc(atendst:11, at_end_of_stream(st_i), 	success,	     		[txtin(st_i,'','')]).
atc(atendst:12, at_end_of_stream(st_i), 	failure,	     		[txtin(st_i,'a','a')]).
atc(atendst:13, at_end_of_stream(st_i), 	success,	     		[binin(st_i,[],[])]).
atc(atendst:14, at_end_of_stream(st_i), 	failure,	     		[binin(st_i,[0],[0])]).

% 8.11.9
atc(setstrp: 1, set_stream_position(_S,Pos), 	inst,                   	[stream_position(Pos)]).
atc(setstrp: 2, set_stream_position(S,_Pos), 	inst,                   	[pre(current_input(S))]).
atc(setstrp: 3, set_stream_position(foo,Pos), 	dom(stream_or_alias,foo), 	[stream_position(Pos)]).
atc(setstrp: 4, set_stream_position(S,Pos), 	exist(stream,S), 		[stream_position(Pos),
										 closed_outstream(S,[])]).
atc(setstrp: 5, set_stream_position(S,foo), 	dom(stream_position,foo), 	[pre(current_input(S))]).
atc(setstrp: 6, set_stream_position(S,Pos), 	perm(reposition,stream,S), 	[pre((open(foo,write,FS),
										  stream_property(FS,position(Pos)),
										  current_input(S))),
										 clean(foo)]).
% missing: set_stream_position/2 functionality test


% 8.12.1.4
stc(getchar: 1, get_char(Char),        		[[Char <-- 'q']],  		[txtin('qwerty','werty')]).
stc(getcode: 2, get_code(Code),        		[[Code <-- 0'q]],  		[txtin('qwerty', 'werty')]).
stc(getchar: 3, get_char(st_i, Char),  		[[Char <-- 'q']],  		[txtin(st_i, 'qwerty','werty')]).
stc(getcode: 4, get_code(st_i, Code),  		[[Code <-- 0'q]],  		[txtin(st_i, 'qwerty','werty')]).
stc(getchar: 5, get_char(st_i, Char),  		[[Char <-- '''']], 		[txtin(st_i,
										      "'qwerty'","qwerty'")]).
% vvv pts: swi-prolog
%stc(getcode: 6, get_code(st_i, Code),  		[[Code <-- 0'\']],  		[txtin(st_i,
stc(getcode: 6, get_code(st_i, Code),  		[[Code <-- 39]],  		[txtin(st_i,
										      "'qwerty'","qwerty'")]).
stc(getchar: 7, get_char(st_i, p),     		failure,           		[txtin(st_i, 'qwerty','werty')]).
stc(getcode: 8, get_code(st_i, 0'p),   		failure,           		[txtin(st_i, 'qwerty','werty')]).
stc(getchar: 9, get_char(st_i, Char),  		[[Char <-- end_of_file]], 	[txtin(st_i,'',{past})]).
stc(getcode:10, get_code(st_i, Code),  		[[Code <-- -1]], 		[txtin(st_i,'',{past})]).
stc(getchar:11, get_char(user_output, _), 	perm(input,stream,user_output), []).
stc(getcode:12, get_code(user_output, _), 	perm(input,stream,user_output), []).

atc(getchar:13, get_char(_, _),        		inst,                           []).
atc(getchar:14, get_char(1),           		type(in_character,1),   	[]).
atc(getchar:15, get_char(user_input,1),		type(in_character,1),   	[]).
atc(getchar:16, get_char(foo,_),       		dom(stream_or_alias,foo),	[]).
atc(getchar:17, get_char(S,_), 			exist(stream,S), 		[closed_outstream(S,[])]).
atc(getchar:18, get_char(S,_), 			perm(input,stream,S),		[pre(current_output(S))]).
atc(getchar:19, get_char(_), 			perm(input,binary_stream,S), 	[binin([]),
										 pre(current_input(S))]).
atc(getchar:20, get_char(_),			perm(input,past_end_of_stream,
						     S),		 	[txtin('',{past}),
										 pre(current_input(S)),
										 pre(get_char(_))]).
atc(getchar:21, (get_char(S, C1),
		 get_char(S, C2)), 		[[C1 <-- end_of_file,
						  C2 <-- end_of_file]], 	[maketxt(t, ''),
										 pre(open(t, read, S,
										      [eof_action(eof_code)]))]).
atc(getchar:22, get_char(S, _),	 		repr(character), 		[makebin(t,[0]),
										 pre(open(t, read, S))]).

atc(getcode:23, get_code(_, _),        		inst,                           []).
atc(getcode:24, get_code(p),           		type(integer,p),                []).
atc(getcode:25, get_code(user_input,p),		type(integer,p),                []).
atc(getcode:26, get_code(-2),          		repr(in_character_code),        []).
atc(getcode:27, get_code(foo,_),       		dom(stream_or_alias,foo),       []).
atc(getcode:28, get_code(S,_), 			exist(stream,S), 		[closed_instream(S,[])]).
atc(getcode:29, get_code(S,_), 			perm(input,stream,S),		[pre(current_output(S))]).
atc(getcode:30, get_code(_), 			perm(input,binary_stream,_S),	[binin([])]).
atc(getcode:31, get_code(_),			perm(input,past_end_of_stream,
						     S),		 	[txtin('',{past}),
										 pre(current_input(S)),
										 pre(get_code(_))]).
atc(getcode:32, (get_code(S, C1),
		 get_code(S, C2)), 		[[C1 <-- -1,
						  C2 <-- -1]], 			[maketxt(t, ''),
										 pre(open(t, read, S,
										      [eof_action(eof_code)]))]).
atc(getcode:33, get_code(S, _),	 		repr(character), 		[makebin(t,[0]),
										 pre(open(t, read, S))]).

% 8.12.2.4
stc(peekchr: 1, peek_char(Char),      		[[Char <-- 'q']],  		[txtin('qwerty', 'qwerty')]).
stc(peekcde: 2, peek_code(Code),      		[[Code <-- 0'q]],  		[txtin('qwerty', 'qwerty')]).
stc(peekchr: 3, peek_char(st_i, Char),		[[Char <-- 'q']],  		[txtin(st_i, 'qwerty','qwerty')]).
stc(peekcde: 4, peek_code(st_i, Code),		[[Code <-- 0'q]],  		[txtin(st_i, 'qwerty','qwerty')]).
stc(peekchr: 5, peek_char(st_i, Char),		[[Char <-- '''']], 		[txtin(st_i,
										      "'qwerty'","'qwerty'")]).
% vvv pts: swi-prolog
%stc(peekcde: 6, peek_code(st_i, Code),		[[Code <-- 0'\']],  		[txtin(st_i,
stc(peekcde: 6, peek_code(st_i, Code),		[[Code <-- 39]],  		[txtin(st_i,
										      "'qwerty'","'qwerty'")]).
stc(peekchr: 7, peek_char(st_i, p),   		failure,           		[txtin(st_i, 'qwerty','qwerty')]).
stc(peekcde: 8, peek_code(st_i, 0'p), 		failure,           		[txtin(st_i, 'qwerty','qwerty')]).
stc(peekchr: 9, peek_char(st_i, Char), 		[[Char <-- end_of_file]], 	[txtin(st_i,'','')]).
stc(peekcde:10, peek_code(st_i, Code), 		[[Code <-- -1]], 		[txtin(st_i,'','')]).
stc(peekchr:11, peek_char(s, _Char), 		perm(input,past_end_of_stream,
						     _S), 			[txtin(s, '', {past}),
										 pre(get_char(s, _))]).
stc(peekchr:12, peek_char(user_output, _), 	perm(input,stream,user_output), []).
stc(peekcde:13, peek_code(user_output, _), 	perm(input,stream,user_output), []).

atc(peekchr:14, peek_char(_, _),      		inst,                           []).
atc(peekchr:15, peek_char(1),         		type(in_character,1),   	[]).
atc(peekchr:16, peek_char(user_input,1),	type(in_character,1),         	[]).
atc(peekchr:17, peek_char(foo,_),     		dom(stream_or_alias,foo),	[]).
atc(peekchr:18, peek_char(S,_), 		exist(stream,S), 		[closed_instream(S,[])]).
atc(peekchr:19, peek_char(S,_), 		perm(input,stream,S),		[pre(current_output(S))]).
atc(peekchr:20, peek_char(s,_), 		perm(input,binary_stream,s), 	[binin(s, [], [])]).
atc(peekchr:21, (peek_char(S, C1),
		 peek_char(S, C1),
		 peek_char(S, C2)), 		[[C1 <-- end_of_file,
						  C2 <-- end_of_file]], 	[maketxt(t, ''),
										 pre(open(t, read, S))]).
atc(peekchr:22, peek_char(S, _), 		repr(character), 		[makebin(t, [0]),
										 pre(open(t, read, S))]).

atc(peekcde:23, peek_code(_, _),      		inst,                           []).
atc(peekcde:24, peek_code(p),         		type(integer,p),                []).
atc(peekcde:25, peek_code(user_input,p),	type(integer,p),      		[]).
atc(peekcde:26, peek_code(-2),        		repr(in_character_code), 	[]).
atc(peekcde:27, peek_code(foo,_),     		dom(stream_or_alias,foo),	[]).
atc(peekcde:28, peek_code(S,_), 		exist(stream,S), 		[closed_instream(S,[])]).
atc(peekcde:29, peek_code(S,_), 		perm(input,stream,S),		[pre(current_output(S))]).
atc(peekcde:30, peek_code(S,_), 		perm(input,binary_stream, S), 	[binin([]), pre(current_input(S))]).
atc(peekcde:31, peek_code(_), 			perm(input,past_end_of_stream,
						     S), 			[txtin('',{past}),
										 pre(current_input(S)),
										 pre(get_code(_))]).
atc(peekcde:32, (peek_code(C1), peek_code(C2)),	[[C1 <-- -1, C2 <-- -1]], 	[txtin('')]).
atc(peekcde:33, peek_code(S, _), 		repr(character), 		[makebin(t, [0]),
										 pre(open(t, read, S))]).

% 8.12.3.4
stc(putchar: 1, put_char(t),   			success,        		[txtout('qwer','qwert')]).
stc(putchar: 2, put_char(st_o, 'A'),		success,   			[txtout(st_o,'qwer','qwerA')]).
stc(putcode: 3, put_code(0't), 			success,        		[txtout('qwer','qwert')]).
stc(putcode: 4, put_code(st_o, 0't),		success,   			[txtout(st_o,'qwer','qwert')]).
stc(putchar: 5, (nl, put_char(a)),		success,     			[txtout('qwer','qwer\na')]).
stc(putchar: 6, (nl(st_o), put_char(st_o, a)), 	success,			[txtout(st_o,'qwer', 'qwer\na')]).
stc(putchar: 7, put_char(my_file, _C), 		inst, 				[txtout(my_file,'','')]).
stc(putchar: 8, put_char(st_o, 'ty'), 		type(character, ty), 		[txtout(st_o,'','')]).
stc(putcode: 9, put_code(my_file, _C), 		inst, 				[txtout(my_file,'','')]).
stc(putcode:10, put_code(st_o, 'ty'), 		type(integer, ty), 		[txtout(st_o,'','')]).
stc(nl:     11, nl(_Sr),       			inst,            		[]).
stc(nl:     12, nl(user_input),			perm(output,stream,user_input), []).
atc(putchar:13, put_char(_S,t),			inst,            		[]).
atc(putchar:14, put_char(_C),  			inst,				[]).
atc(putchar:15, put_char(S,a), 			exist(stream,S), 		[closed_outstream(S,[])]).
atc(putchar:16, put_char(S,a), 			perm(output,stream,S),		[pre(current_input(S))]).
atc(putchar:17, put_char(a), 			perm(output,binary_stream,S), 	[binout([]),
										 pre(current_output(S))]).
atc(putcode:18, put_code(_S,0't),      		inst,            		[]).
atc(putcode:19, put_code(_C),          		inst,            		[]).
atc(putcode:20, put_code(S,0'a), 		exist(stream,S), 		[closed_outstream(S,[])]).
atc(putcode:21, put_code(S,0'a), 		perm(output,stream,S),		[pre(current_input(S))]).
atc(putcode:22, put_code(S,0'a), 		perm(output,binary_stream,S), 	[pre(open(t,write,S,
											  [type(binary)])),
										 clean(t)]).
atc(putcode:23, put_code(-1),  			repr(character_code), 		[]).
atc(putcode:24, put_code(foo,1),       		dom(stream_or_alias,foo),	[]).

% 8.13.1.4
stc(getbyte: 1, get_byte(Byte),			[[Byte <-- 113]],  		[binin([113,119,101,114],
										       [119,101,114])]).
stc(getbyte: 2, get_byte(st_i, Byte), 		[[Byte <-- 113]],  		[binin(st_i, [113,119,101,114],
										       [119,101,114])]).
stc(getbyte: 3, get_byte(st_i, 114), 		failure,           		[binin(st_i,
										       [113,119,101,114,116,121],
										       [119,101,114,116,121])]).
stc(getbyte: 4, get_byte(st_i, Byte), 		[[Byte <-- -1]],   		[binin(st_i, [], {past})]).
stc(getbyte: 5, get_byte(user_output, _Byte), 	perm(input,stream,user_output), []).
atc(getbyte: 6, get_byte(_, _),        		inst,                           []).
atc(getbyte: 7, get_byte(p),      		type(in_byte,p), 		[binin([])]).
atc(getbyte: 8, get_byte(-2),      		type(in_byte,-2), 		[binin([])]).
atc(getbyte: 9, get_byte(foo,_),       		dom(stream_or_alias,foo),	[]).
atc(getbyte:10, get_byte(S,_), 			exist(stream,S), 		[closed_instream(S,
												 [type(binary)])]).
atc(getbyte:11, get_byte(S,_), 			perm(input,stream,S), 		[pre(current_output(S))]).
atc(getbyte:12, get_byte(_), 			perm(input,text_stream,S), 	[txtin(''),
										 pre(current_input(S))]).
atc(getbyte:13, (get_byte(_), get_byte(_)), 	perm(input,past_end_of_stream,
						     S), 			[binin([], {past}),
										 pre(current_input(S))]).

% 8.13.2.4
stc(peekbte: 1, peek_byte(Byte),      		[[Byte <-- 113]],	 	[binin([113,119,101,114],
										       [113,119,101,114])]).
stc(peekbte: 2, peek_byte(st_i, Byte),		[[Byte <-- 113]], 		[binin(st_i,
										       [113,119,101,114],
										       [113,119,101,114])]).
stc(peekbte: 3, peek_byte(st_i, 114), 		failure, 			[binin(st_i,
										       [113,119,101,114],
										       [113,119,101,114])]).
stc(peekbte: 4, peek_byte(st_i, Byte),		[[Byte <-- -1]], 		[binin(st_i, [], [])]).
stc(peekbte: 5, peek_byte(user_output, _Byte), 	perm(input,stream,user_output), []).
atc(peekbte: 6, peek_byte(_, _),      		inst,                           []).
atc(peekbte: 7, peek_byte(p),     		type(in_byte,p), 		[binin([])]).
atc(peekbte: 8, peek_byte(-2),     		type(in_byte,-2), 		[binin([])]).
atc(peekbte: 9, peek_byte(foo,_),     		dom(stream_or_alias,foo),	[]).
atc(peekbte:10, peek_byte(S,_), 		exist(stream,S), 		[closed_instream(S,
												 [type(binary)])]).
atc(peekbte:11, peek_byte(S,_),			perm(input,stream,S),		[pre(current_output(S))]).
atc(peekbte:12, peek_byte(_),	 		perm(input,text_stream,S), 	[txtin(''),
										 pre(current_input(S))]).
atc(peekbte:13, (get_byte(_), peek_byte(_)),	perm(input,
						     past_end_of_stream, S),	[binin([],{past}),
										 pre(current_input(S))]).

% 8.13.3.4
stc(putbyte: 1, put_byte(84),         		success,			variant(literally_iso, [
									          no  - [binout([113,119,101,114],
											[113,119,101,114,84])],
										  yes - [binout([113,119,101,114],
											[113,119,101,114,116])]
												       ])).
stc(putbyte: 2, put_byte(st_o, 116),   		success,			[binout(st_o, [113,119,101,114],
											[113,119,101,114,116])]).
stc(putbyte: 3, put_byte(my_file, _C), 		inst, 				[binout(my_file,[],[])]).

stc(putbyte: 4, put_byte(my_file, 'ty'), 	type(byte, ty), 		[binout(my_file,[],[])]).
                % ISO: user_output
atc(putbyte: 5, put_byte(_S, 118),     		inst,            		[]).
atc(putbyte: 6, put_byte(_C), 			inst,            		[binout([])]).
atc(putbyte: 7, put_byte(S, 77), 		exist(stream,S), 		[closed_outstream(S,
										[type(binary)])]).
atc(putbyte: 8, put_byte(S, 99), 		perm(output,stream,S), 		[binin([]),
										 pre(current_input(S))]).
atc(putbyte: 9, put_byte(99),		 	perm(output,text_stream,S), 	[pre(current_output(S))]).
atc(putbyte:10, put_byte(-1), 			type(byte, -1), 		[binout([])]).
atc(putbyte:11, put_byte(_S,1),        		inst,            		[]).
atc(putbyte:12, put_byte(foo,1),       		dom(stream_or_alias, foo), 	[]).

% 8.14.1.4

stc(read: 1,    read(T),			[[T <-- term1]],		[txtin('term1. term2. ...',
										       ' term2. ...')]).
stc(read: 2,    read(st_o, term1),		success,			[txtin(st_o,'term1. term2. ...',
										       ' term2. ...')]).
stc(read: 3,    read_term(st_o, T,
		[variables(VL),
		 variable_names(VN),
		 singletons(VS)]),		[[ T <-- foo(X1+X2,X1+X3),
						   VL <-- [X1,X2,X3],
						   VN <-- ['A'=X1,'Roger'=X2],
						   VS <-- ['Roger'=X2] ]],	[txtin(st_o,
										       ['foo(A+Roger,A+_). ',
											'term2. ...'],
										       ' term2. ...')]).
stc(read: 4,    read(4.1),			failure,			[txtin('3.1.  term2. ...',
										       '  term2. ...')]).
stc(read: 5,    read(_T),			syntax,				[txtin('foo 123. term2. ...',
										       ' term2. ...')]).
%stc(read: 6,    read(_T),			syntax,				[txtin('3.1',{past})]).
stc(read: 6,    read(_T),			syntax,				[txtin('3.1','')]).

atc(read: 7,    read_term(T, [singletons(S)]),	[[ T <-- foo(bar), S <-- [] ]],	[txtin('foo( bar).')]).
atc(read: 8,    read(_,_),			inst,				[]).
atc(read: 9,    read_term(user_input,_,_),	inst,				[]).
atc(read:10,    read_term(user_input,_,
			  [variables(_)|_]),	inst,				[]).
atc(read:11,    read_term(user_input,_,
			  [variables(_),_]),	inst,				[]).
atc(read:12,    read(foo,_),			dom(stream_or_alias,foo),	[]).
atc(read:13,    read_term(user_input,_,bar),	type(list,bar),			[]).
atc(read:14,    read_term(user_input,_,[bar]),	dom(read_option,bar),	[]).
atc(read:15,    read_term(user_output,_,[]),	perm(input,stream,user_output),	[]).
atc(read:16,    read(T),			[[ T <-- end_of_file ]],	[txtin('',{past})]).
atc(read:17,    read_term(S,_,[]), 		exist(stream,S), 		[closed_instream(S,[])]).
atc(read:18,    read_term(_,[]), 		perm(input,binary_stream,S), 	[binin([]),
										 pre(current_input(S))]).
atc(read:19,    read(_),	 		perm(input,binary_stream,S), 	[binin([]),
										 pre(current_input(S))]).
atc(read:20,    read_term(S,_,[]),		perm(input,past_end_of_stream,
						     S),		 	[txtin('',{past}),
										 pre(current_input(S)),
										 pre(get_code(_))]).
atc(read:21,    read_term(_,[]),		repr(max_arity),	  	[txtin_foo256]).
%atc(read:22,    read_term(_,[]),		syntax,			  	[txtin("'a.", {past})]).
atc(read:22,    read_term(_,[]),		syntax,			  	[txtin("'a.", '')]).
atc(read:23,    read(X),                        [[X <-- M]],                    [txtin_maxmin(max_integer, M)]).
atc(read:24,    read(X),                        [[X <-- M]],                    [txtin_maxmin(min_integer, M)]).

% TODO: read(X), get_code(C) with input a.%

% 8.14.2.4
stc(write: 1,   write_term(S, [1,2,3], []),	success,			[txtout('[1,2,3]'),
										 pre(current_output(S))]).
% stc(write: 2,	write_canonical([1,2,3]),	success,			[txtout('.(1,.(2,.(3,[])))')]).
stc(write: 2,   write_canonical([1,2,3]),	success,			[txtout(rx(
										   ["\\(\\.\\|'\\.'\\)",
										    "(1,\\1(2,\\1(3,\\[\\])))"]))]).
										% this nice regexp is just to
										% allow . to be written as '.'
stc(write: 3,   write_term(S, '1<2', []),	success,			[txtout('1<2'),
										 pre(current_output(S))]).
stc(write: 4,   writeq(S, '1<2'),		success,			[txtout("'1<2'"),
										 pre(current_output(S))]).
stc(write: 5,   writeq('$VAR'(0)),		success,			[txtout('A')]).
stc(write: 6,   write_term(S,'$VAR'(1),
		          [numbervars(false)]),	success,			[txtout('$VAR(1)'),
										 pre(current_output(S))]).
stc(write: 7,   write_term(S,'$VAR'(51),
			   [numbervars(true)]),	success,			[txtout('Z1'),
										 pre(current_output(S))]).

atc(write: 8,   write(_S, foo),			inst,				[]).
atc(write: 9,   write_term(foo, _Opts),		inst,				[]).
atc(write:10,   write_term(user_output,
			     foo, _Opts),	inst,				[]).
atc(write:11,   write_term(foo,
	    	   [quoted(true)|_Opts]),	inst,				[]).
atc(write:12,   write_term(user_output, foo,
	    	   [quoted(true)|_Opts]),	inst,				[]).
atc(write:13,   write_term(foo,
	    	   [quoted(true),_Opts]),	inst,				[]).
atc(write:14,   write_term(user_output, foo,
	    	   [quoted(true),_Opts]),	inst,				[]).
atc(write:15,   write_term(user_output, 1, 2),	type(list,2),			[]).
atc(write:16,   write_term(1, [quoted(true)
	   		      |foo]),		type(list,[quoted(true)|foo]),	[]).
atc(write:17,	write(foo, 1),			dom(stream_or_alias,foo),	[]).
atc(write:18,   write_term(1, [quoted(true),
	   		      foo]),		dom(write_option,foo),		[]).
atc(write:19,   write(S,a), 			exist(stream,S), 		[closed_outstream(S,[])]).
atc(write:20,   write(S,a), 			perm(output,stream,S),		[pre(current_input(S))]).
atc(write:21,   write(a), 			perm(output,binary_stream,S), 	[binout([]),
										 pre(current_output(S))]).

% 8.14.3.1

stc(op: 1,	op(30, xfy, ++),		success,			[post(current_op(30,xfy,++)),
										 post(op(0,xfy,++))]).
stc(op: 2,	op(0 , xfy, ++),		success,			[pre(op(30,xfy,++)),
										 post(\+current_op(_,_,++))]).
stc(op: 3,	op(max, xfy, ++),		type(integer,max),		[]).
stc(op: 4,	op(-30, xfy, ++),		dom(operator_priority,-30),	[]).
stc(op: 5,	op(1201, xfy, ++),		dom(operator_priority,1201),	[]).
stc(op: 6,	op(30, _XFY, ++),		inst,				[]).
stc(op: 7,	op(30, yfy, ++),		dom(operator_specifier,yfy),	[]).
stc(op: 8,	op(30, xfy, 0),			type(list,0),			[]).
stc(op: 9,	(op(30, xfy, ++),
		 op(40, xfx, ++)),		success,			[post(current_op(40,xfx,++)),
										 post(op(0,xfx,++))]).
stc(op:10,	(op(30, xfy, ++),
	 	 op(50, yf, ++)),		perm(create,operator,++),	[post(current_op(30,xfy,++)),
										 post(op(0,xfy,++))]) :-
	\+ context_info(postfix_and_infix_op, allowed).
atc(op:11,	op(_,xfx,++),			inst,				[]).
atc(op:12,	op(100,xfx,_),			inst,				[]).
atc(op:13,	op(100,xfx,[a|_]),		inst,				[]).
atc(op:14,	op(100,xfx,[a,_]),		inst,				[]).
atc(op:15,	op(100,200,[a]),		type(atom,200),			[]).
atc(op:16,	op(100,f(1),[a]),		type(atom,f(1)),		[]).
atc(op:17,	op(100,xfx,[a,a+b]),		type(atom,a+b),			[]).
atc(op:18,	op(100,xfx,','),		perm(modify,operator,','),	[]).
atc(op:19,	op(100,xfx,[a,',']),		perm(modify,operator,','),	[]).

% 8.14.4.4

stc(currop: 1,	current_op(P, xfy, OP),		{[ P <-- 1100, OP <-- (';')],
						 [ P <-- 1050, OP <-- ('->')],
						 [ P <-- 1000, OP <-- (',')],
						 [ P <-- 200, OP <-- ('^')],
						 ...},				[]).

atc(currop: 2,	current_op(1201, _, _), 	dom(operator_priority,1201),	[]).
atc(currop: 3,	current_op(_, yfy, _), 		dom(operator_specifier,yfy),	[]).
atc(currop: 4,	current_op(_, 0, _), 		type(atom,0),                   []).
atc(currop: 5,	current_op(_, _, 5), 		type(atom,5),			[]).

% 8.14.5.4

stc(charcnv: 1, char_conversion('&', ','),	[[(read((a,b)), get_char(' '),
						    get_char('&'))]],		[txtin('a&b. &'),clear_chconv]).
stc(charcnv: 2, char_conversion('^', '\''),	[[(read('b+c'), get_char(' '),
						   get_char('^'))]],		[txtin('^b+c^. ^'),clear_chconv]).
stc(charcnv: 3, char_conversion('A', 'a'),	[[read('A+c'+a)]],		[txtin("'A+c'+A."),clear_chconv]).
stc(charcnv: 4, (read(X), read(Y), read(Z)),	[[ X <-- (a,a), Y <-- 'AAA',
						   Z <-- 'a,a' ]],		[txtin("A&A. 'AAA'. ^A&A^."),
										 pre((char_conversion('&', ','),
										     char_conversion('^', '\''),
										     char_conversion('A', 'a'))),
										 clear_chconv]).
stc(charcnv: 5,	char_conversion('&', '&'),	[[(read('&'),
						  \+ current_char_conversion(
								_,_))]],	[txtin("& ."),
										 pre(char_conversion('&', ','))]).

atc(charcnv: 6, read(X),			[[ X <-- (0'%)+1 ]],		[txtin("0'%%1."),
										 some_chconvs]).
atc(charcnv: 7, read(X),			[[ X <-- '%'+1 ]],		[txtin("'%'%1."),
										 some_chconvs]).
atc(charcnv: 8, read(X),			[[ X <-- "%"+1 ]],		[txtin('"%"%1.'),
										 some_chconvs]).
atc(charcnv: 9, read(X),			[[ X <-- '.'(1,!) ]],		[txtin('1.#.'),
										 pre(op(100,xfx,'.')),
										 post(op(0,xfx,'.')),
										 pre(char_conversion('#','!')),
										 clear_chconv]).
atc(charcnv:10, read(X),			[[ X <--  'aa'+'bb^']],		[txtin("^aa'+'bb^'."),
										 some_chconvs]).
atc(charcnv:11, (set_prolog_flag(
		     char_conversion, off),
		 read(X), set_prolog_flag(
		     char_conversion, on),
		 read(Y)),			[[ X <--  (+), Y <-- (+)]],	[txtin("+ .% ."),
										 some_chconvs]).
atc(charcnv:12, (read(X), read(Y)),		[[ X <--  '-'('.+'),
						   Y <-- end_of_file]],		[txtin("- .% .",{past}),
										 some_chconvs]).

% 8.14.6.4
stc(crchcnv: 1, current_char_conversion(C, a),	{[ C <-- 'A'],
						 [ C <-- Aacute]},		[txtin("'\\341\\'."),
										 pre((read(Aacute),
						 				  char_conversion('A', 'a'),
										  char_conversion(Aacute, 'a'))),
										 clear_chconv]).

% 8.15.1.4
stc(not: 1, 	'\\+'(true),                    failure,                        []).
stc(not: 2, 	'\\+'(!),                       failure,                        []).
stc(not: 3, 	'\\+'((!,fail)),                success,                        []).
stc(not: 4, 	((X=1;X=2),'\\+'((!,fail))),    [[X <-- 1],
            	                                 [X <-- 2]],                    []).
stc(not: 5, 	'\\+'(4 = 5),                   success,                        []).
stc(not: 6, 	'\\+'(3),                       type(callable,3),               []).
stc(not: 7, 	'\\+'(_X),                      inst,                           []).
stc(not: 8, 	'\\+'(X=f(X)),                  sto(failure),                   []).

% 8.15.2.4
stc(once: 1, 	once(!),                        success,                        []).
stc(once: 2, 	(once(!), (X=1; X=2)),          [[X <-- 1],[X <-- 2]],          []).
stc(once: 3, 	once(repeat),                   success,                        []).
stc(once: 4, 	once(fail),                     failure,                        []).
stc(once: 5, 	once((X = f(X))),               sto([[X <-- V@f(V)]]),          []).

etc(once: 6, 	once(3),                        type(callable,3/*once(3)*/),    []).
etc(once: 7, 	once(_X),                       inst,                           []).


% 8.15.3.4
stc(repeat: 1, (repeat, write('hello '), fail), time_out,			[txtout(rx('hello hello .*'))]).
stc(repeat: 2, (repeat, !, fail),               failure,                	[]).

% 8.16.1.4
stc(atomlen: 1, atom_length(
		      'enchanted evening', N), 	[[N <-- 17]],           	[]).
stc(atomlen: 2, atom_length(
		      'enchanted\
 evening', N), 					[[N <-- 17]],           	[]).
stc(atomlen: 3, atom_length('', N),         	[[N <-- 0]],            	[]).
stc(atomlen: 4, atom_length('scarlet', 5),  	failure,                	[]).
stc(atomlen: 5, atom_length(_Atom, 4),      	inst,                   	[]).
stc(atomlen: 6, atom_length(1.23, 4),       	type(atom, 1.23),       	[]).
stc(atomlen: 7, atom_length(atom, '4'),     	type(integer, '4'),		[]).

etc(atomlen: 8, atom_length(atom, -4),      	dom(not_less_than_zero,-4),	[]).
atc(atomlen: 9, atom_length('Bartk Bla', L),  [[L <-- 11]],                   []).
atc(atomlen:10, atom_length(_Atom, N),          inst,                   	[]) :-
	context_info(max_atom_length, Max), N is Max+1.
atc(atomlen:11, atom_length(atom, N),           failure,                   	[]) :-
	context_info(max_atom_length, Max), N is Max+1.

% 8.16.2.4
stc(atomcat: 1, atom_concat('hello',
			    ' world', S3), 	[[S3 <-- 'hello world']],	[]).
stc(atomcat: 2, atom_concat(T, ' world',
			    'small world'), 	[[T <-- 'small']],      	[]).
stc(atomcat: 3, atom_concat('hello',' world',
			    'small world'), 	failure,                	[]).
stc(atomcat: 4, atom_concat(T1, T2, 'hello'),	[[T1 <-- '',T2 <-- 'hello'],
                                                 [T1 <-- 'h',T2 <-- 'ello'],
                                                 [T1 <-- 'he',T2 <-- 'llo'],
                                                 [T1 <-- 'hel',T2 <-- 'lo'],
                                                 [T1 <-- 'hell',T2 <-- 'o'],
                                                 [T1 <-- 'hello',T2 <-- '']], 	[]).
stc(atomcat: 5, atom_concat(small, _V2, _V4), 	inst,                 		[]).

etc(atomcat: 6, atom_concat(_A, 'iso', _C), 	inst,                   	[]).
etc(atomcat: 7, atom_concat('iso', _B, _C), 	inst,                   	[]).
etc(atomcat: 8, atom_concat(f(a), 'iso', _C),	type(atom,f(a)),       		[]).
etc(atomcat: 9, atom_concat('iso', f(a), _C),	type(atom,f(a)),       		[]).
etc(atomcat:10, atom_concat(_A, _B, f(a)),  	type(atom,f(a)),        	[]).
atc(atomcat:11, atom_concat('Bartk ', 'Bla', N), [[N <-- 'Bartk Bla']],     []).
atc(atomcat:12, atom_concat(N, 'Bla', 'Bartk Bla'), [[N <-- 'Bartk ']],     []).
atc(atomcat:13, atom_concat('Bartk ', N, 'Bartk Bla'), [[N <-- 'Bla']],     []).
atc(atomcat:14, atom_concat(T1, T2, 'Pcs'),    [[T1 <-- '',T2 <-- 'Pcs'],
						 [T1 <-- 'P',T2 <-- 'cs'],
						 [T1 <-- 'P',T2 <-- 'cs'],
						 [T1 <-- 'Pc',T2 <-- 's'],
						 [T1 <-- 'Pcs',T2 <-- '']],    []).
% non-standard test
atc(atomcat:15, atom_concat(LA,LA,_),		repr(exceeded_atom_size),       [long_atom(LA)]) :-
	context_info(max_atom_length, _).

% 8.16.3.4
stc(subatom: 1, sub_atom(abracadabra,0,5,X,S2), [X^[S2 <-- 'abrac']],   	[]).
stc(subatom: 2, sub_atom(abracadabra,X,5,0,S2), [X^[S2 <-- 'dabra']],   	[]).
stc(subatom: 3, sub_atom(abracadabra,3,
			 Length,3,S2), 		[[Length <-- 5,
                                                  S2 <-- 'acada']],     	[]).
stc(subatom: 4, sub_atom(abracadabra,
			 Before, 2, After, ab), [[Before <-- 0, After <-- 9],
						 [Before <-- 7, After <-- 2]], 	[]).
stc(subatom: 5, sub_atom('Banana',3,2,X,S2), 	[X^[S2 <-- 'an']],      	[]).
stc(subatom: 6, sub_atom('charity',B,3,A,S2), 	[A^B^[S2 <-- 'cha'],
                                                 A^B^[S2 <-- 'har'],
                                                 A^B^[S2 <-- 'ari'],
                                                 A^B^[S2 <-- 'rit'],
                                                 A^B^[S2 <-- 'ity']],   	[]).
stc(subatom: 7, sub_atom('ab', St, Len, X, A), 	[X^[St<--0, Len<--0, A<--''],
						 X^[St<--0, Len<--1, A<--'a'],
						 X^[St<--0, Len<--2, A<--'ab'],
						 X^[St<--1, Len<--0, A<--''],
						 X^[St<--1, Len<--1, A<--'b'],
						 X^[St<--2, Len<--0, A<--'']],	[]).

etc(subatom: 8, sub_atom(_Banana, 3, 2, _, _S),	inst,                  		[]).
etc(subatom: 9, sub_atom(f(a), 2, 2, _, _S2),  	type(atom,f(a)),        	[]).
etc(subatom:10, sub_atom('Banana', 4, 2, _, 2),	type(atom,2),   		[]).
etc(subatom:11, sub_atom('Banana', a, 2, _, _),	type(integer,a),		[]).
etc(subatom:12, sub_atom('Banana', 4, n, _, _),	type(integer,n),		[]).
etc(subatom:13, sub_atom('Banana', 4, _, m, _),	type(integer,m),		[]).
atc(subatom:14, sub_atom('Banana', -2, 3, 4,_), dom(not_less_than_zero, -2),    []).
atc(subatom:15, sub_atom('Banana', 2, -3, 4,_), dom(not_less_than_zero, -3),    []).
atc(subatom:16, sub_atom('Banana', 2, 3, -4,_), dom(not_less_than_zero, -4),    []).
atc(subatom:17, sub_atom('Banana', 2, 3, A, 'nan'), [[A<--1]],                  []).
atc(subatom:18, sub_atom('Banana', B, 3, 1, 'nan'), [[B<--2]],                  []).
atc(subatom:19, sub_atom('Banana', 2, L, 1, 'nan'), [[L<--3]],                  []).
atc(subatom:20, sub_atom('Banana', 2, L, A, 'nan'), [[A<--1,L<--3]],            []).
atc(subatom:21, sub_atom('Banana', B, L, 1, 'nan'), [[B<--2,L<--3]],            []).
atc(subatom:22, sub_atom('Banana', 2, 3, 1, 'ana'), failure,                    []).
atc(subatom:23, sub_atom('Banana', 2, 3, 2, 'nan'), failure,                    []).
atc(subatom:24, sub_atom('Banana', 2, 3, 2, _), failure,                        []).
atc(subatom:25, sub_atom('Banana', 2, 3, 1, 'anan'), failure,                   []).
atc(subatom:26, sub_atom('Banana', 0, 7, 0, _), failure,                        []).
atc(subatom:27, sub_atom('Banana', 7, 0, 0, _), failure,                        []).
atc(subatom:28, sub_atom('Banana', 0, 0, 7, _), failure,                        []).
atc(subatom:29, sub_atom(a, N, 1, 2, 1+2), type(atom,1+2),              []) :-
	context_info(max_atom_length, Max), N is Max+1.
atc(subatom:30, sub_atom(a, N, 1, 2, 1+2),  type(atom,1+2),              []) :-
	context_info(max_atom_length, Max), N is Max+1.
atc(subatom:31, sub_atom('Bartk Bla', 4, 2, A, S), [[A <-- 5,S <-- 'k']],    []).
atc(subatom:32, sub_atom('Bartk Bla', 4, L, 5, S), [[L <-- 2,S <-- 'k']],    []).
atc(subatom:33, sub_atom('Bartk Bla', B, 2, 5, S), [[B <-- 4,S <-- 'k']],    []).
atc(subatom:34, sub_atom('Pcs',B,2,A,S),       [[B <-- 0,A <-- 2,S <-- 'P'],
						 [B <-- 1,A <-- 1,S <-- 'c'],
						 [B <-- 2,A <-- 0,S <-- 'cs']], []).
atc(subatom:35, sub_atom(abracadabra,B,L,A,abra), [[B <-- 0,L <-- 4,A <-- 7],
						   [B <-- 7,L <-- 4,A <-- 0]],  []).
atc(subatom:36, sub_atom(a, N, 1, 2, _), failure,			[]) :-
	context_info(max_atom_length, Max), N is Max+1.
atc(subatom:37, sub_atom(a, 1, N, 2, _), failure,			[]) :-
	context_info(max_atom_length, Max), N is Max+1.
atc(subatom:38, sub_atom(a, 1, 2, N, _), failure,			[]) :-
	context_info(max_atom_length, Max), N is Max+1.


% 8.16.4.4
stc(atomchs: 1, atom_chars('', L),           	[[L <-- []]],           	[]).
stc(atomchs: 2, atom_chars([], L),           	[[L <-- ['[',']']]],    	[]).
stc(atomchs: 3, atom_chars('''', L),         	[[L <-- ['''']]],       	[]).
stc(atomchs: 4, atom_chars('ant', L),        	[[L <-- ['a','n','t']]],	[]).
stc(atomchs: 5, atom_chars(Str, ['s','o','p']), [[Str <-- 'sop']],    		[]).
stc(atomchs: 6, atom_chars('North', ['N'|X]),	[[X <-- ['o','r','t','h']]],	[]).
stc(atomchs: 7, atom_chars('soap',
			   ['s','o','p']),	failure,           		[]).
stc(atomchs: 8, atom_chars(_X, _Y),          	inst,                   	[]).
etc(atomchs: 9, atom_chars(_A, [a,_E,c]),    	inst,                   	[]).
etc(atomchs:10, atom_chars(_A, [a,b|_L]),    	inst,                   	[]).
etc(atomchs:11, atom_chars(f(a), _L),        	type(atom,f(a)),        	[]).
etc(atomchs:12, atom_chars(_A, iso),         	type(list,iso), 		[]).
etc(atomchs:13, atom_chars(_A, [a,f(b)]),    	type(character,f(b)), 		[]).
atc(atomchs:14, atom_chars('Pcs', L),          [[L <-- ['P','','c','s']]],    []).
atc(atomchs:15, atom_chars(A,
			   ['P','','c','s']),  [[A <-- 'Pcs']],               []).

% 8.16.5.4
stc(atomcds: 1, atom_codes('', L),           	[[L <-- []]],           	[]).
stc(atomcds: 2, atom_codes([], L),           	[[L <-- [0'[, 0']]]],   	[]).
% vvv pts: swi-prolog
%stc(atomcds: 3, atom_codes('''',L),          	[[L <-- [0'\' ]]],      	[]).
stc(atomcds: 3, atom_codes('''',L),          	[[L <-- [39 ]]],      	[]).
stc(atomcds: 4, atom_codes('ant', L),        	[[L <-- [0'a, 0'n, 0't]]], 	[]).
stc(atomcds: 5, atom_codes(Str,
			   [0's,0'o,0'p]),   	[[Str <-- 'sop']],      	[]).
stc(atomcds: 6, atom_codes('North',[0'N|X]), 	[[X <-- [0'o,0'r,0't,0'h]]],	[]).
stc(atomcds: 7, atom_codes('soap',
			   [0's, 0'o, 0'p]), 	failure,       			[]).
stc(atomcds: 8, atom_codes(_X, _Y),          	inst,                   	[]).
etc(atomcds: 9, atom_codes(f(a), _L),        	type(atom,f(a)),        	[]).
etc(atomcds:10, atom_codes(_, 0'x),          	type(list,0'x), 		[]).
etc(atomcds:11, atom_codes(_A, [0'i,0's,-1]),	repr(character_code),      	[]).
atc(atomcds:12, atom_codes('Pcs', C),          [[C <-- [0'P,0',0'c,0's]]],    []).
atc(atomcds:13, atom_codes(A,
			   [0'P,0',0'c,0's]),  [[A <-- 'Pcs']],               []).
atc(atomcds:14, atom_codes(_A,
			   [N]),                repr(character_code),           []) :-
	context_info(max_char_code, Max), N is Max+1.
%atc(atomcds:15, atom_codes(A,
%	         [0'\x10000000\,0'a,0'b]), 	[[A <-- '\x10000000\ab']],      []).
% vvv pts: swi-prolog
%atc(atomcds:15, atom_codes(A,
%	         [0'\xFF\,0'a,0'b]), 	        [[A <-- '\xFF\ab']],            []).
atc(atomcds:16, atom_codes(_A, [a,b,c]),	(repr(character_code)
						;type(integer,@@)),          	[]).

% 8.16.6.4
stc(charcde: 1, char_code(a, Code),             [[Code <-- 0'a]],       	[]).
stc(charcde: 2, (char_code(Char, 99),
                   atom_codes(Atom, [99])),     [Ch^[Char <-- Ch,
                                                      Atom <-- Ch]],    	[]).
stc(charcde: 3, char_code(Char, 0'c),           [[Char <-- c]],         	[]).
stc(charcde: 4, char_code(Char, 163),           [[read(Char)]],    		[txtin("'\\243\\'.")]).
stc(charcde: 5, char_code(b, 0'b),              success,                	[]).
stc(charcde: 6, char_code('ab', _Code),         type(character, 'ab'),  	[]).
stc(charcde: 7, char_code(_Char, _Code),        inst,                   	[]).
etc(charcde: 8, char_code(a, x),                type(integer, x),       	[]).
etc(charcde: 9, char_code(_Char, -2),           repr(character_code),    	[]).
%atc(charcde:10, char_code(Char, 999),           [[read(Char)]],                 [txtin("'\\1747\\'.")]).       
%atc(charcde:11, char_code(Char, 0o177777),      [[read(Char)]],   		[txtin("'\\177777\\'.")]).     
%atc(charcde:12, char_code(Char, 0o7777777),     [[read(Char)]],   		[txtin("'\\7777777\\'.")]).    
%atc(charcde:13, char_code(Char, 0o377777777),   [[read(Char)]],   		[txtin("'\\377777777\\'.")]).  
%atc(charcde:14, char_code(Char, 0o17777777777), [[read(Char)]],   		[txtin("'\\17777777777\\'.")]).
atc(charcde:15, (char_code(Char, 225),
		    atom_chars(Char, [Y])),     [[Char <-- , Y <-- '\341\']],  []). 

% 8.16.7.4
stc(numbchs: 1, number_chars(33, L),       	[[L <-- ['3','3']]],    	[]).
stc(numbchs: 2, number_chars(33, ['3','3']),	success,	               	[]).
stc(numbchs: 3, number_chars(33.0, L),		[[number_chars(N, L), N=33.0]], []).
stc(numbchs: 4, number_chars(X,
		 ['3','.','3','E','+','0']),	[[X <-- 3.3]],          	[]).
stc(numbchs: 5, number_chars(3.3,  ['3'|_L]),	success,                	[]).
stc(numbchs: 6, number_chars(A, ['-','2','5']), [[A <-- -25]],          	[]).
stc(numbchs: 7, number_chars(A,['\n',' ','3']), [[A <-- 3]],            	[]).
stc(numbchs: 8, number_chars(_A, ['3',' ']),	syntax,        			[]).
stc(numbchs: 9, number_chars(A, ['0',x,f]),	[[A <-- 15]],           	[]).
stc(numbchs:10, number_chars(A, ['0','''',a]),  [[A <-- 0'a]],          	[]).
stc(numbchs:11, number_chars(A, ['4','.','2']), [[A <-- 4.2]],          	[]).
stc(numbchs:12, number_chars(A,
		['4','2','.','0','e','-','1']), [[A <-- 4.2]],          	[]).

etc(numbchs:13, number_chars(_A, _L),      	inst,                        	[]).
etc(numbchs:14, number_chars(a, _L),       	type(number, a),             	[]).
etc(numbchs:15, number_chars(_A, 4),       	type(list, 4),       		[]).
etc(numbchs:16, number_chars(_A, ['4',2]), 	type(character, 2),  		[]).

atc(numbchs:17, number_chars(_A, [a|_L]),  	inst,                        	[]).
atc(numbchs:18, number_chars(_A, [a,_L]),  	inst,                        	[]).

atc(numbchs:19, number_chars(X, [' ', '0', 'o', '1', '1']), [[X <-- 9]],        []).
atc(numbchs:20, number_chars(X, [' ', '0', 'x', '1', '1']), [[X <-- 17]],       []).
atc(numbchs:21, number_chars(X, [' ', '0', 'b', '1', '1']), [[X <-- 3]],        []).
atc(numbchs:22, number_chars(_X, ['0', 'o', '8']), syntax,                      []).
atc(numbchs:23, number_chars(_X, ['0', 'b', '2']), syntax,                      []).
atc(numbchs:24, number_chars(_X, ['0', 'x', 'g']), syntax,                      []).
atc(numbchs:25, number_chars(_X, ['']),        syntax,                         []).
atc(numbchs:26, number_chars(_X, ['a']),        syntax,                         []).
atc(numbchs:27, number_chars(_X,
		    ['0','x','0','.','0']),     syntax,                         []).

% 8.16.8.4
stc(numbcds: 1, number_codes(33, L),       	[[L <-- [0'3,0'3]]],    	[]).
stc(numbcds: 2, number_codes(33, [0'3,0'3]),	success,               		[]).
stc(numbcds: 3, number_codes(33.0, L),		[[number_codes(N, L), N=33.0]], []).
stc(numbcds: 4, number_codes(33.0, [0'3|_L]),	 success,                	[]).
stc(numbcds: 5, number_codes(A, [0'-,0'2,0'5]), [[A <-- -25]],          	[]).
stc(numbcds: 6, number_codes(A, [0' , 0'3]),    [[A <-- 3]],            	[]).
stc(numbcds: 7, number_codes(A, [0'0,0'x,0'f]), [[A <-- 15]],           	[]).
% vvv pts: swi-prolog
%stc(numbcds: 8, number_codes(A,[0'0,0'\',0'a]), [[A <-- 0'a]],          	[]).
stc(numbcds: 8, number_codes(A,[0'0,39,0'a]), [[A <-- 0'a]],          	[]).
stc(numbcds: 9, number_codes(A, [0'4,0'.,0'2]), [[A <-- 4.2]],          	[]).
stc(numbcds:10, number_codes(A, [
		 0'4,0'2,0'.,0'0,0'e,0'-,0'1]), [[A <-- 4.2]],          	[]).

etc(numbcds:11, number_codes(_A, _L),      	inst,                   	[]).
etc(numbcds:12, number_codes(a, _L),       	type(number,a), 		[]).
etc(numbcds:13, number_codes(_A, 4),       	type(list,4),   		[]).
etc(numbcds:14, number_codes(_A, [0'4,-1]),	repr(character_code),      	[]).

atc(numbcds:15, number_codes(_A, [0'a|_L]),	inst,                   	[]).
atc(numbcds:16, number_codes(_A, [0'a,_L]),	inst,                   	[]).

atc(numbcds:17, (number_chars(X, [' ', '0', 'x', '1', '1', '1']),
		    number_codes(X, Y)), [[X <-- 273, Y <-- [50,55,51]]],       []).
atc(numbcds:18, (number_chars(X, [' ', '0', 'o', '1', '1', '1']),
		    number_codes(X, Y)), [[X <-- 73, Y <-- [55,51]]],           []).
atc(numbcds:19, (number_chars(X, [' ', '0', 'b', '1', '1', '1']),
		    number_codes(X, Y)), [[X <-- 7, Y <-- [55]]],               []).
% vvv pts  !!
atc(numbcds:20, number_codes(N, "0'\\n"),       [[N <-- 10]],                   []).
atc(numbcds:21, number_codes(_N, ""),          syntax,                         []).
atc(numbcds:22, number_codes(_X,
		    [0'0,0'x,0'0,0'.,0'0]),     syntax,                         []).


% 8.17.1.4
stc(setpflg: 1, set_prolog_flag(unknown, fail), [[current_prolog_flag(unknown,
								      fail)]], 	[]).
stc(setpflg: 2, set_prolog_flag(_X, off),	inst,                  		[]).
stc(setpflg: 3, set_prolog_flag(5, decimals), 	type(atom,5),   		[]).
stc(setpflg: 4, set_prolog_flag(date,
				'July 1999'),   dom(prolog_flag,date),  	[]).
stc(setpflg: 5, set_prolog_flag(debug, trace),  dom(flag_value,debug+trace),    []).
etc(setpflg: 6, set_prolog_flag(max_arity, 40), perm(modify,flag,max_arity),	[]).

% 8.17.2.4
stc(curpflg: 1, current_prolog_flag(debug,off), success,                	[]).
stc(curpflg: 2, current_prolog_flag(F, V), 	{[F<--debug, V<--off],
						 [F<--char_conversion, V<--on],
						 [F<--unknown, V<--error],...}, []).
stc(curpflg: 3, current_prolog_flag(5, _V), 	type(atom,5),   		[]).

etc(curpflg: 4, set_prolog_flag(unknown,
				warning), 	[[current_prolog_flag(unknown,
								    warning)]], []).
etc(curpflg: 5, (set_prolog_flag(unknown,
				warning),
		 current_prolog_flag(
		           unknown,error)),	failure,                	[]).
etc(curpflg: 6, current_prolog_flag(debug, V),  [[V <-- off]],          	[]).
etc(curpflg: 7, current_prolog_flag(warning,_), dom(prolog_flag,warning), 	[]).
etc(curpflg: 8, current_prolog_flag(1+2, flag), type(atom,1+2), 		[]).

% missing: testing the effect of flag 'unknown'


% 8.17.3.4
% stc(halt: 1, 	halt,                           implementation_defined, 	[]).

% 8.17.4.4
% stc(halt1: 1, halt(1),                        implementation_defined, 	[]).
%stc(halt1: 2, 	halt(a),                        type(integer,a),        	[]). SWI
%etc(halt1: 3, 	halt(_),                        inst,                   	[]). SWI


% 9.1.7

stc(eval: 1,	X is '+'(7, 35), 		[[ X <-- 42 ]],			[]).
stc(eval: 2,	X is '+'(0, 3+11), 		[[ X <-- 14 ]],			[]).
stc(eval: 3,	X is '+'(0, 3.2+11), 		[[ X <-- 14.2 ]],		[]).
stc(eval: 4,	_X is '+'(77, _N), 		inst,				[]).
%stc(eval: 5,	_X is '+'(foo, 77), 		type(number,foo),		[]).
stc(eval: 5,	_X is '+'(foo, 77), 		type(evaluable,foo/0),		[]).
stc(eval: 6,	X is '-'(7), 			[[ X <-- -7 ]],			[]).
stc(eval: 7,	X is '-'(3-11), 		[[ X <-- 8 ]],			[]).
stc(eval: 8,	X is '-'(3.2-11), 		[[ X <-- 7.8 ]],		[]).
stc(eval: 9,	_X is '-'(_N),	 		inst,				[]).
%stc(eval:10,	_X is '-'(foo), 		type(number,foo),		[]).
stc(eval:10,	_X is '-'(foo), 		type(evaluable,foo/0),		[]).
stc(eval:11,	X is '-'(7,35), 		[[ X <-- -28 ]],		[]).
stc(eval:12,	X is '-'(20,3+11), 		[[ X <-- 6 ]],			[]).
stc(eval:13,	X is '-'(0,3.2+11), 		[[ X <-- -14.2 ]],		[]).
stc(eval:14,	_X is '-'(77, _N), 		inst,				[]).
%stc(eval:15,	_X is '-'(foo, 77), 		type(number,foo),		[]).
stc(eval:15,	_X is '-'(foo, 77), 		type(evaluable,foo/0),		[]).
stc(eval:16,	X is '*'(7,35), 		[[ X <-- 245 ]],		[]).
stc(eval:17,	X is '*'(0,3+11), 		[[ X <-- 0 ]],			[]).
stc(eval:18,	X is '*'(1.5,3.2+11), 		[[ X =~~ 21.3 ]],		[]).

stc(eval:19,	_X is '*'(77, _N), 		inst,				[]).
%stc(eval:20,	_X is '*'(foo, 77), 		type(number,foo),		[]).
stc(eval:20,	_X is '*'(foo, 77), 		type(evaluable,foo/0),		[]).
%stc(eval:21,	X is '/'(7,35), 		[[ X <-- 0 ]],			[]).
stc(eval:21,	X is '//'(7,35), 		[[ X <-- 0 ]],			[]).
stc(eval:22,	X is '/'(7.0,35), 		[[ X <-- 0.2 ]],		[]).
%stc(eval:23,	X is '/'(140,3+11), 		[[ X <-- 10 ]],			[]).
stc(eval:23,	X is '//'(140,3+11), 		[[ X <-- 10 ]],			[]).
%stc(eval:24,	X is '/'(20.164,3.2+11),	[[ X =~~ 14.2 ]],		[]).
stc(eval:24,	X is '/'(20.164,3.2+11),	[[ X =~~ 1.42 ]],		[]).
%stc(eval:25,	_X is '/'(7,-3), 		undefined,			[]).
stc(eval:25,	X is '//'(7,-3), 		[[X <-- (-2;-3) ]],		[]).
%stc(eval:26,	_X is '/'(-7,3), 		undefined,			[]).
stc(eval:26,	X is '//'(-7,3), 		[[X <-- (-2;-3) ]],		[]).
stc(eval:27,	_X is '/'(77, _N), 		inst,				[]).
%stc(eval:28,	_X is '/'(foo, 77), 		type(number,foo),		[]).
stc(eval:28,	_X is '/'(foo, 77), 		type(evaluable,foo/0),		[]).
stc(eval:29,	_X is '/'(3, 0), 		eval(zero_divisor),		[]).
stc(eval:30,	X is mod(7,3),	 		[[ X <-- 1 ]],			[]).
stc(eval:31,	X is mod(0,3+11), 		[[ X <-- 0 ]],			[]).
stc(eval:32,	X is mod(7,-2),	 		[[ X <-- -1 ]],			[]).
stc(eval:33,	_X is mod(77, _N), 		inst,				[]).
%stc(eval:34,	_X is mod(foo, 77), 		type(number,foo),		[]).
stc(eval:34,	_X is mod(foo, 77), 		type(evaluable,foo/0),		[]).
stc(eval:35,	_X is mod(7.5, 2), 		type(integer,7.5),		[]).
stc(eval:36,	_X is mod(7, 0), 		eval(zero_divisor),		[]).
stc(eval:37,	X is floor(7.4), 		[[ X <-- 7 ]],			[]).
stc(eval:38,	X is floor(-0.4), 		[[ X <-- -1 ]],			[]).
stc(eval:39,	X is round(7.5), 		[[ X <-- 8 ]],			[]).
stc(eval:40,	X is round(7.6), 		[[ X <-- 8 ]],			[]).
stc(eval:41,	X is round(-0.6), 		[[ X <-- -1 ]],			[]).

stc(eval:42,	_X is round(_N), 		inst,				[]).
stc(eval:43,	X is ceiling(-0.5), 		[[ X <-- 0 ]],			[]).
stc(eval:44,	X is truncate(-0.5), 		[[ X <-- 0 ]],			[]).
%stc(eval:45,	_X is truncate(foo), 		type(number,foo),		[]).
stc(eval:45,	_X is truncate(foo), 		type(evaluable,foo/0),		[]).
stc(eval:46,	X is float(7), 			[[ X <-- 7.0 ]],		[]).
stc(eval:47,	X is float(7.3), 		[[ X <-- 7.3 ]],		[]).
%stc(eval:48,	X is float(5/3), 		[[ X <-- 1.0 ]],		[]).
stc(eval:48,	X is float(5//3), 		[[ X <-- 1.0 ]],		[]).
stc(eval:49,	_X is float(_N), 		inst,				[]).
%stc(eval:50,	_X is float(foo), 		type(number,foo),		[]).
stc(eval:50,	_X is float(foo), 		type(evaluable,foo/0),		[]).
stc(eval:51,	X is abs(7), 			[[ X <-- 7 ]],			[]).
stc(eval:52,	X is abs(3-11), 		[[ X <-- 8 ]],			[]).
stc(eval:53,	X is abs(3.2-11.0), 		[[ X <-- 7.8 ]],		[]).
stc(eval:54,	_X is abs(_N), 			inst,				[]).
%stc(eval:55,	_X is abs(foo), 		type(number,foo),		[]).
stc(eval:55,	_X is abs(foo), 		type(evaluable,foo/0),		[]).
stc(eval:56,	(current_prolog_flag(
	  	  max_integer, MI),
	  	 _X is '+'(MI,1)),		eval(int_overflow),		[]).
stc(eval:57,	(current_prolog_flag(
	  	  max_integer, MI),
	  	 _X is '-'('+'(MI,1),1)),	eval(int_overflow),		[]).
stc(eval:58,	(current_prolog_flag(
	  	  max_integer, MI),
	  	 _X is '-'(/*-1*/-2,MI)),	eval(int_overflow),		[]).
		 % ISO allows min_integer = -(max_integer + 1)
stc(eval:59,	(current_prolog_flag(
	  	  max_integer, MI),
	  	 _X is '*'(MI,2)),		eval(int_overflow),		[]).
stc(eval:60,	(current_prolog_flag(
	  	  max_integer, MI),
	  	 R is float(MI)*2,
	  	 _X is floor(R)),		eval(int_overflow),		[]).

% 9.3.1.4
stc(pow: 1,	X is '**'(5,3),			[[ X =~~ 125.0 ]], 		[]).
stc(pow: 2,	X is '**'(-5.0,3),		[[ X =~~ -125.0 ]], 		[]).
stc(pow: 3,	X is '**'(5,-1),		[[ X =~~ 0.2 ]], 		[]).
stc(pow: 4,	_X is '**'(77,_N),		inst,		 		[]).
stc(pow: 5, 	_X is '**'(foo,2),		type(evaluable,foo/0),		[]).
stc(pow: 6, 	X is '**'(5,3.0),		[[ X =~~ 125.0 ]], 		[]).
stc(pow: 7, 	X is '**'(0,0.0),		[[ X =~~ 1.0 ]], 		[]).

% 9.3.2.4
stc(sin: 1,	X is sin(0.0),			[[ X <-- 0.0 ]], 		[]).
stc(sin: 2,	_X is sin(_N),			inst,		 		[]).
stc(sin: 3,	X is sin(0),			[[ X <-- 0.0 ]], 		[]).
stc(sin: 4, 	_X is sin(foo),			type(evaluable,foo/0),		[]).
stc(sin: 5, 	(PI is atan(1.0)*4,
		 X is sin(PI/2.0)),	   	[[  X =~~ 1, PI =~~ 3.14159 ]],	[]).

% 9.3.3.4
stc(cos: 1,	X is cos(0.0),			[[ X <-- 1.0 ]], 		[]).
stc(cos: 2,	_X is cos(_N),			inst,		 		[]).
stc(cos: 3,	X is cos(0),			[[ X <-- 1.0 ]], 		[]).
stc(cos: 4, 	_X is cos(foo),			type(evaluable,foo/0),		[]).
stc(cos: 5, 	(PI is atan(1.0)*4,
		 X is cos(PI/2.0)),	   	[[  X =~~ 0, PI =~~ 3.14159 ]],	[]).

% 9.3.4.4
stc(atan: 1,	X is atan(0.0),			[[ X <-- 0.0 ]], 		[]).
stc(atan: 2, 	PI is atan(1.0)*4,	   	[[ PI =~~ 3.14159 ]],		[]).
stc(atan: 3,	_X is atan(_N),			inst,		 		[]).
stc(atan: 4,	X is atan(0),			[[ X <-- 0.0 ]], 		[]).
stc(atan: 5, 	_X is atan(foo),		type(evaluable,foo/0),		[]).

% 9.3.5.4
stc(exp1: 1,	X is exp(0.0),			[[ X <-- 1.0 ]], 		[]).
%stc(exp1: 2, 	E is exp(1.0),		   	[[ E =~~ 2.7818 ]],		[]).
stc(exp1: 2, 	E is exp(1.0),		   	[[ E =~~ 2.71828 ]],		[]).
stc(exp1: 3,	_X is exp(_N),			inst,		 		[]).
stc(exp1: 4,	X is exp(0),			[[ X <-- 1.0 ]], 		[]).
stc(exp1: 5, 	_X is exp(foo),			type(evaluable,foo/0),		[]).

% 9.3.6.4
stc(log: 1,	X is log(1.0),			[[ X <-- 0.0 ]], 		[]).
%stc(log: 2, 	E is log(2.7818),	   	[[ E =~~ 1.0 ]],		[]).
stc(log: 2, 	E is log(2.71828),	   	[[ E =~~ 1.0 ]],		[]).
stc(log: 3,	_X is log(_N),			inst,		 		[]).
stc(log: 4,	_X is log(0),			eval(undefined), 		[]).
stc(log: 5, 	_X is log(foo),			type(evaluable,foo/0),		[]).
stc(log: 6,	_X is log(0.0),			eval(undefined), 		[]).

% 9.3.7.4
stc(sqrt: 1,	X is sqrt(0.0),			[[ X <-- 0.0 ]], 		[]).
stc(sqrt: 2, 	E is sqrt(1),		   	[[ E <-- 1.0 ]],		[]).
stc(sqrt: 3, 	E is sqrt(1.21),		[[ E =~~ 1.1 ]],		[]).
stc(sqrt: 4,	_X is sqrt(_N),			inst,		 		[]).
stc(sqrt: 5,	_X is sqrt(-1.0),		eval(undefined), 		[]).
stc(sqrt: 6, 	_X is sqrt(foo),		type(evaluable,foo/0),		[]).

% 9.4.1.4
stc(bit_rsh: 1,	X is '>>'(16, 2),		[[ X <-- 4 ]], 			[]).
stc(bit_rsh: 2,	X is '>>'(19, 2),		[[ X <-- 4 ]], 			[]).
stc(bit_rsh: 3,	X is '>>'(-16, 2),		[[(compl2 -> X = -4 ; true)]],	[]).
stc(bit_rsh: 4,	_X is '>>'(77, _N),		inst,	 			[]).
%stc(bit_rsh: 5,_X is '>>'(foo, 2),		type(integer,foo),		[]).
stc(bit_rsh: 5,	_X is '>>'(foo, 2),		type(evaluable,foo/0),		[]).
atc(bit_rsh: 6,	_X is '>>'(1.0, 2),		type(integer,1.0),		[]).

% 9.4.2.4
stc(bit_lsh: 1,	X is '<<'(16, 2),		[[ X <-- 64 ]], 		[]).
stc(bit_lsh: 2,	X is '<<'(19, 2),		[[ X <-- 76]], 			[]).
stc(bit_lsh: 3,	X is '<<'(-16, 2),		[[(compl2 -> X = -64 ; true)]],	[]).
stc(bit_lsh: 4,	_X is '<<'(77, _N),		inst,	 			[]).
%stc(bit_lsh: 5,_X is '<<'(foo, 2),		type(integer,foo),		[]).
stc(bit_lsh: 5,	_X is '<<'(foo, 2),		type(evaluable,foo/0),		[]).
atc(bit_lsh: 6,	_X is '<<'(1.0, 2),		type(integer,1.0),		[]).

% 9.4.3.4
stc(bit_and: 1,	X is '/\\'(10, 12),		[[ X <-- 8 ]], 			[]).
stc(bit_and: 2,	X is /\(10, 12),		[[ X <-- 8 ]], 			[]).
stc(bit_and: 3,	X is '/\\'(17*256+125, 255),	[[ X <-- 125 ]], 		[]).
stc(bit_and: 4,	X is /\(-10, 12),		[[(compl2 -> X = 4 ; true)]],	[]).
stc(bit_and: 5,	_X is '/\\'(77, _N),		inst,	 			[]).
%stc(bit_and: 6,_X is '/\\'(foo, 2),		type(integer,foo),		[]).
stc(bit_and: 6,	_X is '/\\'(foo, 2),		type(evaluable,foo/0),		[]).
atc(bit_and: 7,	_X is '/\\'(1.0, 2),		type(integer,1.0),		[]).

% 9.4.4.4
stc(bit_or: 1,	X is '\\/'(10, 12),		[[ X <-- 14 ]],			[]).
stc(bit_or: 2,	X is \/(10, 12),		[[ X <-- 14 ]],			[]).
stc(bit_or: 3,	X is '\\/'(125, 255),		[[ X <-- 255 ]], 		[]).
stc(bit_or: 4,	X is \/(-10, 12),		[[(compl2 -> X = -2 ; true)]],	[]).
stc(bit_or: 5,	_X is '\\/'(77, _N),		inst,	 			[]).
%stc(bit_or: 6,	_X is '\\/'(foo, 2),		type(integer,foo),		[]).
stc(bit_or: 6,	_X is '\\/'(foo, 2),		type(evaluable,foo/0),		[]).
atc(bit_or: 7,	_X is '\\/'(1.0, 2),		type(integer,1.0),		[]).

% 9.4.5.4
stc(bit_neg: 1,	X is '\\'('\\'(10)),		[[ X <-- 10 ]],			[]).
stc(bit_neg: 2,	X is \(\(10)),			[[ X <-- 10 ]],			[]).
stc(bit_neg: 4,	X is \(10),			[[(compl2 -> X = -11 ; true)]],	[]).
stc(bit_neg: 5,	_X is '\\'(_N),			inst,	 			[]).
atc(bit_neg: 6,	_X is '\\'(2.5),		type(integer,2.5),		[]).




% ---------------------------------------------------------------------------
% Programs and macros

% 6.3.4

program(fxops,[
(	:- op(100, fx, fx)
),(	:- op(100, xf, xf)
),(	:- op(100, xfx, xfx)
)]).

program(fyops,[
(	:- op(100, fy, fy)
),(	:- op(100, yf, yf)
),(	:- op(100, xfy, xfy)
),(	:- op(100, yfx, yfx)
)]).

% 6.3.7.1
macro(set_double_quotes(Flag), [pre((current_prolog_flag(double_quotes, Old),
				     set_prolog_flag(double_quotes, Flag))),
				post(set_prolog_flag(double_quotes, Old))]).

macro(double_quotes_ex1, txtin([
'(      current_prolog_flag(double_quotes, chars), atom_chars(\'jim\', "jim")',
';      current_prolog_flag(double_quotes, codes), atom_codes(\'jim\', "jim")',
';      current_prolog_flag(double_quotes, atom), \'jim\' == "jim"',
').'])).

macro(double_quotes_ex2, txtin([
'(      current_prolog_flag(double_quotes, chars), [] == ""',
';      current_prolog_flag(double_quotes, codes), [] == ""',
';      current_prolog_flag(double_quotes, atom), \'\' == ""',
').'])).

% 7.8.3.4
program(call,
[(
        b(X) :-
                Y = (write(X), X),
                call(Y)
),(
        a(1)
),(
        a(2)
)]).

% 7.8.4.4
program(cut,
[(
        twice(!) :- write('C ')
),(
        twice(true) :- write('Moss ')
),(
        goal((twice(_), !))
),(
        goal(write('Three '))
)]).

% 7.8.9.4
program(catch,
[(
        foo(X) :-
                Y is X * 2, throw(test(Y))
),(
        bar(X) :-
                X = Y, throw(Y)
),(
        coo(X) :-
                throw(X)
),(
        car(X) :-
                X = 1, throw(X)
),(
        g :-
                catch(p, _B, write(h2)),
                coo(c)
),(
        p
),(     p :-
                throw(b)
)]).

% 8.5.3.4
macro(list_of(N, A, L),
      [prog(build_list),pre(my_list_of(N,A,L))]).

program(build_list,
[(
        my_list_of(0, _, [])
),(
        my_list_of(N, A, [A|L]) :-
                  N > 0, N1 is N-1,
                  my_list_of(N1, A, L)
)]
       ).


% 8.8.1.4
program(clause, [
(       :- dynamic(cat/0)
),(     cat

),(     :- dynamic(dog/0)
),(     dog :- true

),(     elk(X) :- moose(X)

),(     :- dynamic(legs/2)
),(     legs(A,6) :- insect(A)
),(     legs(A,7) :- A, call(A)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)
)
        ]).

% 8.9.1.4
program(dynpreds1, [
(       :- dynamic(legs/2)
),(     legs(A, 6) :- insect(A)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)
)
]).

program(dynpreds2, [
(       :- dynamic(legs/2)
),(     legs(octopus, 8)
),(     legs(A, 6) :- insect(A)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)
)
]).

program(dynpreds3, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(octopus, 8)
),(     legs(A, 6) :- insect(A)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)
)
]).

program(dynpreds4, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(octopus, 8)
),(     legs(A, 6) :- insect(A)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
)
]).

% 8.9.2.4
program(dynpreds5, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(octopus, 8)
),(     legs(A, 6) :- insect(A)
),(     legs(spider, 8)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
)
]).

program(dynpreds6, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(octopus, 8)
),(     legs(A, 6) :- insect(A)
),(     legs(spider, 8)
),(     legs(B, 2) :- bird(B)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
)
]).

program(dynpreds7, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(octopus, 8)
),(     legs(A, 6) :- insect(A)
),(     legs(spider, 8)
),(     legs(B, 2) :- bird(B)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
),(     foo(X) :- call(@X) -> call(@X)
)
]).

% 8.9.3.4
program(dynpreds8, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(A, 6) :- insect(A)
),(     legs(spider, 8)
),(     legs(B, 2) :- bird(B)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
),(     foo(X) :- call(@X) -> call(@X)
)
]).

program(dynpreds9, [
(       :- dynamic(legs/2)
),(     legs(A, 4) :- animal(A)
),(     legs(A, 6) :- insect(A)
),(     legs(spider, 8)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
),(     foo(X) :- call(@X) -> call(@X)
)
]).

program(dynpreds10, [
(       :- dynamic(legs/2)

),(     :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
),(     foo(X) :- call(@X) -> call(@X)
)
]).

program(dynpreds11, [
(       :- dynamic(legs/2)

),(     :- dynamic(insect/1)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
),(     foo(X) :- call(@X) -> call(@X)
)
]).

program(dynpreds12, [
(       :- dynamic(legs/2)

),(     :- dynamic(insect/1)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(@X), call(@X)
)
]).

% 8.9.4.4
program(abolish, [
(       :- dynamic(insect/1)
),(     insect(ant)
),(     insect(bee)

),(     :- dynamic(foo/1)
),(     foo(X) :- call(X), call(X)
),(     foo(X) :- call(X) -> call(X)

),(     bar(X) :- true
)
                 ]).

macro(over_arity(N),
      [pre((current_prolog_flag(max_arity,A), N is A + 1))]).

% 8.10.3.4
program(member, [
(       my_member(X, [X|_])
),(     my_member(X, [_|L]) :-
                my_member(X, L)
)
                ]).

macro(stream_position(Pos), 
      [pre((open(bar,write,S,[reposition(true)]), stream_property(S, position(Pos)))), clean(bar)]).

macro(closed_instream(S, Opts),
      [maketxt(foo,''),pre((open('foo', read, S, Opts), close(S)))]).
macro(closed_outstream(S, Opts),
      [pre((open('foo', write, S, Opts), close(S))), clean(foo)]).

macro(txtin_foo256,
      txtin(['foo(',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,',
	'1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1).'])).

macro(clear_chconv,
      post((current_char_conversion(A,_B), char_conversion(A, A), fail; true))).

macro(some_chconvs, [pre((char_conversion('%', '+'), char_conversion('^', '\''))),clear_chconv]).

macro(long_atom(A),
      [prog(build_list),pre((my_list_of(Len,0'a,L),atom_codes(A, L)))]) :-
	context_info(max_atom_length, Max), Len is Max*2//3.

macro(txtin_maxmin(Type, Val), 
      [pre(((current_prolog_flag(Type, Val) -> true ; Val = 0), number_codes(Val, L), atom_codes(Atm, L))), txtin([Atm,'.'])]).
